Package: scDesign3
Type: Package
Title: A unified framework of realistic in silico data generation and
        statistical model inference for single-cell and spatial omics
Version: 1.10.0
Authors@R: 
    c(person("Dongyuan", "Song", , "dongyuansong@ucla.edu", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-1114-1215")),
    person("Qingyang", "Wang", , "qw802@g.ucla.edu",  role = c("aut"),
           comment = c(ORCID = "0000-0002-1051-609X")),
    person("Chenxin", "Jiang", , "cflorajiang@g.ucla.edu",  role = c("aut"),
           comment = c(ORCID = "0009-0005-7369-4116")))
Description: We present a statistical simulator, scDesign3, to generate realistic single-cell and spatial omics data, including various cell states, experimental designs, and feature modalities, by learning interpretable parameters from real data. Using a unified probabilistic model for single-cell and spatial omics data, scDesign3 infers biologically meaningful parameters; assesses the goodness-of-fit of inferred cell clusters, trajectories, and spatial locations; and generates in silico negative and positive controls for benchmarking computational tools.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: false
Depends: R (>= 4.3.0)
Imports: dplyr, tibble, stats, methods, mgcv, gamlss, gamlss.dist,
        SummarizedExperiment, SingleCellExperiment, mclust, mvtnorm,
        parallel, pbmcapply, umap, ggplot2, irlba, viridis,
        BiocParallel, matrixStats, Matrix, sparseMVN, coop
Suggests: mvnfast, igraph, rvinecopulib, knitr, rmarkdown, testthat (>=
        3.0.0), RefManageR, sessioninfo, BiocStyle
biocViews: Software, SingleCell, Sequencing, GeneExpression, Spatial
URL: https://github.com/SONGDONGYUAN1994/scDesign3
BugReports: https://github.com/SONGDONGYUAN1994/scDesign3/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/scDesign3
git_branch: RELEASE_3_23
git_last_commit: c1f38d2
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:27:40 UTC; biocbuild
Author: Dongyuan Song [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1114-1215>),
  Qingyang Wang [aut] (ORCID: <https://orcid.org/0000-0002-1051-609X>),
  Chenxin Jiang [aut] (ORCID: <https://orcid.org/0009-0005-7369-4116>)
Maintainer: Dongyuan Song <dongyuansong@ucla.edu>

# Generated by roxygen2: do not edit by hand

S3method(predict,gamlss)
export(ba)
export(construct_data)
export(extract_para)
export(fit_copula)
export(fit_marginal)
export(ga)
export(gamlss.ba)
export(gamlss.ga)
export(perform_lrt)
export(plot_reduceddim)
export(scdesign3)
export(simu_new)
export(sparse_cov)
import(gamlss)
import(ggplot2)
import(mclust)

## scDesign3

------------------------------------------------------------------------

The R package **scDesign3** is an all-in-one single-cell data simulation tool by using reference datasets with different cell states (cell types, trajectories or and spatial coordinates), different modalities (gene expression, chromatin accessibility, protein abundance, DNA methylation, etc), and complex experimental designs. The transparent parameters enable users to alter models as needed; the model evaluation metrics (AIC, BIC) and convenient visualization function help users select models. <span style="color:blue"> **Detailed tutorials that illustrate various functionalities of scDesign3 are available at this [website](https://songdongyuan1994.github.io/scDesign3/docs/index.html)**</span>. The following illustration figure summarizes the usage of scDesign3:

<img src="man/figures/scDesign3_illustration.png" width="600"/>


To find out more details about **scDesign3**, you can check out our manuscript on Nature Biotechnology:

[Song, D., Wang, Q., Yan, G. *et al.* scDesign3 generates realistic in silico data for multimodal single-cell and spatial omics. *Nat Biotechnol* **42**, 247–252 (2024).](https://www.nature.com/articles/s41587-023-01772-1)

The computational time is quadratic to the number of features used in copula modeling. Reducing this number will greatly speed up the calculation.

Please note that the parallel computing of scDesign3 is mainly designed for **UNIX OS**; be careful when you set `n_cores`. Please note that you should consider **the balance** between `n_cores` and your ROM (memory). Simply increasing the number of cores without the increase of memory will slow down or froze your program. We recommend that you should allocate at least 1 GB for 1 core.

# Table of contents
1. [Installation](#installation-)
2. [Quick Start](#quick-start)
3. [Tutorials](#tutorials)
4. [Contact](#contact)
5. [Related Manuscripts](#related-manuscripts)


## Installation<a name="installation-"></a>

To install the development version from GitHub, please run:

``` r
if (!require("devtools", quietly = TRUE))
    install.packages("devtools")
devtools::install_github("SONGDONGYUAN1994/scDesign3")
```

We are now working on submitting it to Bioconductor and will provide the link once online.

## Quick Start<a name="quick-start"></a>

The following code is a quick example of running our simulator. The function `scdesign3()` takes in a `SinglecellExperiment` object with the cell covariates(such as cell types, pseudotime, or spatial coordinates) stored in the `colData` of the `SinglecellExperiment` object. For more details on the `SinlgeCellExperiment` object, please check on its [Bioconductor link](https://bioconductor.org/packages/release/bioc/html/SingleCellExperiment.html).

``` r
example_simu <- scdesign3(
    sce = example_sce,
    assay_use = "counts",
    celltype = "cell_type",
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    mu_formula = "s(pseudotime, k = 10, bs = 'cr')",
    sigma_formula = "s(pseudotime, k = 5, bs = 'cr')",
    family_use = "nb",
    n_cores = 2,
    correlation_function = "default",
    usebam = FALSE,
    corr_formula = "1",
    copula = "gaussian",
    fastmvn = FALSE,
    DT = TRUE,
    pseudo_obs = FALSE,
    family_set = c("gauss", "indep"),
    important_feature = "all",
    nonnegative = TRUE,
    return_model = FALSE,
    nonzerovar = FALSE,
    parallelization = "mcmapply",
    BPPARAM = NULL,
    trace = FALSE
  )
```

The parameters of `scdesign3()` are:

- `sce`: A SingleCellExperiment object.
- `assay_use`: A string which indicates the assay you will use in the sce. Default is 'counts'.
- `celltype`: A string of the column name of the cell type variable in the colData of the sce. Default is 'cell_type'. The cell type variable in the colData of the sce should be a factor variable. Use NULL if there is no column in the colData that contains the cell-type information. 
- `pseudotime`: A string or a string vector of the name of pseudotime and (if exist) multiple lineages. Default is NULL. If the data only has one lineage, then this parameter should be the column name of the pseudotime variable in the colData of the sce. If the data has multiple lingaes, then this parameter be the column names of the pseudotime variables for each lineage and the variables indicating which lineage that a cell belongs to. The pseudotime variables should be continuous numeric variables.
- `spatial`: A length-two string vector of the column names of spatial coordinates in the colData of sce. Default is NULL. 
- `other_covariate`: A string or a string vector of the other covariates in the colData of sce you want to include in the data. For example, you can put the column names of the batch variables and/or condition variables in the colData of sce here if your sce contains these information and you want to include these variables in`mu_formula` or `sigma_formula` or `corr_formula`.
- `mu_formula`: A string of the mu parameter formula for fitting each gene's marginal distribution.
- `sigma_formula`: A string of the sigma parameter formula for fitting each gene's marginal distribution.
- `family_use`: A string of the marginal distribution you want to use when fitting each gene's marginal distribution. Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian'.
- `n_cores`: An integer. The number of cores to use.
- `correlation_function`: A string. If 'default', the function from Rfast; if 'coop', the function from coop, which calls BLAS.
- `usebam`: A logic variable. If TRUE, use bam (generalized additive models for very large datasets) for acceleration.
- `edf_flexible`: A logic variable. If TRUE, the degree of freedom for each gene's regression model will be automatically selected for acceleration.
- `corr_formula`: A string of the correlation structure. For example, if you want to obtain a correlation structure for each cell type, then this parameter should be the column name of the cell type variable in the colData of sce.
- `copula`: A string of the copula choice. Must be one of 'gaussian' or 'vine'. Default is 'gaussian'. Note that vine copula may have better modeling of high-dimensions, but can be very slow when features are >1000.
- `fastmvn`: An logical variable. If TRUE, the sampling of multivariate Gaussian is done by mvnfast, otherwise by mvtnorm. Default is FALSE. It only matters for Gaussian copula.
- `DT`: A logic variable. If TRUE, perform the distributional transformation to make the discrete data 'continuous'. This is useful for discrete distributions (e.g., Poisson, NB). Default is TRUE. Note that for continuous data (e.g., Gaussian), DT does not make sense and should be set as FALSE.
- `pseudo_obs`: A logic variable. If TRUE, use the empirical quantiles instead of theoretical quantiles for fitting copula. Default is FALSE.
- `family_set`: A string or a string vector of the bivariate copula families. Default is c("gauss", "indep").
- `important_feature`: A string or vector which indicates whether a gene will be used in correlation estimation or not. If this is a string, then this string must be "all" or "auto", which indicates that all genes are used or the genes will be automatically selected based on the proportion of zero expression across cells for each gene. Gene with zero proportion greater than 0.8 will be excluded form gene-gene correlation estimation. If this is a vector, then this should be a logical vector with length equal to the number of genes in sce. TRUE in the logical vector means the corresponding gene will be included in gene-gene correlation estimation and FALSE in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation. The default value is "all".
- `nonnegative`: A logical variable. If TRUE, values < 0 in the synthetic data will be converted to 0. Default is TRUE (since the expression matrix is nonnegative).
- `return_model`: A logic variable. If TRUE, the marginal models and copula models will be returned. Default is FALSE.
- `nonzerovar`: A logical variable. If TRUE, for any gene with zero variance, a cell will be replaced with 1. This is designed for avoiding potential errors, for example, PCA.
- `parallelization`: A string indicating the specific parallelization function to use. Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package parallel,BiocParallel, and pbmcapply respectively. The default value is 'mcmapply'.
- `BPPARAM`: A MulticoreParam object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply', this parameter must be NULL. When the parameter parallelization = 'bpmapply', this parameter must be one of the MulticoreParam object offered by the package 'BiocParallel. The default value is NULL.
- `TRACE`: A logic variable. If TRUE, the warning/error log and runtime for gam/gamlss will be returned, FALSE otherwise. Default is FALSE.

The output of `scdesign3()` is a list which includes:

-   `new_count`: This is the synthetic count matrix generated by `scdesign3()`.
-   `new_covariate`:
    -   If the parameter `ncell` is set to a number that is different from the number of cells in the input data, this will be a matrix that has the new cell covariates that are used for generating new data.
    -   If the parameter `ncell` is the default value, this will be `NULL`.
-   `model_aic`: This is a vector include the genes' marginal models' AIC, fitted copula's AIC, and total AIC, which is the sum of the previous two.
-   `model_bic`: This is a vector include the genes' marginal models' BIC, fitted copula's BIC, and total BIC, which is the sum of the previous two.
-   `marginal_list`:
    -   If the parameter `return_model` is set to `TRUE`, this will be a list which contains the fitted gam or gamlss model for all genes in the input data. This may greatly increase the object size.
    -   If the parameter `return_model` is set to the default value `FALSE`, this will be `NULL`.
-   `corr_list`:
    -   If the parameter `return_model` is set to `TRUE`, this will be a list which contains either a correlation matrix (when `copula = "gaussian"`) or the fitted Vine copula (when `copula = "vine"`) for each user specified correlation groups (based on the parameter `corr_by`).
    -   If the parameter `return_model` is set to the default value `FALSE`, this will be `NULL`.

For more details about the `mu_formula` and `sigma_formula` formula specification, please check online materials about the package [mgcv](https://cran.r-project.org/web/packages/mgcv/index.html). Technically speaking, you can try any formulas as long as they are available for **mgcv**.

## Tutorials<a name="tutorials"></a>

For all detailed tutorials, please check the [website](https://songdongyuan1994.github.io/scDesign3/docs/index.html). The tutorials will demonstrate the applications of **scDesign3** from the following four perspectives: data simulation, model parameters, model selection, and model alteration.

-   Data simulation
    -   [Simulate datasets with cell type and modified cell-type proportions](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-cellType-vignette.html)
    -   [Simulate datasets with cell library size](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-librarySize-vignette.html)
    -   [Simulate datasets with multiple lineages](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-multipleLineages-vignette.html)
    -   [Simulate spatial transcriptomic data](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-spatial-vignette.html)
    -   [Simulate spot-resolution spatial data for cell-type deconvolution](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-spatial-deconvolution.html)
    -   [Simulate single-cell ATAC-seq data](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-scATACseq-vignette.html)
    -   [Simulate CITE-seq data](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-CITEseq-vignette.html)
    -   [Simulate multi-omics data from multiple single-omic datasets](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-multiomics-vignette.html)
-   Model parameter
    -   [scDesign3 introduction](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-introduction-vignette.html)
    -   [scDesign3 marginal distribution for genes](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-marginal-vignette.html)
    -   [Compare Gaussian copula and Vine copula](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-copulaCompare-vignette.html)
    -   [Parallelization in scDesign3](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-parallelization-vignette.html)
-   Model selection
    -   [Evaluate clustering goodness-of-fit by scDesign3](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-clusterGOF-vignette.html)
    -   [Evaluate pseudotime goodness-of-fit by scDesign3](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-pseudotimeGOF-vignette.html)
-   Model alteration
    -   [Simulate datasets with/without batch effect](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-batchEffect-vignette.html)
    -   [Simulate datasets with/without condition effect](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-conditionEffect-vignette.html)
    -   [Simulate datasets for DE test](https://songdongyuan1994.github.io/scDesign3/docs/articles/scDesign3-DEanalysis-vignette.html)

## Contact<a name="contact"></a>

Any questions or suggestions on `scDesign3` are welcomed! Please report it on [issues](https://github.com/SONGDONGYUAN1994/scDesign3/issues), or contact Dongyuan Song ([dongyuansong\@ucla.edu](mailto:dongyuansong@ucla.edu){.email}) or Qingyang Wang ([qw802\@g.ucla.edu](mailto:qw802@g.ucla.edu){.email}).

## Changelog
-   2024-11-07
    - Add options for correlation estimation. The alternative is from R package [coop](https://cran.rstudio.com/web/packages/coop/index.html), which requires BLAS
    
-   2024-06-22 Important changes
    - Add sparse Gaussian copula for `fit_copula`
    - Add automatic k selection for `fit_marginal`
    
## Related Manuscripts<a name="related-manuscripts"></a>
-   The original **scDesign3** paper
    -   **scDesign3**: [Song, D., Wang, Q., Yan, G. *et al.* scDesign3 generates realistic in silico data for multimodal single-cell and spatial omics. *Nat Biotechnol* **42**, 247–252 (2024).](https://www.nature.com/articles/s41587-023-01772-1)
-   The predecessors of **scDesign3**
    -   **scDesign**: [Li, W. V., & Li, J. J. (2019). A statistical simulator scDesign for rational scRNA-seq experimental design. *Bioinformatics*, **35**(14), i41-i50.](https://academic.oup.com/bioinformatics/article/35/14/i41/5529133)
    -   **scDesign2**: [Sun, T., Song, D., Li, W. V., & Li, J. J. (2021). scDesign2: a transparent simulator that generates high-fidelity single-cell gene expression count data with gene correlations captured. *Genome biology*, **22**(1), 1-37.](https://link.springer.com/article/10.1186/s13059-021-02367-2)
-   The simulator for single-cell multi-omics reads developed by our lab member Guanao Yan
    -   **scReadSim**: [Yan, G., Song, D. & Li, J.J. scReadSim: a single-cell RNA-seq and ATAC-seq read simulator. *Nat Commun* **14**, 7482 (2023)](https://doi.org/10.1038/s41467-023-43162-w)

#' Simulate new data
#'
#' \code{simu_new} generates new simulated data based on fitted marginal and copula models.
#'
#' The function takes the new covariate (if use) from \code{\link{construct_data}},
#' parameter matrices from \code{\link{extract_para}} and multivariate Unifs from \code{\link{fit_copula}}.
#'
#' @param sce A \code{SingleCellExperiment} object.
#' @param assay_use A string which indicates the assay you will use in the sce. Default is 'counts'.
#' @param mean_mat A cell by feature matrix of the mean parameter.
#' @param sigma_mat A cell by feature matrix of the sigma parameter.
#' @param zero_mat A cell by feature matrix of the zero-inflation parameter.
#' @param quantile_mat A cell by feature matrix of the multivariate quantile.
#' @param copula_list A list of copulas for generating the multivariate quantile matrix. If provided, the \code{quantile_mat} must be NULL.
#' @param n_cores An integer. The number of cores to use.
#' @param fastmvn An logical variable. If TRUE, the sampling of multivariate Gaussian is done by \code{mvnfast}, otherwise by \code{mvtnorm}. Default is FALSE.
#' @param family_use A string of the marginal distribution.
#' Must be one of 'poisson', "binomial", 'nb', 'zip', 'zinb' or 'gaussian'.
#' @param nonnegative A logical variable. If TRUE, values < 0 in the synthetic data will be converted to 0. Default is TRUE (since the expression matrix is nonnegative).
#' @param nonzerovar A logical variable. If TRUE, for any gene with zero variance, a cell will be replaced with 1. This is designed for avoiding potential errors, for example, PCA.
#' @param input_data A input count matrix.
#' @param new_covariate A data.frame which contains covariates of targeted simulated data from  \code{\link{construct_data}}.
#' @param important_feature important_feature A string or vector which indicates whether a gene will be used in correlation estimation or not. If this is a string, then
#' this string must be either "all" (using all genes) or "auto", which indicates that the genes will be automatically selected based on the proportion of zero expression across cells
#' for each gene. Gene with zero proportion greater than 0.8 will be excluded form gene-gene correlation estimation. If this is a vector, then this should
#' be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
#' gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
#' The default value for is "all".
#' @param parallelization A string indicating the specific parallelization function to use.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#' @param filtered_gene A vector or NULL which contains genes that are excluded in the marginal and copula fitting 
#' steps because these genes only express in less than two cells. This can be obtain from  \code{\link{construct_data}}
#' @return A feature by cell matrix of the new simulated count (expression) matrix or sparse matrix.
#' @examples
#'   data(example_sce)
#'   my_data <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "1"
#'   )
#'   my_marginal <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#'   )
#'   my_copula <- fit_copula(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   marginal_list = my_marginal,
#'   family_use = c(rep("nb", 5), rep("zip", 5)),
#'   copula = "vine",
#'   n_cores = 1,
#'   input_data = my_data$dat
#'   )
#'   my_para <- extract_para(
#'     sce = example_sce,
#'     marginal_list = my_marginal,
#'     n_cores = 1,
#'     family_use = c(rep("nb", 5), rep("zip", 5)),
#'     new_covariate = my_data$new_covariate,
#'     data = my_data$dat
#'   )
#'   my_newcount <- simu_new(
#'   sce = example_sce,
#'   mean_mat = my_para$mean_mat,
#'   sigma_mat = my_para$sigma_mat,
#'   zero_mat = my_para$zero_mat,
#'   quantile_mat = NULL,
#'   copula_list = my_copula$copula_list,
#'   n_cores = 1,
#'   family_use = c(rep("nb", 5), rep("zip", 5)),
#'   input_data = my_data$dat,
#'   new_covariate = my_data$new_covariate,
#'   important_feature = my_copula$important_feature,
#'   filtered_gene = my_data$filtered_gene
#'   )
#'
#' @export simu_new

simu_new <- function(sce,
                     assay_use = "counts",
                     mean_mat,
                     sigma_mat,
                     zero_mat,
                     quantile_mat = NULL,
                     copula_list,
                     n_cores,
                     fastmvn = FALSE,
                     family_use,
                     nonnegative = TRUE,
                     nonzerovar = FALSE,
                     input_data,
                     new_covariate,
                     important_feature = "all",
                     parallelization = "mcmapply",
                     BPPARAM = NULL,
                     filtered_gene){
  if(!is.null(quantile_mat) & !is.null(copula_list)) {
    stop("You can only provide either the quantile_mat or the copula_list!")
  }

  # check if user inputted new covariates
  data_temp <- input_data[,colnames(new_covariate), drop = FALSE]
  if(identical(data_temp, new_covariate)){
    new_covariate <- NULL
  }
  
  qc_gene_idx <- which(!rownames(sce) %in% filtered_gene)
  if(length(family_use) != 1){
    family_use <- family_use[qc_gene_idx]
  }
 
    if(!is.null(quantile_mat)) {
      message("Multivariate quantile matrix is provided")
      
      newmvn_full <- matrix(NA, nrow = dim(quantile_mat)[1], ncol = dim(sce)[1])
      rownames(newmvn_full) <- rownames(quantile_mat)
      colnames(newmvn_full) <- rownames(sce)
      newmvn_full[rownames(quantile_mat), colnames(quantile_mat)] <- quantile_mat
      quantile_mat <- as.matrix(newmvn_full)
      
      
    } else {
      message("Use Copula to sample a multivariate quantile matrix")
      
      group_index <- unique(input_data$corr_group)
      corr_group <- as.data.frame(input_data$corr_group)
      colnames(corr_group) <- "corr_group"
      ngene <- length(qc_gene_idx)
      if (is.null(new_covariate)) {
        new_corr_group <- NULL
      } else{
        new_corr_group <- as.data.frame(new_covariate$corr_group)
        colnames(new_corr_group) <- "corr_group"
      }
      ind <- group_index[1] == "ind"
      newmvn.list <-
        lapply(group_index, function(x,
                                     sce,
                                     corr_group,
                                     new_corr_group,
                                     ind,
                                     n_cores,
                                     copula_list) {
          message(paste0("Sample Copula group ", x, " starts"))
          curr_index <- which(corr_group[, 1] == x)
          if (is.null(new_covariate)) {
            curr_ncell <- length(curr_index)
            curr_ncell_idx <- curr_index
          } else{
            curr_ncell <- length(which(new_corr_group[, 1] == x))
            curr_ncell_idx <-which(new_corr_group[, 1] == x)
            #paste0("Cell", which(new_corr_group[, 1] == x))
          }
          cor.mat <- copula_list[[x]]
          
          if(curr_ncell == 0) {
            new_mvu <- NULL
          } else {
            if (methods::is(cor.mat, "matrix") | methods::is(cor.mat, "dsCMatrix")) {
              #message(paste0("Group ", group_index, " Start"))
              
              #message("Sample MVN")
              #sample from mvn for important genes only
              corr_gene_idx <- apply(cor.mat, 2, function(x) length(which(x < 1e-5)) != length(x)-1)
              corr_gene <- colnames(cor.mat)[which(corr_gene_idx)]
              if(length(corr_gene)!=0) {
                new_mvn_important <- sampleMVN(n = curr_ncell,
                                               Sigma = cor.mat[corr_gene, corr_gene],
                                               n_cores = n_cores,
                                               fastmvn = fastmvn)
              
              colnames(new_mvn_important) <- corr_gene} else {
                new_mvn_important <- NULL
              }
              #message("MVN Sampling End")
              ind_gene <- colnames(cor.mat)[which(corr_gene_idx==FALSE)]
              if(length(ind_gene) > 0){
                new_mvn_non_important <- lapply(ind_gene, function(x) return(stats::rnorm(n = curr_ncell)))
                new_mvn_non_important_mat <- do.call("cbind",new_mvn_non_important)
                colnames(new_mvn_non_important_mat) <- ind_gene
                
                mvnrvq <- apply(new_mvn_non_important_mat, 2, stats::pnorm)
                new_mvu <- cbind(new_mvn_important, mvnrvq)
                new_mvu <- new_mvu[,colnames(cor.mat)]
              }else{
                new_mvu <- new_mvn_important
              }

              rownames(new_mvu) <- curr_ncell_idx
            } else if (methods::is(cor.mat, "vinecop")) {
              new_mvu <- matrix(0, nrow = curr_ncell, ncol = ngene)
              #message("Sampling Vine Copula Starts")
              mvu <- rvinecopulib::rvinecop(
                curr_ncell,
                vine = cor.mat,
                cores = n_cores,
                qrng = TRUE
              )
              new_mvu[, which(important_feature)] <- mvu
              if(length(which(important_feature)) != ngene){
                cor.mat <- diag(rep(1, length(which(!important_feature))))
                mvu2 <- sampleMVN(n = curr_ncell,
                                  Sigma = cor.mat,
                                  n_cores = n_cores,
                                  fastmvn = fastmvn)
                new_mvu[, which(!important_feature)] <- mvu2
              }
              #message("Sampling Vine Copula Ends")
              rownames(new_mvu) <- curr_ncell_idx
            } else if (ind) {
              "Use independent copula (random Unif)."
              new_mvu <-
                matrix(data = stats::runif(curr_ncell * ngene),
                       nrow = curr_ncell)
              rownames(new_mvu) <- curr_ncell_idx
            } else{
              stop("Copula must be one from 'vine' or 'gaussian', or assume gene-gene is independent")
            }
          }
          return(
            list(
              new_mvu = new_mvu
            )
          )
        }, sce = sce, ind = ind, n_cores = n_cores, corr_group = corr_group, new_corr_group = new_corr_group, copula_list = copula_list)
      
      newmvn <-
        do.call(rbind, lapply(newmvn.list, function(x)
          x$new_mvu))
      newmvn[as.numeric(rownames(newmvn)),] <- newmvn
      rownames(newmvn) <- as.character(1:dim(newmvn)[1])
      colnames(newmvn) <- rownames(sce)[qc_gene_idx]
      newmvn_full <- matrix(NA, nrow = dim(newmvn)[1], ncol = dim(sce)[1])
      rownames(newmvn_full) <- rownames(newmvn)
      colnames(newmvn_full) <- rownames(sce)
      newmvn_full[rownames(newmvn), colnames(newmvn)] <- newmvn
      quantile_mat <- as.matrix(newmvn_full)
    }
    
    mat_function <- function(x, y) {
      
      idx <- which(mean_mat[,x] !=0)
      para_mat <- cbind(mean_mat[idx, x], sigma_mat[idx, x], quantile_mat[idx, x], zero_mat[idx, x])
      
      if (y == "binomial") {
        qfvec <- stats::qbinom(p = para_mat[, 3], prob = para_mat[, 1], size = 1)
      } else if (y == "poisson") {
        
        qfvec <- stats::qpois(p = para_mat[, 3], lambda = para_mat[, 1])
      } else if (y == "gaussian") {
        qfvec <-
          gamlss.dist::qNO(p = para_mat[, 3],
                           mu = para_mat[, 1],
                           sigma = abs(para_mat[, 2]))
      } else if (y == "nb") {
        qfvec <-
          gamlss.dist::qNBI(p = para_mat[, 3],
                            mu = para_mat[, 1],
                            sigma = para_mat[, 2])
      } else if (y == "zip") {
        qfvec <-
          gamlss.dist::qZIP(p = para_mat[, 3],
                            mu = para_mat[, 1],
                            sigma = ifelse(para_mat[, 4] != 0, para_mat[, 4],  2.2e-16))## Avoid zero zero-inflated prob
      } else if (y == "zinb") {
        
        qfvec <-
          gamlss.dist::qZINBI(p = para_mat[, 3],
                              mu = para_mat[, 1],
                              sigma = para_mat[, 2],
                              nu = ifelse(para_mat[, 4] != 0, para_mat[, 4],  2.2e-16))
      } else {
        stop("Distribution of gamlss must be one of gaussian, poisson, nb, zip or zinb!")
      }
      
      #message(paste0("Gene ", x , " End!"))
      
      r <- as.vector(qfvec)
      if(length(r) < total_cells){
        new_r <- rep(0, total_cells)
        new_r[idx] <- r
        names(new_r) <- cell_names
        r <- new_r
      }
      r
    }

  


  ## New count
  paraFunc <- parallel::mcmapply
  if(.Platform$OS.type == "windows"){
    BPPARAM <- BiocParallel::SnowParam()
    parallelization <- "bpmapply"
  }
  if(parallelization == "bpmapply"){
    paraFunc <- BiocParallel::bpmapply
  }
  if(parallelization == "pbmcmapply"){
    paraFunc <- pbmcapply::pbmcmapply
  }

  if(is.null(new_covariate)){
    total_cells <- dim(sce)[2]
    cell_names <- colnames(sce)
  }else{
    total_cells <- dim(new_covariate)[1]
    cell_names <- rownames(new_covariate)
  }
  if(parallelization == "bpmapply"){
    if(class(BPPARAM)[1] != "SerialParam"){
    #if(!is(BPPARAM, "SerialParam")){
      BPPARAM$workers <- n_cores
    }
    mat <-  paraFunc(mat_function, x = seq_len(dim(sce)[1])[qc_gene_idx], y = family_use, SIMPLIFY = TRUE, BPPARAM = BPPARAM)
  }else{
    mat <- paraFunc(mat_function, x = seq_len(dim(sce)[1])[qc_gene_idx], y = family_use, SIMPLIFY = TRUE
                  , mc.cores = n_cores
                   )
 }
  new_count <- mat #simplify2array(mat)
  rownames(new_count) <- cell_names
  colnames(new_count) <- rownames(sce)[qc_gene_idx]

  if(length(qc_gene_idx) < dim(sce)[1]){
    temp_count <- matrix(0, total_cells, dim(sce)[1])
    rownames(temp_count) <- cell_names
    colnames(temp_count) <- rownames(sce)
    temp_count[rownames(new_count),colnames(new_count)] <- new_count
    new_count <- temp_count
  }
  new_count <- as.matrix(t(new_count))

  if(nonnegative) new_count[new_count < 0] <- 0

  if(nonzerovar) {
    row_vars <- matrixStats::rowVars(new_count[qc_gene_idx,])
    if(sum(row_vars == 0) > 0) {
      message("Some genes have zero variance. Replace a random one with 1.")
      row_vars_index <- which(row_vars == 0)
      col_index <- seq_len(dim(new_count)[2])
      for(i in row_vars_index) {
        new_count[i, sample(col_index, 1)] <- 1
      }
    }
  }
  new_count <- as.matrix(new_count)
  if(methods::is(SummarizedExperiment::assay(sce, assay_use), "sparseMatrix")){
    new_count<- Matrix::Matrix(new_count, sparse = TRUE)
  }

  return(new_count)
}

#context("Run scDesign3")
library(scDesign3)

test_that("Run scDesign3", {
  data(example_sce)
  my_data <- construct_data(
    sce = example_sce,
    assay_use = "counts",
    celltype = "cell_type",
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    corr_by = "cell_type"
  )

  my_data2 <- construct_data(
    sce = example_sce,
    assay_use = "counts",
    celltype = "cell_type",
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    corr_by = "pseudotime",
    ncell = 10000
  )

  my_marginal1 <- fit_marginal(
    data = my_data,
    mu_formula = "1",
    sigma_formula = "1",
    family_use = "nb",
    n_cores = 1,
    usebam = FALSE
  )

   my_marginal2 <- fit_marginal(
    data = my_data,
    mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
    sigma_formula = "1",
    family_use = "nb",
    n_cores = 1,
    usebam = FALSE
  )

  my_fit1 <- lapply(my_marginal1, function(x)x$fit)
  my_fit2 <- lapply(my_marginal2, function(x)x$fit)

  my_pvalue <- perform_lrt(my_fit2, my_fit1)

  my_marginal3 <- fit_marginal(
    data = my_data,
    mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
    sigma_formula = "1", #s(pseudotime, bs = 'cr', k = 3)
    family_use = c(rep("nb", 9), rep("zip", 1)),
    n_cores = 2,
    usebam = TRUE, trace = TRUE, simplify = TRUE
  )

  my_copula <- fit_copula(
    sce = example_sce,
    assay_use = "counts",
    marginal_list = my_marginal3,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    copula = "vine",
    n_cores = 1,
    input_data = my_data$dat
  )

  my_quantile_mat <- fit_copula(
    sce = example_sce,
    assay_use = "counts",
    empirical_quantile = TRUE,
    marginal_list = my_marginal3,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    copula = "vine",
    n_cores = 2,
    input_data = my_data$dat
  )
  
  my_copula1 <- fit_copula(
    sce = example_sce,
    assay_use = "counts",
    marginal_list = my_marginal3,
    family_use = c(rep("nb", 5), rep("zip", 5)),
    copula = "gaussian",
    n_cores = 1,
    input_data = my_data$dat,
    if_sparse = TRUE
  )

  my_para <- extract_para(
    sce = example_sce,
    marginal_list = my_marginal3,
    n_cores = 1,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    new_covariate = my_data2$new_covariate,
    data = my_data$dat
  )

  my_newcount <- simu_new(
    sce = example_sce,
    mean_mat = my_para$mean_mat,
    sigma_mat = my_para$sigma_mat,
    zero_mat = my_para$zero_mat,
    quantile_mat = NULL,
    copula_list = my_copula$copula_list,
    n_cores = 1,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    input_data = my_data$dat,
    new_covariate = my_data$new_covariate,
    important_feature = my_copula$important_feature,
    filtered_gene = my_data$filtered_gene
  )
  
  my_newcount2 <- simu_new(
    sce = example_sce,
    mean_mat = my_para$mean_mat,
    sigma_mat = my_para$sigma_mat,
    zero_mat = my_para$zero_mat,
    quantile_mat = my_quantile_mat$quantile_mat,
    copula_list = NULL,
    n_cores = 1,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    input_data = my_data$dat,
    new_covariate = my_data2$new_covariate,
    important_feature = my_copula$important_feature,
    filtered_gene = my_data$filtered_gene
  )

  my_newcount3 <- simu_new(
    sce = example_sce,
    mean_mat = my_para$mean_mat,
    sigma_mat = my_para$sigma_mat,
    zero_mat = my_para$zero_mat,
    quantile_mat = NULL,
    copula_list = my_copula1$copula_list,
    n_cores = 1,
    family_use = c(rep("nb", 9), rep("zip", 1)),
    input_data = my_data$dat,
    new_covariate = my_data2$new_covariate,
    important_feature = my_copula$important_feature,
    filtered_gene = my_data$filtered_gene
  )
  
  my_simu <- scdesign3(
    sce = example_sce,
    assay_use = "counts",
    celltype = "cell_type",
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
    sigma_formula = "1",
    family_use = c(rep("nb", 9), rep("zip", 1)),
    n_cores = 2,
    usebam = FALSE,
    corr_formula = "pseudotime",
    copula = "vine",
    DT = TRUE,
    pseudo_obs = FALSE,
    ncell = 1000,
    return_model = TRUE, simplify = TRUE, 
    n_rep = 2
  )

  # my_simu2 <- scdesign3(
  #   sce = example_sce,
  #   assay_use = "counts",
  #   celltype = "cell_type",
  #   pseudotime = "pseudotime",
  #   spatial = NULL,
  #   other_covariates = NULL,
  #   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  #   sigma_formula = "s(pseudotime, bs = 'cr', k = 3)",
  #   family_use = c(rep("nb", 5), rep("zip", 5)),
  #   n_cores = 2,
  #   usebam = FALSE,
  #   corr_formula = "pseudotime",
  #   empirical_quantile = TRUE,
  #   copula = "vine",
  #   DT = TRUE,
  #   pseudo_obs = FALSE,
  #   return_model = FALSE
  # )
})

#' Extract the parameters of each cell's distribution
#'
#' \code{extract_para} generates parameter matrices which determine each cell's distribution
#'
#' The function takes the new covariate (if use) from \code{\link{construct_data}} and
#' marginal models from \code{\link{fit_marginal}}.
#'
#' @param sce A \code{SingleCellExperiment} object.
#' @param assay_use A string which indicates the assay you will use in the sce. Default is 'counts'.
#' @param marginal_list A list of fitted regression models from \code{\link{fit_marginal}} for each gene in sce.
#' @param n_cores An integer. The number of cores to use.
#' @param family_use A string of the marginal distribution.
#' Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian', which represent 'poisson distribution',
#' 'negative binomial distribution', 'zero-inflated poisson distribution', 'zero-inflated negative binomial distribution',
#' and 'gaussian distribution' respectively.
#' @param new_covariate A data.frame which contains covariates of targeted simulated data from  \code{\link{construct_data}} and the
#' correlation group assignment for each cell in the column 'corr_group'.
#' @param parallelization A string indicating the specific parallelization function to use.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#' @param data A dataframe which is used when fitting the gamlss model
#' @return A list with the components:
#' \describe{
#'   \item{\code{mean_mat}}{A cell by feature matrix of the mean parameter.}
#'   \item{\code{sigma_mat}}{A cell by feature matrix of the sigma parameter (for Gaussian, the variance; for NB, the dispersion.).}
#'   \item{\code{zero_mat}}{A cell by feature matrix of the zero-inflation parameter (only non-zero for ZIP and ZINB).}
#' }
#' @examples
#'   data(example_sce)
#'   my_data <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "1"
#'   )
#'   my_marginal <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#'   )
#'   my_copula <- fit_copula(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   marginal_list = my_marginal,
#'   family_use = c(rep("nb", 5), rep("zip", 5)),
#'   copula = "vine",
#'   n_cores = 1,
#'   input_data = my_data$dat
#'   )
#'   my_para <- extract_para(
#'     sce = example_sce,
#'     marginal_list = my_marginal,
#'     n_cores = 1,
#'     family_use = c(rep("nb", 5), rep("zip", 5)),
#'     new_covariate = my_data$new_covariate,
#'     data = my_data$dat
#'   )
#'
#' @export extract_para

extract_para <-  function(sce,
                          assay_use = "counts",
                          marginal_list,
                          n_cores,
                          family_use,
                          new_covariate,
                          parallelization = "mcmapply",
                          BPPARAM = NULL,
                          data) {
  removed_cell_list <- lapply(marginal_list, function(x){x$removed_cell})
  marginal_list <- lapply(marginal_list, function(x){x$fit})

  # find gene whose marginal is fitted
  qc_gene_idx <- which(!is.na(marginal_list))
  if(length(family_use) != 1){
    family_use <- family_use[qc_gene_idx]
  }
  # check if user inputted new covariates
  data_temp <- data[,colnames(new_covariate), drop = FALSE]
  if(identical(data_temp, new_covariate)){
    new_covariate <- NULL
  }
  
  count_mat <- SummarizedExperiment::assay(sce, assay_use)
  
  mat_function <-function(x, y) {
    fit <- marginal_list[[x]]
    removed_cell <- removed_cell_list[[x]]
    #count_mat <-
    #  t(as.matrix(SummarizedExperiment::assay(sce, assay_use)))
    #data$gene <- count_mat[,x]
    data$gene <- as.numeric(count_mat[x, ])
    # if(!"gamlss" %in% class(fit)){
    #   modelframe <- model.frame(fit)
    # }else{
    #   modelframe <- fit$mu.x
    # }
    if(is.null(new_covariate)){
      total_cells <- dim(sce)[2]
      cell_names <- colnames(sce)
    }else{
      total_cells <- dim(new_covariate)[1]
      cell_names <- rownames(new_covariate)
    }

    if(length(removed_cell) > 0 && !any(is.na(removed_cell))){
      if(is.null(new_covariate)){
        data <- data[-removed_cell,]
      }else{
        if (methods::is(fit, "gamlss")){
          all_covariates <- all.vars(fit$mu.formula)[-1]
        }else{
          all_covariates <- all.vars(fit$formula)[-1]
        }
        remove_idx <- lapply(all_covariates, function(x){
          curr_x <- tapply(data$gene, data[,x], sum)
          zero_group <- which(curr_x==0)
          if(length(zero_group) == 0){
            return(NA)
          }else{
            type <- names(curr_x)[zero_group]
            return(which(new_covariate[,x] %in% type))
          }

        })
        remove_cell_idx <- unlist(remove_idx)
        remove_cell_idx <- unique(stats::na.omit(remove_cell_idx))
        if(length(remove_cell_idx) > 0){
          new_covariate <- new_covariate[-remove_cell_idx,]
        }
      }
    }

    if (methods::is(fit, "gamlss")) {
      mean_vec <-
        stats::predict(fit,
                       type = "response",
                       what = "mu",
                       newdata = new_covariate, data = data)
      if (y == "poisson" | y == "binomial") {
        theta_vec <- rep(NA, total_cells)
      } else if (y == "gaussian") {
        theta_vec = stats::predict(fit,
                                   type = "response",
                                   what = "sigma",
                                   newdata = new_covariate, data = data) # this thete_vec is used for sigma_vec
      } else if (y == "nb") {
        theta_vec <-
          stats::predict(fit,
                         type = "response",
                         what = "sigma",
                         newdata = new_covariate, data = data)
        #theta_vec[theta_vec < 1e-3] <- 1e-3
      } else if (y == "zip") {
        theta_vec <- rep(NA,  total_cells)
        zero_vec <-
          stats::predict(fit,
                         type = "response",
                         what = "sigma",
                         newdata = new_covariate, data = data)
      } else if (y == "zinb") {
        theta_vec <-
          stats::predict(fit,
                         type = "response",
                         what = "sigma",
                         newdata = new_covariate, data = data)
        zero_vec <-
          stats::predict(fit,
                         type = "response",
                         what = "nu",
                         newdata = new_covariate, data = data)
      } else {
        stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb, zip or zinb!")
      }
    } else {
      ## Fit is mgcv::gam
      if (is.null(new_covariate)) {
        y <- stats::family(fit)$family[1]
        if (grepl("Negative Binomial", y)) {
          y <- "nb"
        }

        mean_vec <- stats::predict(fit, type = "response")
        if (y == "poisson" | y == "binomial") {
          theta_vec <- rep(NA,  total_cells)
        } else if (y == "gaussian") {
          theta_vec <- rep(sqrt(fit$sig2), total_cells) # this thete_vec is used for sigma_vec
        } else if (y == "nb") {
          theta <- fit$family$getTheta(TRUE)
          theta_vec <- 1/rep(theta, total_cells)
        } else {
          stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb!")
        }
      } else{
        y <- stats::family(fit)$family[1]
        if (grepl("Negative Binomial", y)) {
          y <- "nb"
        }


        mean_vec <-
          stats::predict(fit, type = "response", newdata = new_covariate)
        if (y == "poisson" | y == "binomial") {
          theta_vec <- rep(NA, total_cells)
        } else if (y == "gaussian") {
          theta_vec = stats::predict(fit,
                                     type = "response",
                                     what = "sigma",
                                     newdata = new_covariate) # this thete_vec is used for sigma_vec
        } else if (y == "nb") {
          theta <- fit$family$getTheta(TRUE)
          theta_vec <- 1/rep(theta, total_cells)
        } else {
          stop("Distribution of gam must be one of gaussian, binomial, poisson, nb!")
        }
      }
    }

    if (!exists("zero_vec")) {
      zero_vec <- rep(0, length(mean_vec))
      names(zero_vec) <- names(mean_vec)
    }

    if(length(mean_vec) < total_cells){
      full_means <- rep(0, total_cells)
      names(full_means) <- cell_names
      full_means[names(mean_vec)] <- mean_vec
      full_theta <- rep(NA, total_cells)
      names(full_theta) <- cell_names
      full_zero <- rep(NA, total_cells)
      names(full_zero) <- cell_names
      if(is.null(names(theta_vec))){
        if(length(theta_vec) == length(mean_vec)){
          names(theta_vec) <- names(mean_vec)  ## for gamlss case
        }else{
          names(theta_vec) <- cell_names ## for gam case
        }

      }
      full_theta[names(theta_vec)] <- theta_vec
      full_zero[names(zero_vec)] <- zero_vec
      mean_vec <- full_means
      theta_vec <- full_theta
      zero_vec <- full_zero
    }

    #q_vec <- quantile_mat[, x]


    para_mat <- cbind(mean_vec, theta_vec, zero_vec)
    rownames(para_mat) <- cell_names
    para_mat
  }
  paraFunc <- parallel::mcmapply
  if(.Platform$OS.type == "windows"){
    BPPARAM <- BiocParallel::SnowParam()
    parallelization <- "bpmapply"
  }
  if(parallelization == "bpmapply"){
    paraFunc <- BiocParallel::bpmapply
  }
  if(parallelization == "pbmcmapply"){
    paraFunc <- pbmcapply::pbmcmapply
  }
  if(parallelization == "bpmapply"){
    if(class(BPPARAM)[1] != "SerialParam"){
    #if(!is(BPPARAM, "SerialParam")){
      BPPARAM$workers <- n_cores
    }
    mat <- suppressMessages(paraFunc(mat_function, x = seq_len(dim(sce)[1])[qc_gene_idx], y = family_use,BPPARAM = BPPARAM,SIMPLIFY = FALSE))
  }else{
    mat <- suppressMessages(paraFunc(mat_function, x = seq_len(dim(sce)[1])[qc_gene_idx], y = family_use,SIMPLIFY = FALSE
                                  ,mc.cores = n_cores
                                   ))
  }
  mean_mat <- sapply(mat, function(x)
    x[, 1])
  sigma_mat <- sapply(mat, function(x)
    x[, 2])
  zero_mat <- sapply(mat, function(x)
    x[, 3])

  if(length(qc_gene_idx) < dim(sce)[1]){
    colnames(mean_mat) <-
      colnames(sigma_mat) <- colnames(zero_mat) <- rownames(sce)[qc_gene_idx]
    zeros <- matrix(0, nrow = dim(mean_mat)[1], ncol = dim(sce)[1] - length(qc_gene_idx))
    na_mat <- matrix(NA, nrow = dim(mean_mat)[1], ncol = dim(sce)[1] - length(qc_gene_idx))
    rownames(zeros) <- rownames(mean_mat)
    colnames(zeros) <- rownames(sce)[-qc_gene_idx]
    rownames(na_mat) <- rownames(mean_mat)
    colnames(na_mat) <- rownames(sce)[-qc_gene_idx]
    mean_mat <- cbind(mean_mat,zeros)
    sigma_mat <- cbind(sigma_mat,na_mat)
    zero_mat <- cbind(zero_mat,na_mat)
    mean_mat <- mean_mat[,rownames(sce)]
    sigma_mat <- sigma_mat[,rownames(sce)]
    zero_mat <- zero_mat[,rownames(sce)]
  }else{
    colnames(mean_mat) <-
      colnames(sigma_mat) <- colnames(zero_mat) <- rownames(sce)
  }

  zero_mat <- Matrix::Matrix(zero_mat, sparse = TRUE)
  return(list(
    mean_mat = mean_mat,
    sigma_mat = sigma_mat,
    zero_mat = zero_mat
  ))
}

#' This function computes the thresholding sparse covariance/correlation estimator 
#' with the optimal threshold level.
#'
#' Part from Chenxin Jiang
#'
#' @param data The data matrix.
#' @param method The choice of method to select the optimal threshold level.
#' @param operator The choice of the thresholding operator.
#' @param corr The indicator of computing correlation or covariance matrix.
#'
#' @return The thresholding sparse covariance/correlation estimator.
#' @export sparse_cov
#' 
#' @examples
#' print("No example")

sparse_cov <- function(data, 
                       method=c('cv', 'qiu'),
                       operator=c('hard', 'soft', 'scad', 'al'),
                       corr=TRUE){
  p <- dim(data)[2]
  n <- dim(data)[1]
  
  # sample covariance
  z <- covariance(data)
  
  # select the optimal thresholding level
  delta <- est_delta(data, method=method, operator=operator)
  s <- thresh_op(z, operator=operator, delta=delta, n=n)
  
  # Modify s to make it psd
  tol <- 1e-6
  ev <- eigen(s, symmetric=TRUE, only.values = TRUE)$values
  s1 <- s + (tol-min(ev))*diag(dim(s)[1]) 
  
  if(corr){
    # make corr
    s1_corr <- Matrix::cov2cor(s1)
    output <- s1_corr
  }else{
    output <- s1
  }
  
  return(output)
}





## This function computes the thresholding sparse covariance estimator for a given threshold level.
thresh_op <- function(z, operator, delta, n){
  if(operator == 'hard'){
    s_method <- s_hard
  }else if(operator == 'soft'){
    s_method <- s_soft
  }else if(operator == 'scad'){
    s_method <- s_scad
  }else if(operator == 'al'){
    s_method <- s_al
  }else{
    stop('Please specify a valid thresholding operator.')
  }
  s_method(z, delta, n)
}

# Operator 1: Hard Thresholding
s_hard <- function(z, delta, n){
  p<-dim(z)[1]
  lambda <- sqrt(log(p)/n)*delta
  output <- (z>lambda)*z
  diag(output) <- diag(z)
  return(output)
}

# Operator 2: Soft Thresholding
s_soft <- function(z, delta, n){
  p <- dim(z)[1]
  lambda <- sqrt(log(p)/n)*delta
  z0 <- abs(z)-lambda
  output <- sign(z)*(z0>0)*z0
  diag(output) <- diag(z)
  return(output)
}


# Operator 3: SCAD (smoothly clipped absolute deviation)
s_scad <- function(z, delta, n, a=3.7){
  p <- dim(z)[1]
  lambda <- sqrt(log(p)/n)*delta
  
  output <- matrix(NA, dim(z)[1], dim(z)[2])
  
  index1 <- which(abs(z)<=2*lambda)
  z0 <- abs(z)-lambda
  output[index1] <- sign(z[index1])*(z0[index1]>0)*z0[index1]
  
  index2 <- which(abs(z)>2*lambda & abs(z)<=a*lambda)
  output[index2] <- ((a-1)*z[index2]-sign(z[index2])*a*lambda)/(a-2)
  
  index3 <- which(abs(z)>a*lambda)
  output[index3] <- z[index3]
  
  diag(output) <- diag(z)
  return(output)
}


# Operator 4: Adaptive lasso
s_al <- function(z, delta, n){
  p <- dim(z)[1]
  lambda <- sqrt(log(p)/n)*delta
  
  eta <- 3
  z0 <- abs(z) - lambda^(eta+1)*abs(z)^(-eta)
  output <- sign(z)*(z0>0)*z0
  diag(output) <- diag(z)
  return(output)
}






## This function select the optimal thresholding level delta
est_delta <- function(data, 
                      method=c('cv', 'qiu'),
                      operator=c('hard', 'soft', 'scad', 'al')){
  n <- dim(data)[1]
  if((method=='qiu') ){
    s <- covariance(data)
    delta <- qiu.select(data, s)
  }else if(method=='cv'){
    delta <- cv.min(data, operator)
  }else{
    stop('Please specify a valid thresholding method and an operator function.')
  }
  return(delta)
}


### Qiu function to tune delta
qiu.select = function(data, s=NULL){
  n <- dim(data)[1]
  p <- dim(data)[2]
  
  if(is.null(s)){
    s <- covariance(data)
  }
  
  # standardized covariance of Sigma
  eta <- sqrt(n/log(p))*s 
  # select lower triangular
  ltrig <- abs(eta[lower.tri(eta, diag = FALSE)])
  
  # parameter to select the optimal thr
  a <- min(sqrt(2+log(n)/log(p)), 2)
  a1 <- 2 - a
  a0 <- (log(log(p)))^(-1/2)
  M <- sum(((a1+a0)<ltrig) & (ltrig<=2))
  q <- (p-1)*p/2
  plog <- (log(p))^(1/2)
  V <- 2*q*(stats::pnorm(2*plog)-stats::pnorm((a1+a0)*plog))
  N2 <- max(M-V, plog)
  
  delta <- sqrt(2*(2-log(N2*1/plog)/log(p)))
  
  return(delta)
}

#### CV to tune delta
# The tuning parameter lambda was selected by minimizing the Frobenius norm of the 
# difference between s(sample_cov(train)) and sample_cov(test) by cross validation

# cv.select: For implementation, we choose the optimal delta from a delta.list 
# instead of really solving the optimization problem.
# lambda.list ranges in [0,lambda.max], with a user specified length.

# cv.min: solve the optimization problem directly (currently use this)

cv.select <- function(data, 
                      operator, 
                      fold=5, 
                      delta.length=10, 
                      delta.max=2){
  
  n <- dim(data)[1]
  p <- dim(data)[2]
  
  
  # Create delta list
  delta.list <- seq(0, delta.max, length.out=delta.length)
  
  # Create equal size folds
  folds <- cut(seq(1, nrow(data)), breaks=fold, labels=FALSE)
  
  fold_losses <- rep(0, fold) # Record fold loss 
  cv_errors <- rep(0, delta.length) # Record average estimated loss
  
  
  ## Perform k fold cross validation
  
  # iterate over the values of lambda
  for(j in 1:delta.length){
    
    # iterate over CV folds
    for(i in 1:fold){
      # Segment the data by fold 
      testIndexes <- which(folds==i, arr.ind=TRUE) # i-th fold as the testing data
      #trainIndexes <- which(folds!=i, arr.ind=TRUE) # i-th fold as the testing data
      testData <- data[testIndexes, ]
      trainData <- data[-testIndexes, ]
      # Get the covariance matrix estimator s based on the training data
      sample.cov.train <- covariance(trainData)
      s <- thresh_op(sample.cov.train, operator=operator, delta = delta.list[j], n=n)
      # Compute Frobenius risk = norm(distance of s and covariance of testing data)
      sample.cov.test <- covariance(testData)
      fold_losses[i] <- norm(s - sample.cov.test, type='F')
    }
    # Get the average estimated loss of CV
    cv_errors[j] <-  mean(fold_losses)
  }
  
  # Find the lambda that minimize cv_errors
  delta <- delta.list[which.min(cv_errors)]
  
  delta
}



cv.min <- function(data, 
                   operator, 
                   fold=5,
                   delta.max=2){
  n <- dim(data)[1]
  p <- dim(data)[2]
  
  
  ## Perform k fold cross validation
  cv.loss <- function(delta){
    # Create equal size folds
    folds <- cut(seq(1, nrow(data)), breaks=fold, labels=FALSE)
    fold_losses <- rep(0, fold) # Record fold loss 
    
    # iterate over CV folds
    for(i in seq_len(fold)){
      # Segment the data by fold 
      testIndexes <- which(folds==i, arr.ind=TRUE) # i-th fold as the testing data
      testData <- data[testIndexes, ]
      trainData <- data[-testIndexes, ]
      # Get the covariance matrix estimator s based on the training data
      sample.cov.train <- covariance(trainData)
      s <- thresh_op(sample.cov.train, operator=operator, delta = delta, n=n)
      # Compute Frobenius risk = norm(distance of s and covariance of testing data)
      sample.cov.test <- covariance(testData)
      fold_losses[i] <- norm(s - sample.cov.test, type='F')
    }
    # Get the average estimated loss of CV
    cv_errors <- mean(fold_losses)
    cv_errors
  }
  
  # Solve the optimization problem
  res <- stats::optimize(cv.loss, c(0, delta.max))
  delta <- res$minimum
  
  delta
}

#' Functions from gamlss/gamlss.add with bugs fixed
#'
#' An additive function to be used while fitting GAMLSS models. The interface for \code{gam()} in the \pkg{mgcv} package.
#' @section ga
#' @param formula A formula of the model.
#' @param envir The environment.
#' @param control The control of the model fitting.
#' @param ... Other arguments.
#' 
#' @return A xvar list.
#' 
#' @examples
#' print("No example")
#' 
#' @export
ga <- function(formula, envir, control = ga.control(...), ...)
{
  #------------------------------------------
  # function starts here
  #------------------------------------------
  scall <- deparse(sys.call(), width.cutoff = 500L)
  if (!methods::is(formula, "formula"))
    stop("formula argument in ga() needs a formula starting with ~")
  # get where "gamlss" is in system call, it can be in gamlss() or predict.gamlss()
  rexpr <- grepl("gamlss",sys.calls())
  #rexpr <- grepl("stats::model.frame.default", sys.calls()) ##



  for (i in length(rexpr):1) {
    position <- i # get the position
    if (rexpr[i] == TRUE)
      break
  }

  gamlss.env <- sys.frame(position) #gamlss or predict.gamlss
  #gamlss.env <- environment(formula)

  ## get the data
  ## this has been modified on the 12-12-14 to make sure that
  ##  if model.frame.gamlss() is used as for example in term.plot() the
  ## function does not fail (It need to implemented to all smoother using formula?)
  if (sys.call(position)[1] == "predict.gamlss()") {
    # if stats::predict is used
    Data <- get("data", envir = gamlss.env)
  } else if (sys.call(position)[1] == "gamlss()") {
    # if gamlss() is used
    if (is.null(get("gamlsscall", envir = gamlss.env)$data)) {
      # if no data argument but the formula can be interpreted
      Data <- stats::model.frame(formula)
    } else {
      # data argument in gamlss
      Data <- get("gamlsscall", envir = gamlss.env)$data
    }
  } else {
    Data <- get("data", envir = gamlss.env)
  }
  Data <- data.frame(base::eval(substitute(Data)))
  #-------------------------------------------------
  # new Daniil and Vlasis
  # Initialize gam
  formula <-
    stats::as.formula(paste0("Y.var", deparse(formula, width.cutoff = 500L)))

  Data$Y.var <- rep(0, nrow(Data))
  G <- mgcv::gam(
    formula,
    data = Data,
    offset = control$offset,
    method = control$method,
    optimizer = control$optimizer,
    control = control$control,
    scale =  control$scale,
    select = control$select,
    knots = control$knots,
    sp = control$sp,
    min.sp = control$min.sp,
    H = control$H,
    gamma = control$gamma,
    paraPen = control$paraPen,
    in.out = control$in.out,
    drop.unused.levels = control$drop.unused.levels,
    drop.intercept = control$drop.intercept,
    discrete = control$discrete,
    G = NULL,
    fit = FALSE
  )
  #--------------------------------------------------
  xvar <- rep(0,  dim(Data)[1])
  attr(xvar, "formula")     <- formula
  attr(xvar, "control")     <- control
  attr(xvar, "gamlss.env") <- gamlss.env
  attr(xvar, "data")       <- as.data.frame(Data)
  attr(xvar, "call")       <-
    substitute(gamlss.ga(data[[scall]], z, w, ...))
  attr(xvar, "class")      <- "smooth"
  attr(xvar, "G")          <- G
  xvar
}

#' Functions from gamlss/gamlss.add with bugs fixed
#'
#' An additive function to be used while fitting GAMLSS models. The interface for \code{bam()} in the \pkg{mgcv} package.
#' @section ba
#' @param formula A formula of the model.
#' @param control The control of the model fitting.
#' @param ... Other arguments.
#' @return A xvar list.
#' @examples
#' print("No example")
#' @export
ba <-function(formula, control = ba.control(...), ...)
{
  #------------------------------------------
  # function starts here
  #------------------------------------------
  scall <- Reduce(paste, deparse(sys.call(), width.cutoff = 500L))
  if (!methods::is(formula, "formula"))
    stop("formula argument in ba() needs a formula starting with ~")
  # get where "gamlss" is in system call, it can be in gamlss() or predict.gamlss()
  rexpr <- grepl("gamlss",sys.calls()) ##
  #rexpr <- grepl("fitModel", sys.calls())
  #rexpr <- grepl("stats::model.frame.default", sys.calls())

  for (i in length(rexpr):1) {
    position <- i # get the position
    if (rexpr[i]==TRUE) break
  }
  gamlss.env <- sys.frame(position) #gamlss or predict.gamlss
  ## get the data
  ## this has been modified on the 12-12-14 to make sure that
  ##  if model.frame.gamlss() is used as for example in term.plot() the
  ## function does not fail (It need to implemented to all smoother using formulea?)
  if (sys.call(position)[1]=="predict.gamlss()") { # if stats::predict is used
    Data <- get("data", envir=gamlss.env)
  } else if (sys.call(position)[1]=="gamlss()") { # if gamlss() is used
    if (is.null(get("gamlsscall", envir=gamlss.env)$data)) { # if no data argument but the formula can be interpreted
      Data <- stats::model.frame(formula)
    } else {# data argument in gamlss
      Data <- get("gamlsscall", envir=gamlss.env)$data
    }
  } else {
    Data <- get("data", envir=gamlss.env)
  }
  Data <- data.frame(eval(substitute(Data)))
  #-------------------------------------------------
  # new Daniil and Vlasis
  # Initialize bam
  formula <- stats::as.formula(paste0("Y.var", Reduce(paste,deparse(formula, width.cutoff = 500L)) ))
  Data$Y.var = rep(0, nrow(Data))
  #browser()
  G = mgcv::bam(formula,
                data = Data,
                offset = control$offset,
                method = control$method,
                control = control$control,
                select = control$select,
                scale = control$scale,
                gamma = control$gamma,
                knots = control$knots,
                sp = control$sp,
                min.sp = control$min.sp,
                paraPen = control$paraPen,
                chunk.size = control$chunk.size,
                rho = control$rho,
                AR.start = control$AR.start,
                discrete = control$discrete,
                cluster = control$cluster,
                nthreads = control$nthreads,
                gc.level = control$gc.level,
                use.chol = control$use.chol,
                samfrac = control$samfrac,
                coef = control$coef,
                drop.unused.levels = control$drop.unused.levels,
                drop.intercept = control$drop.intercept,
                G = NULL,
                fit = FALSE
  )
  ##bam(formula, family=gaussian(),
  ##      data=list()#, weights=NULL, subset=NULL,
  #    na.action=na.omit, offset=NULL#, method="fREML"#,control=list()#,
  #    select=FALSE#, scale=0#,gamma=1,knots=NULL,sp=NULL,min.sp=NULL,
  #    paraPen=NULL,chunk.size=10000,rho=0,AR.start=NULL,discrete=FALSE,
  #    cluster=NULL,nthreads=1,gc.level=1,use.chol=FALSE,samfrac=1,
  #    coef=NULL,drop.unused.levels=TRUE,G=NULL,fit=TRUE,drop.intercept=NULL,...)
  #
  #--------------------------------------------------
  xvar <- rep(0,  dim(Data)[1])
  attr(xvar,"formula")     <- formula
  attr(xvar,"control")     <- control
  attr(xvar, "gamlss.env") <- gamlss.env
  attr(xvar, "data")       <- as.data.frame(Data)
  attr(xvar, "call")       <- substitute(gamlss.ba(data[[scall]], z, w, ...))
  attr(xvar, "class")      <- "smooth"
  attr(xvar, "G")          <- G
  xvar
}

#' Functions from gamlss/gamlss.add with bugs fixed
#'
#' The control for \code{ba()}. From \code{gamlss.add::ba.control()} and \code{gamlss::bam()}.
#' @section ba.control
#'
#' @param offset The offset in the formula.
#' @param method The method argument in \code{bam()}.
#' @param control A list of fit control parameters to replace defaults returned by gam.control. Any control parameters not supplied stay at their default values.
#' @param select The \code{select} argument in \code{bam()}. Determine should selection penalties be added to the smooth effects, so that they can in principle be penalized out of the model.
#' @param scale For the scale parameter. If this is positive then it is taken as the known scale parameter. Negative signals that the scale parameter is unknown. 0 signals that the scale parameter is 1 for Poisson and binomial and unknown otherwise.
#' @param gamma The \code{gamma} argument in \code{bam()}. Increase above 1 to force smoother fits.
#' @param knots The \code{knots} argument in \code{bam()}. An optional list containing user specified knot values to be used for basis construction.
#' @param sp The \code{sp} argument in \code{bam()}. A vector of smoothing parameters can be provided here.
#' @param min.sp The \code{min.sp} argument in \code{bam()}. Lower bounds can be supplied for the smoothing parameters.
#' @param paraPen The \code{paraPen} argument in \code{bam()}. Optional list specifying any penalties to be applied to parametric model terms.
#' @param chunk.size The model matrix is created in chunks of this size, rather than ever being formed whole.
#' @param rho An AR1 error model can be used for the residuals (based on dataframe order), of Gaussian-identity link models. This is the AR1 correlation parameter.
#' @param AR.start Logical variable of same length as data, \code{TRUE} at first observation of an independent section of AR1 correlation.
#' @param discrete With \code{method="fREML"} it is possible to discretize covariates for storage and efficiency reasons. If \code{discrete} is \code{TRUE}, a number or a vector of numbers for each smoother term, then discretization happens. If numbers are supplied they give the number of discretization bins.
#' @param cluster \code{bam} can compute the computationally dominant QR decomposition in parallel using parLapply from the \code{parallel} package, if it is supplied with a cluster on which to do this (a cluster here can be some cores of a single machine).
#' @param nthreads Number of threads to use for non-cluster computation (e.g. combining results from cluster nodes).
#' @param gc.level To keep the memory footprint down, it can help to call the garbage collector often, but this takes a substatial amount of time. Setting this to zero means that garbage collection only happens when R decides it should. Setting to 2 gives frequent garbage collection. 1 is in between.
#' @param use.chol By default \code{bam} uses a very stable QR update approach to obtaining the QR decomposition of the model matrix. For well conditioned models an alternative accumulates the crossproduct of the model matrix and then finds its Choleski decomposition, at the end. This is somewhat more efficient, computationally.
#' @param samfrac For very large sample size Generalized additive models the number of iterations needed for the model fit can be reduced by first fitting a model to a random sample of the data, and using the results to supply starting values. This initial fit is run with sloppy convergence tolerances, so is typically very low cost. \code{samfrac} is the sampling fraction to use. 0.1 is often reasonable.
#' @param coef Initial values for model coefficients.
#' @param drop.unused.levels By default unused levels are dropped from factors before fitting. For some smooths involving factor variables you might want to turn this off.
#' @param drop.intercept Set to \code{TRUE} to force the model to really not have the a constant in the parametric model part, even with factor variables present.
#' @param ... Other arguments.
#'
#' @return A control object
#' @noRd
ba.control <- function(offset = NULL,
                       method = "fREML",
                       control = list(),
                       select = FALSE,
                       scale = 0,
                       gamma = 1,
                       knots = NULL,
                       sp = NULL,
                       min.sp = NULL,
                       paraPen = NULL,
                       chunk.size = 10000,
                       rho = 0,
                       AR.start = NULL,
                       discrete = TRUE,
                       cluster = NULL,
                       nthreads = 2,
                       gc.level = 1,
                       use.chol = FALSE,
                       samfrac = 1,
                       coef = NULL,
                       drop.unused.levels = TRUE,
                       drop.intercept = NULL,
                       ...)
{
  #gam()
  control <- mgcv::gam.control(...)
  #ga()
  list( offset=offset, method=method, control=control,  select=select,
        scale=scale, gamma=gamma, knots=knots, sp=sp, min.sp=min.sp,
        paraPen = paraPen, chunk.size = chunk.size, rho = rho,
        AR.start = AR.start,
        discrete=discrete,  cluster=cluster, nthreads=nthreads,
        gc.level=gc.level, use.chol=use.chol, samfrac=samfrac,
        coef= coef,
        drop.unused.levels = drop.unused.levels,
        drop.intercept=drop.intercept, ...)
}

#' Functions from gamlss/gamlss.add with bugs fixed
#'
#' The control for \code{ga()}. From \code{gamlss.add::ga.control()} and \code{gamlss::gam()}.
#' @section ga.control
#' @param offset The offset in the formula.
#' @param method The smoothing parameter estimation method.
#' @param optimizer An array specifying the numerical optimization method to use to optimize the smoothing parameter estimation criterion (given by \code{method})
#' @param control A list of fit control parameters to replace defaults returned by \code{gam.control}.
#' @param scale If this is positive then it is taken as the known scale parameter. Negative signals that the scale parameter is unknown. 0 signals that the scale parameter is 1 for Poisson and binomial and unknown otherwise.
#' @param select If this is \code{TRUE} then \code{gam()} can add an extra penalty to each term so that it can be penalized to zero.
#' @param knots This is an optional list containing user specified knot values to be used for basis construction.
#' @param sp A vector of smoothing parameters can be provided here.
#' @param min.sp Lower bounds can be supplied for the smoothing parameters.
#' @param H A user supplied fixed quadratic penalty on the parameters of the GAM can be supplied, with this as its coefficient matrix.
#' @param gamma Increase this beyond 1 to produce smoother models.
#' @param paraPen Optional list specifying any penalties to be applied to parametric model terms.
#' @param in.out Optional list for initializing outer iteration.
#' @param drop.unused.levels By default unused levels are dropped from factors before fitting. For some smooths involving factor variables you might want to turn this off.
#' @param drop.intercept Set to \code{TRUE} to force the model to really not have the a constant in the parametric model part, even with factor variables present. Can be vector when \code{formula} is a list.
#' @param discrete Experimental option for setting up models for use with discrete methods employed in \code{bam}.
#' @param ... Other arguments
#'
#' @return A control object
#' @noRd
ga.control <- function(offset = NULL,
                       method = "REML",
                       optimizer = c("outer","newton"),
                       control = list(),
                       scale = 0,
                       select = FALSE,
                       knots = NULL,
                       sp = NULL,
                       min.sp = NULL,
                       H = NULL,
                       gamma = 1,
                       paraPen = NULL,
                       in.out = NULL,
                       drop.unused.levels = TRUE,
                       drop.intercept = NULL,
                       discrete = FALSE,
                       ...)
{
  #gam()
  control <- mgcv::gam.control(...)
  #ga()
  list(offset=offset, method=method, optimizer=optimizer, control=control,
       scale= scale,
       select=select, knots=knots, sp=sp, min.sp=min.sp, H=H, gamma=gamma,
       paraPen=paraPen, in.out=in.out,   drop.unused.levels = drop.unused.levels,
       drop.intercept=drop.intercept, discrete = discrete, ...)
}





#' Functions from gamlss/gamlss.add with bugs fixed
#'
#' The gamlss versions of the generic function \code{model.frame}
#' @section model.frame.gamlss
#' @param formula A formula of the model.
#' @param what For which parameter to extract the model.frame, terms or model.frame.
#' @param parameter Equivalent to \code{what}.
#' @param ... Other arguments.
#'
#' @return a vector or matrix of predicted values.
#' @noRd
model.frame.gamlss <- function(formula, what = c("mu", "sigma", "nu", "tau"), parameter = NULL, ...)
{
  object <- formula
  dots <- list(...)
  what <- if (!is.null(parameter)) {
    match.arg(parameter, choices = c("mu", "sigma", "nu", "tau"))
  } else match.arg(what)
  Call <- object$call
  parform <- stats::formula(object, what)
  data <- if (!is.null(Call$data)) {
    # problem here, as Call$data is .
    #eval(Call$data)
    # instead, this would work:
    if(what == "mu") {
      eval(Call$data, environment(formula$mu.terms))
    }
    else if (what == "sigma") {
      eval(Call$data, environment(formula$sigma.terms))
    } else if (what == "nu") {
      eval(Call$data, environment(formula$nu.terms))
    } else if (what == "tau") {
      eval(Call$data, environment(formula$tau.terms))
    }
    # (there is no formula$terms, just mu.terms and sigma.terms)
  } else {
    environment(formula$terms)
  }
  Terms <- stats::terms(parform)
  mf <- stats::model.frame(
    Terms,
    data,
    xlev = object[[paste(what, "xlevels", sep = ".")]]
  )
  mf
}

#' Support for Function ga()
#'
#'This is support for the  smoother functions \code{ga()} interfaces for Simon Wood's \code{gam()} functions from package \pkg{mgcv}. It is not intended to be called directly by users. From \code{gamlss.add::gamlss.ga}.
#' @param x The explanatory variables
#' @param y Iterative y variable
#' @param w Iterative weights
#' @param xeval If xeval=TRUE then prediction is used
#' @param ... Other arguments
#' @return Not used
#' @examples
#' print("No example")
#' @export
gamlss.ga <-function(x, y, w, xeval = NULL, ...) {
  if (is.null(xeval))
  {#fitting
    #formula <- attr(x,"formula")
    #control <- as.list(attr(x, "control"))
    Y.var <- y
    W.var <- w
    G <- attr(x,"G")
    G$y <- Y.var
    G$w <- W.var
    G$mf$Y.var <- Y.var
    G$mf$`(weights)` <- W.var
    fit <-  mgcv::gam(G=G, fit=TRUE)
    df <- sum(ifelse(is.null(fit$edf2), yes = fit$edf, fit$edf2) + fit$edf1 - fit$edf)-1
    fv <- stats::fitted(fit)
    residuals <- y-fv
    list(fitted.values=fv, residuals=residuals,
         nl.df = df, lambda=fit$sp[1], #
         coefSmo = fit, var=NA)    # var=fv has to fixed
  } else { # predict
    gamlss.env <- as.environment(attr(x, "gamlss.env"))
    obj <- get("object", envir=gamlss.env ) # get the object from predict
    TT <- get("TT", envir=gamlss.env ) # get wich position is now
    SL <- get("smooth.labels", envir=gamlss.env) # all the labels of the smoother
    fit <- eval(parse(text=paste("obj$", get("what", envir=gamlss.env), ".coefSmo[[",as.character(match(TT,SL)), "]]", sep="")))
    OData <- attr(x,"data")
    ll <- dim(OData)[1]
    pred <- stats::predict(fit,newdata = OData[seq(length(y)+1,ll),])
  }
}


#' Support for Function ba()
#'
#'This is support for the  smoother functions \code{ba()} interfaces for Simon Wood's \code{bam()} functions from package \pkg{mgcv}. It is not intended to be called directly by users. From \code{gamlss.add::gamlss.ba}.
#' @param x The explanatory variables
#' @param y Iterative y variable
#' @param w Iterative weights
#' @param xeval If xeval=TRUE then prediction is used
#' @param ... Other arguments
#' @return Not used
#' @examples
#' print("No example")
#' @export
gamlss.ba <-function(x, y, w, xeval = NULL, ...) {
  if (is.null(xeval))
  {#fitting
    Y.var <- y
    W.var <- w
    G <- attr(x,"G")
    control = attr(x,"control")
    G$y <- Y.var
    G$w <- W.var
    G$mf$Y.var <- Y.var
    G$mf$`(weights)` <- W.var
    fit <-  mgcv::bam(G=G, fit=TRUE,
                      offset=control$offset, method=control$method,
                      control=control$control, select=control$select,
                      scale=control$scale, gamma=control$gamma,
                      knots=control$knots, sp=control$sp, min.sp=control$min.sp,
                      paraPen=control$paraPen, chunk.size=control$chunk.size,
                      rho=control$rho, AR.start=control$AR.start,
                      discrete=control$discrete,
                      cluster=control$cluster, nthreads=control$nthreads,
                      gc.level=control$gc.level, use.chol=control$use.chol,
                      samfrac=control$samfrac,
                      drop.unused.levels=control$bam$drop.unused.levels)
    df <- sum(ifelse(is.null(fit$edf2), yes = fit$edf, fit$edf2) + fit$edf1 - fit$edf)-1
    fv <- stats::fitted(fit)
    residuals <- y-fv
    list( fitted.values=fv, residuals=residuals,
          nl.df = df, lambda=fit$sp[1], #
          coefSmo = fit, var=NA)    # var=fv has to fixed
  } else { # predict
    gamlss.env <- as.environment(attr(x, "gamlss.env"))
    obj <- get("object", envir=gamlss.env ) # get the object from predict
    TT <- get("TT", envir=gamlss.env ) # get wich position is now
    SL <- get("smooth.labels", envir=gamlss.env) # all the labels of the smoother
    fit <- eval(parse(text=paste("obj$", get("what", envir=gamlss.env), ".coefSmo[[",as.character(match(TT,SL)), "]]", sep="")))
    OData <- attr(x,"data")
    ll <- dim(OData)[1]
    pred <- stats::predict(fit,newdata = OData[seq(length(y)+1,ll),])
  }
}

# Accessed from gamlss's github 04/10/2023
#' @export
predict.gamlss <- function(object,
                           what = c("mu", "sigma", "nu", "tau"),
                           parameter = NULL,
                           newdata = NULL,
                           type = c("link", "response", "terms"), # terms not working
                           terms = NULL,
                           se.fit = FALSE,
                           data = NULL, ...)
{
  ## this little function put data frames together
  ##  originated from an the R-help reply by B. Ripley
  ##------------------------------------------------------------------------------
  ##------------------------------------------------------------------------------
  ##-------- concat starts here
  concat <- function(..., names=NULL)
  {
    tmp <- list(...)
    if(is.null(names)) names <- names(tmp)
    if(is.null(names)) names <- sapply( as.list(match.call()), deparse)[-1]
    if( any(
      sapply(tmp, is.matrix)
      |
      sapply(tmp, is.data.frame) ) )
    {
      len <- sapply(tmp, function(x) c(dim(x),1)[1] )
      len[is.null(len)] <- 1
      data <- rbind( ... )
    }
    else
    {
      len <- sapply(tmp,length)
      data <- unlist(tmp)
    }
    namelist <- factor(rep(names, len), levels=names)
    return( data.frame( data, source=namelist) )
  }
  ##----------concat finish here
  ##--------------------------------------------------------------------------------------------
  ##--------------------------------------------------------------------------------------------
  ##   main function starts here
  ##----------------------------
  ## If no new data just use lpred() and finish
  if (is.null(newdata))  #
  {
    newdata = data
    #predictor<- lpred(object, what = what, type = type, terms = terms, se.fit = se.fit, ... )
    #return(predictor)
    #newdata <- object$mu.x
  }
  ## at the moment se.fit is not supported for new data
  if (se.fit)
    warning(" se.fit = TRUE is not supported for new data values at the moment \n")
  ##  stop if newdata is not data frame
  ## note that atomic is not working here so better to take it out Mikis 23-10-13
  ## if (!(is.atomic(newdata) | inherits(newdata, "data.frame")))
  #if (!(inherits(newdata, "data.frame")))
   # stop("newdata must be a data frame ") # or a frame mumber
  ## getting which parameter and type
  what <- if (!is.null(parameter))  {
    match.arg(parameter, choices=c("mu", "sigma", "nu", "tau"))} else  match.arg(what)
  type <- match.arg(type)
  ## get the original call
  Call <- object$call
  ## we need both the old and the new data sets
  ## the argument data can be provided by predict
  data<-if (is.null(data))
  {        ## if it is not provided then get it from the original call
    if (!is.null(Call$data)) eval(Call$data)
    else stop("define the original data using the option data")
  }else{
    data
  }
  # else data # if it provide get it
  ## keep only the same variables
  ## this assumes that all the relevant variables will be in newdata
  ## what happens if not?
  data <- data[,match(names(newdata),names(data)), drop=FALSE]
  ## merge the two data together
  data <- concat(data,newdata)
  ## get the formula
  parform <- stats::formula(object, what)# object[[paste(what, "formula", sep=".")]]
  ## put response to NULL
  if (length(parform)==3)
    parform[2] <- NULL
  ## define the terms
  Terms <- terms(parform)
  ## get the offset
  offsetVar <- if (!is.null(off.num <- attr(Terms, "offset"))) # new
    eval(attr(Terms, "variables")[[off.num + 1]], data)
  ## model frame
  if(is.null(environment(Terms))){
    environment(Terms) <- globalenv()
  }
  m <- stats::model.frame(Terms, data, xlev = object[[paste(what,"xlevels",sep=".")]])
  ## model design matrix y and w
  X <- stats::model.matrix(Terms, data, contrasts = object$contrasts)
  y <- object[[paste(what,"lp",sep=".")]]
  w <- object[[paste(what,"wt",sep=".")]]
  ## leave for future checks
  #  aN <- dim(newdata)[1]
  #zeros <- rep(0,aN)
  #ones <- rep(1,aN)
  #yaug <- as.vector(c(y,zeros))
  #waug <- as.vector(c(w,zeros))
  ## for keeping only the original data
  onlydata <- data$source == "data" # TRUE or FALSE
  ## whether additive terms are involved in the fitting
  smo.mat <- object[[paste(what,"s",sep=".")]]
  ## if offset take it out from fitting
  if (!is.null(off.num))
    y <- (y - offsetVar[onlydata])
  ## if smoothing
  if (!is.null(smo.mat))
  {
    n.smooths <- dim(smo.mat)[2]
    y <- (y - smo.mat %*% rep(1, n.smooths))
  }
  ## refit the model
  # refit <- lm.wfit(X[onlydata,  , drop = FALSE], y, w)
  # ## ckeck the residuals if they are zero
  # ##if (any(abs(resid(refit))>1e-005))
  # if (abs(sum(resid(refit)))>1e-001||abs(sum(coef(object, what=what)-coef(refit), na.rm=TRUE))>1e-005)
  #   warning(paste("There is a discrepancy  between the original and the re-fit",
  #                 " \n used to achieve 'safe' predictions \n ", sep = "" ))
  ## this is disturbing fit and refit have different coefficients  why?
  ## fit <- lm.wfit(X, yaug, waug)
  ## get the coefficients
  coef <- coef(object, what=what)        ## save the coefficints
  nX <- dimnames(X)        ## the names of rows and columns
  rownames <- nX[[1]][!onlydata] ## only the newdata rows
  nrows <- sum(!onlydata)     ## the number of rows in the new data
  nac <- is.na(coef)        ## whether they are NA in coefficients
  assign.coef <- attr(X, "assign")  ## X is a matrix
  collapse <- type != "terms"## !collapse is for whether type is not "terms"
  Xpred <- X[!onlydata,]
  Xpred <- matrix(Xpred, nrow=nrows) # I think this probably is not needed sinse allready a matrix
  # I will check this later
  if (!collapse)       ## whether type=="terms"
  {
    aa <- attr(X, "assign")
    ll <- attr(Terms, "term.labels")
    if (attr(Terms, "intercept") > 0)  ll <- c("(Intercept)", ll)
    aaa <- factor(aa, labels = ll)
    asgn <- split(order(aa), aaa)
    hasintercept <- attr(Terms, "intercept") > 0
    fit_qr <- object[[paste0(what,".qr")]]
    p <- fit_qr$rank
    p1 <- seq(len = p)
    piv <- fit_qr$pivot[p1]
    if (hasintercept)
    {
      asgn$"(Intercept)" <- NULL
      avx <- colMeans(X[onlydata, ])
      termsconst <- sum(avx[piv] * coef[piv])
    }
    # TT <- sum(onlydata)
    # xbar <- drop(array(1/TT, c(1, TT)) %*% X[onlydata, !nac])
    nterms <- length(asgn)
    #    if (nterms > 0)
    # define the prediction matrix
    pred <- matrix(ncol = nterms, nrow = nrows)
    dimnames(pred) <- list(rownames(newdata), names(asgn))
    #          if (se.fit )
    #          {
    #              ip <- matrix(ncol = nterms, nrow = NROW(X))
    #              dimnames(ip) <- list(rownames(X), names(asgn))
    #              Rinv <- qr.solve(qr.R(obj[[paste(what,"qr",sep=".")]])[p1, p1])
    #          }
    if (hasintercept)
      Xpred <- sweep(Xpred, 2, avx)
    unpiv <- rep.int(0, NCOL(Xpred))
    unpiv[piv] <- p1
    for (i in seq(1, nterms, length = nterms))
    {
      iipiv <- asgn[[i]]
      ii <- unpiv[iipiv]
      iipiv[ii == 0] <- 0
      pred[, i] <- if (any(iipiv > 0)) # ms Thursday, May 1, 2008 at 10:12
        Xpred[, iipiv, drop = FALSE] %*% coef[iipiv]
      else 0
      #              if (se.fit )
      #                ip[, i] <- if (any(iipiv > 0))
      #                  as.matrix(X[, iipiv, drop = FALSE] %*% Rinv[ii,
      #                    , drop = FALSE])^2 %*% rep.int(1,p)
      #                else 0
    }
    attr(pred, "constant") <- if (hasintercept)  termsconst
    else 0
    #   Xpred <- Xpred - outer(rep(1, sum(!onlydata)), avx)
    if (!is.null(terms))
    {
      pred <- pred[, terms, drop = FALSE]
      #   if (se.fit)
      #     ip <- ip[, terms, drop = FALSE]
    }
  } ## end if for terms
  else  ## if type is not terms but "link" or "response"
  {
    pred <- drop(Xpred[, !nac, drop = FALSE] %*% coef[!nac])
    if (!is.null(off.num) && collapse)
      pred <- pred + offsetVar[!onlydata]
  }
  ##
  ## now the smoothing part
  ##
  if (!is.null(smo.mat))
  {
    #cat("new prediction", "\n")
    smooth.labels <- dimnames(smo.mat)[[2]]       ## getting the labels i.e. "pb(Fl)" "pb(A)"
    pred.s <- array(0, c(nrows, n.smooths), list(names(pred),
                                                 dimnames(smo.mat)[[2]])) ## creating the prediction matrix
    # smooth.labels[smooth.labels%in%colnames(X)]
    # smooth.wanted <- smooth.labels[match(smooth.labels, colnames(X), 0) > 0]
    ## getting the smoothing call
    smooth.calls <- lapply(m[smooth.labels], attr, "call") # i.e $`pb(Fl)`
    #     gamlss.pb(data[["pb(Fl)"]], z, w)
    data <- subset(m, onlydata, drop=FALSE)        ## get the  original data
    attr(data, "class") <- NULL                                   ## note that m is the data.frame with all data
    new.m <- subset(m, !onlydata, drop=FALSE)       ## get the new data
    attr(new.m, "class") <- NULL
    residuals <-  if (!is.null(off.num)) object[[paste(what,"wv",sep=".")]] - object[[paste(what,"lp",sep=".")]]+offsetVar[onlydata]
    else object[[paste(what,"wv",sep=".")]] - object[[paste(what,"lp",sep=".")]]
    for(TT in smooth.labels)
    {
      if (is.matrix(m[[TT]])) # the problem is that for some smoother the m[[TT]] is a matrix (for example pvc())
      { # MS 27-6-11         # in this case  we have to protect the dim attributes of data[[tt]]
        nm <- names(attributes(m[[TT]])) # first we get the names of all attributes
        attributes(data[[TT]]) <- attributes(m[[TT]])[nm[-c(1,2)]]# then we pass all but
      }                                 # 1 and 2 i.e. dim and names
      else   attributes(data[[TT]]) <- attributes(m[[TT]])
      Call <- smooth.calls[[TT]] #
      Call$xeval <- substitute(new.m[[TT]], list(TT = TT))
      z <- residuals + smo.mat[, TT]
      # debug(gamlss.pvc)
      pred.s[, TT] <- eval(Call)
    }
    if(type == "terms")
    {
      # pred[, smooth.wanted] <- pred[, smooth.wanted] + pred.s[, smooth.wanted]
      pred[, smooth.labels] <- pred[, smooth.labels] + pred.s[, smooth.labels]
    }
    else pred <- drop(pred + pred.s %*% rep(1, n.smooths))
  }
  if(type == "response")
  {
    FAM <- eval(object$call$family)#
    if (!methods::is(FAM,"gamlss.family"))
    {
      FAM <- stats::family(object)[1]
    }
    # else
    # {
    FAM <- gamlss.dist::as.gamlss.family(FAM)# this should get a gamlss family but not alway
    pred <- FAM[[paste0(what,".linkinv")]](pred)
  }
  names(pred) <- rownames(newdata)
  pred
}

#' The wrapper for the whole scDesign3 pipeline
#'
#' \code{scdesign3} takes the input data, fits the model and
#'
#' @param sce A \code{SingleCellExperiment} object.
#' @param assay_use A string which indicates the assay you will use in the sce. Default is 'counts'.
#' @param celltype A string of the name of cell type variable in the \code{colData} of the sce. Default is 'cell_type'.
#' @param pseudotime A string or a string vector of the name of pseudotime and (if exist)
#' multiple lineages. Default is NULL.
#' @param spatial A length two string vector of the names of spatial coordinates. Default is NULL.
#' @param other_covariates A string or a string vector of the other covariates you want to include in the data.
#' @param ncell The number of cell you want to simulate. Default is \code{dim(sce)[2]} (the same number as the input data).
#' @param mu_formula A string of the mu parameter formula
#' @param sigma_formula A string of the sigma parameter formula
#' @param family_use A string of the marginal distribution.
#' Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian'.
#' @param n_cores An integer. The number of cores to use.
#' @param correlation_function A string. If 'default', the function from \code{Rfast}; if 'coop', the function from \code{coop}, which calls BLAS.
#' @param usebam A logic variable. If use \code{\link[mgcv]{bam}} for acceleration in marginal fitting.
#' @param edf_flexible A logic variable. It is used for accelerating for spatial model if k is large in 'mu_formula'. Default is FALSE.
#' @param corr_formula A string of the correlation structure.
#' @param empirical_quantile Please only use it if you clearly know what will happen! A logic variable. If TRUE, DO NOT fit the copula and use the EMPIRICAL CDF values of the original data; it will make the simulated data fixed (no randomness). Default is FALSE. Only works if ncell is the same as your original data.
#' @param copula A string of the copula choice. Must be one of 'gaussian' or 'vine'. Default is 'gaussian'. Note that vine copula may have better modeling of high-dimensions, but can be very slow when features are >1000.
#' @param if_sparse A logic variable. Only works for Gaussian copula (\code{family_set = "gaussian"}). If TRUE, a thresholding strategy will make the corr matrix sparse.
#' @param fastmvn An logical variable. If TRUE, the sampling of multivariate Gaussian is done by \code{mvnfast}, otherwise by \code{mvtnorm}. Default is FALSE. It only matters for Gaussian copula.
#' @param DT A logic variable. If TRUE, perform the distributional transformation
#' to make the discrete data 'continuous'. This is useful for discrete distributions (e.g., Poisson, NB).
#' Default is TRUE. Note that for continuous data (e.g., Gaussian), DT does not make sense and should be set as FALSE.
#' @param pseudo_obs A logic variable. If TRUE, use the empirical quantiles instead of theoretical quantiles for fitting copula.
#' Default is FALSE.
#' @param family_set A string or a string vector of the bivariate copula families. Default is c("gauss", "indep"). For more information please check package \code{rvinecoplib}.
#' @param important_feature A numeric value or vector which indicates whether a gene will be used in correlation estimation or not. If this is a numeric value, then
#' gene with zero proportion greater than this value will be excluded form gene-gene correlation estimation. If this is a vector, then this should
#' be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
#' gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
#' The default value is "all" (a special string which means no filtering).
#' @param nonnegative A logical variable. If TRUE, values < 0 in the synthetic data will be converted to 0. Default is TRUE (since the expression matrix is nonnegative).
#' @param nonzerovar A logical variable. If TRUE, for any gene with zero variance, a cell will be replaced with 1. This is designed for avoiding potential errors, for example, PCA. Default is FALSE.
#' @param return_model A logic variable. If TRUE, the marginal models and copula models will be returned. Default is FALSE.
#' @param simplify A logic variable. If TRUE, the fitted regression model will only keep the essential contains for \code{predict}, otherwise the fitted models can be VERY large. Default is FALSE.
#' @param parallelization A string indicating the specific parallelization function to use.
#' @param n_rep An integer number. The number of replicates of simulated new count matrix. Default is 1.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#' @param trace A logic variable. If TRUE, the warning/error log and runtime for gam/gamlss
#' will be returned, FALSE otherwise. Default is FALSE.
#' @return A list with the components:
#' \describe{
#'   \item{\code{new_count}}{A matrix of the new simulated count (expression) matrix.}
#'   \item{\code{new_covariate}}{A data.frame of the new covariate matrix.}
#'   \item{\code{model_aic}}{The model AIC.}
#'   \item{\code{marginal_list}}{A list of marginal regression models if return_model = TRUE.}
#'   \item{\code{corr_list}}{A list of correlation models (conditional copulas) if return_model = TRUE.}
#' }
#' @examples
#' data(example_sce)
#' my_simu <- scdesign3(
#' sce = example_sce,
#' assay_use = "counts",
#' celltype = "cell_type",
#' pseudotime = "pseudotime",
#' spatial = NULL,
#' other_covariates = NULL,
#' mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#' sigma_formula = "1",
#' family_use = "nb",
#' n_cores = 2,
#' usebam = FALSE,
#' edf_flexible = FALSE,
#' corr_formula = "pseudotime",
#' copula = "gaussian",
#' if_sparse = TRUE,
#' DT = TRUE,
#' pseudo_obs = FALSE,
#' ncell = 1000,
#' return_model = FALSE
#' )
#'
#' @export scdesign3
scdesign3 <- function(sce,
                      assay_use = "counts",
                      celltype,
                      pseudotime = NULL,
                      spatial = NULL,
                      other_covariates,
                      ncell = dim(sce)[2],
                      mu_formula,
                      sigma_formula = "1",
                      family_use = "nb",
                      n_cores = 2,
                      correlation_function = "default",
                      usebam = FALSE,
                      edf_flexible = FALSE,
                      corr_formula,
                      empirical_quantile = FALSE,
                      copula = "gaussian",
                      if_sparse = FALSE,
                      fastmvn = FALSE,
                      DT = TRUE,
                      pseudo_obs = FALSE,
                      family_set = c("gauss", "indep"),
                      important_feature = "all",
                      nonnegative = TRUE,
                      nonzerovar = FALSE,
                      return_model = FALSE,
                      simplify = FALSE,
                      parallelization = "mcmapply",
                      n_rep = 1,
                      BPPARAM = NULL,
                      trace = FALSE) {
  message("Input Data Construction Start")

  input_data <- construct_data(
    sce = sce,
    assay_use = assay_use,
    celltype = celltype,
    pseudotime = pseudotime,
    spatial = spatial,
    other_covariates = other_covariates,
    ncell = ncell,
    corr_by = corr_formula,
    parallelization = parallelization,
    BPPARAM = BPPARAM
  )
  message("Input Data Construction End")

  message("Start Marginal Fitting")
  marginal_res <- fit_marginal(
    mu_formula = mu_formula,
    sigma_formula = sigma_formula,
    n_cores = n_cores,
    data = input_data,
    family_use = family_use,
    usebam = usebam,
    edf_flexible = edf_flexible,
    parallelization = parallelization,
    BPPARAM = BPPARAM,
    trace = trace, 
    simplify = simplify
  )
  message("Marginal Fitting End")

  if(empirical_quantile == TRUE) {
    message("Extract Empirical Quantile Matrices")
    copula_res <- fit_copula(
      sce = sce,
      assay_use = assay_use,
      input_data = input_data$dat,
      marginal_list = marginal_res,
      family_use = family_use,
      empirical_quantile = TRUE,
      copula = copula,
      family_set = family_set,
      n_cores = n_cores,
      important_feature = important_feature,
      if_sparse = if_sparse,
      parallelization = parallelization,
      BPPARAM = BPPARAM
    )
  } else {
    message("Start Copula Fitting")
    copula_res <- fit_copula(
      sce = sce,
      assay_use = assay_use,
      input_data = input_data$dat,
      marginal_list = marginal_res,
      family_use = family_use,
      copula = copula,
      family_set = family_set,
      n_cores = n_cores,
      correlation_function = correlation_function,
      important_feature = important_feature,
      if_sparse = if_sparse,
      parallelization = parallelization,
      BPPARAM = BPPARAM
    )
    message("Copula Fitting End")
  }
  
  

  message("Start Parameter Extraction")
  para_list <- extract_para(
    sce = sce,
    assay_use = assay_use,
    marginal_list = marginal_res,
    n_cores = n_cores,
    family_use = family_use,
    new_covariate = input_data$newCovariate,
    parallelization = parallelization,
    BPPARAM = BPPARAM,
    data = input_data$dat
  )
  message("Parameter
Extraction End")

  message("Start Generate New Data")
  
  if(empirical_quantile == TRUE) {
    new_count <- simu_new(
      sce = sce,
      assay_use= assay_use,
      mean_mat = para_list$mean_mat,
      sigma_mat = para_list$sigma_mat,
      zero_mat = para_list$zero_mat,
      quantile_mat = copula_res$quantile_mat,
      copula_list = NULL,
      n_cores = n_cores,
      family_use = family_use,
      nonnegative = nonnegative,
      nonzerovar = nonzerovar,
      input_data = input_data$dat,
      new_covariate = input_data$newCovariate,
      important_feature = copula_res$important_feature,
      parallelization = parallelization,
      BPPARAM = BPPARAM,
      filtered_gene = input_data$filtered_gene
    )
  } else {
    if(n_rep == 1) {
      new_count <- simu_new(
        sce = sce,
        assay_use= assay_use,
        mean_mat = para_list$mean_mat,
        sigma_mat = para_list$sigma_mat,
        zero_mat = para_list$zero_mat,
        quantile_mat = NULL,
        copula_list = copula_res$copula_list,
        n_cores = n_cores,
        family_use = family_use,
        nonnegative = nonnegative,
        nonzerovar = nonzerovar,
        input_data = input_data$dat,
        new_covariate = input_data$newCovariate,
        important_feature = copula_res$important_feature,
        parallelization = parallelization,
        BPPARAM = BPPARAM,
        filtered_gene = input_data$filtered_gene
      )
    } else {
      new_count <- lapply(seq_len(n_rep), function(x) {
        current_count <- simu_new(
          sce = sce,
          assay_use= assay_use,
          mean_mat = para_list$mean_mat,
          sigma_mat = para_list$sigma_mat,
          zero_mat = para_list$zero_mat,
          quantile_mat = NULL,
          copula_list = copula_res$copula_list,
          n_cores = n_cores,
          family_use = family_use,
          nonnegative = nonnegative,
          nonzerovar = nonzerovar,
          input_data = input_data$dat,
          new_covariate = input_data$newCovariate,
          important_feature = copula_res$important_feature,
          parallelization = parallelization,
          BPPARAM = BPPARAM,
          filtered_gene = input_data$filtered_gene
        )
        current_count
      })
    }
  }
  
  message("New Data Generating End")

  scdesign3_res <- list(
    new_count = new_count,
    new_covariate = input_data$newCovariate,
    model_aic = copula_res$model_aic,
    model_bic = copula_res$model_bic,
    marginal_list = if (return_model)
      marginal_res
    else
      NULL,
    corr_list = if (return_model)
      copula_res$copula_list
    else
      NULL
  )
  return(scdesign3_res)
}

#' Fit the copula model
#'
#' \code{fit_copula} fits the copula model.
#'
#' This function takes the result from \code{\link{fit_marginal}} as the input and
#' and fit the copula model on the residuals.
#'
#' @param sce A \code{SingleCellExperiment} object.
#' @param assay_use A string which indicates the assay you will use in the sce.
#' Default is 'counts'.
#' @param input_data The input data, which is one of the output from \code{\link{construct_data}}.
#' @param empirical_quantile Please only use it if you clearly know what will happen! A logic variable. If TRUE, DO NOT fit the copula and use the EMPIRICAL CDF values of the original data; it will make the simulated data fixed (no randomness). Default is FALSE. Only works if ncell is the same as your original data.
#' @param marginal_list A list of fitted regression models from \code{\link{fit_marginal}}.
#' @param family_use A string or a vector of strings of the marginal distribution. Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian'.
#' @param copula A string of the copula choice. Must be one of 'gaussian' or 'vine'. Default is 'gaussian'. Note that vine copula may have better modeling of high-dimensions, but can be very slow when features are >1000.
#' @param DT A logic variable. If TRUE, perform the distributional transformation
#' to make the discrete data 'continuous'. This is useful for discrete distributions (e.g., Poisson, NB).
#' Default is TRUE. Note that for continuous data (e.g., Gaussian), DT does not make sense and should be set as FALSE.
#' @param pseudo_obs A logic variable. If TRUE, use the empirical quantiles instead of theoretical quantiles for fitting copula.
#' Default is FALSE.
#' @param epsilon A numeric variable for preventing the transformed quantiles to collapse to 0 or 1.
#' @param family_set A string or a string vector of the bivariate copula families. Default is c("gaussian", "indep").
#' @param important_feature A numeric value or vector which indicates whether a gene will be used in correlation estimation or not. If this is a numeric value, then
#' gene with zero proportion greater than this value will be excluded form gene-gene correlation estimation. If this is a vector, then this should
#' be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
#' gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
#' The default value for is "all" (a special string which means no filtering).
#' @param if_sparse A logic variable. Only works for Gaussian copula (\code{family_set = "gaussian"}). If TRUE, a thresholding strategy will make the corr matrix sparse. 
#' @param n_cores An integer. The number of cores to use.
#' @param correlation_function A string. If 'default', the function from \code{Rfast}; if 'coop', the function from \code{coop}, which calls BLAS.
#' @param parallelization A string indicating the specific parallelization function to use.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#'
#' @return A list with the components:
#' \describe{
#'   \item{\code{new_mvu}}{A matrix of the new multivariate uniform distribution from the copula.}
#'   \item{\code{copula_list}}{A list of the fitted copula model. If using Gaussian copula, a list of correlation matrices; if vine, a list of vine objects.}
#'   \item{\code{model_aic}}{A vector of the marginal AIC and the copula AIC.}
#'   \item{\code{model_bic}}{A vector of the marginal BIC and the copula BIC.}
#' }
#' @examples
#'   data(example_sce)
#'   my_data <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "1"
#'   )
#'   my_marginal <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#'   )
#'   my_copula <- fit_copula(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   marginal_list = my_marginal,
#'   family_use = c(rep("nb", 5), rep("zip", 5)),
#'   copula = "vine",
#'   n_cores = 1,
#'   input_data = my_data$dat
#'   )
#'   
#' @import mclust
#' @import gamlss
#'
#' @export fit_copula

fit_copula <- function(sce,
                       assay_use,
                       input_data,
                       empirical_quantile = FALSE,
                       marginal_list,
                       family_use,
                       copula = 'gaussian',
                       DT = TRUE,
                       pseudo_obs = FALSE,
                       epsilon = 1e-6,
                       family_set = c("gaussian", "indep"),
                       important_feature = "all",
                       if_sparse = FALSE,
                       correlation_function = "default",
                       n_cores,
                       parallelization = "mcmapply",
                       BPPARAM = NULL) {

  if(empirical_quantile == TRUE) {
      message("Use the empirical quantile matrices from the original data; do not fit copula. This will make the result FIXED.")
  }
  
  if(identical(important_feature, "all")) {
    important_feature <- rep(TRUE, dim(sce)[1])
  }
  
  marginals <- lapply(marginal_list, function(x){x$fit})
  # find gene whose marginal is fitted
  qc_gene_idx <- which(!is.na(marginals))
  if(length(family_use) != 1){
    family_use <- family_use[qc_gene_idx]
  }
  group_index <- unique(input_data$corr_group)
  ind <- group_index[1] == "ind"
  if(ind) {
    copula.aic <- 0
    copula.bic <- 0
    marginal.aic <- sum(sapply(marginals[qc_gene_idx], stats::AIC))
    marginal.bic <- sum(sapply(marginals[qc_gene_idx], stats::BIC))
    model_aic <- c(marginal.aic, copula.aic, marginal.aic + copula.aic)
    names(model_aic) <- c("aic.marginal", "aic.copula", "aic.total")
    model_bic <- c(marginal.bic, copula.bic, marginal.bic + copula.bic)
    names(model_bic) <- c("bic.marginal", "bic.copula", "bic.total")
    copula_list <- NULL
    return(
      list(
        #new_mvu = new_mvu,
        model_aic = model_aic,
        model_bic = model_bic,
        copula_list = copula_list,
        important_feature = important_feature
      )
    )
  }

  if(empirical_quantile == TRUE) {
    ###
    if(is.vector(important_feature) & methods::is(important_feature,"logical")){
      if(length(important_feature) != dim(sce)[1]){
        stop("The important_feature should either be 'auto' or a logical vector with the length equals to the number of genes in the input data")
      }
    }else{
      if(is.numeric(important_feature)){
        stopifnot(important_feature <= 1)
        gene_zero_prop <- apply(as.matrix(SummarizedExperiment::assay(sce, assay_use)), 1, function(y){
          sum(y < 1e-5) / dim(sce)[2]
        })
        important_feature <- gene_zero_prop < important_feature ## default zero proportion in scDesign2 is 0.8.
        names(important_feature) <- rownames(sce)
      }else{
        stop("The important_feature should either be a numeric value or a logical vector with the length equals to the number of genes in the input data")
      }
    }
    
    important_feature <- important_feature[qc_gene_idx]
    corr_group <- as.data.frame(input_data$corr_group)
    colnames(corr_group) <- "corr_group"

    
    newmvq.list <- lapply(group_index, function(x,
                                                sce,
                                                corr_group) {
      message(paste0("Empirical quantile group ", x, " starts"))
      sce <- sce[important_feature, ]
      curr_index <- which(corr_group[, 1] == x)
      
      newmat <- SummarizedExperiment::assay(sce[, curr_index], assay_use)
      newmat <- t(as.matrix(newmat))
      
      newmvq <- rvinecopulib::pseudo_obs(newmat)
      newmvq
    }, sce = sce,
    corr_group = corr_group)
      
    newmvq <- do.call("rbind", newmvq.list)
    newmvq <- newmvq[colnames(sce),] 
    
    return(list(model_aic = 0,
           model_bic = 0,
           quantile_mat = newmvq,
           important_feature = important_feature))
    ###
  } else {
    if (copula == "gaussian") {
      message("Convert Residuals to Multivariate Gaussian")
      newmat <- convert_n(
        sce = sce[qc_gene_idx,],
        assay_use = assay_use,
        marginal_list = marginal_list[qc_gene_idx],
        data = input_data,
        DT = DT,
        pseudo_obs = pseudo_obs,
        n_cores = n_cores,
        family_use = family_use,
        epsilon = epsilon,
        parallelization = parallelization,
        BPPARAM = BPPARAM
      )
      message("Converting End")
    } else{
      message("Convert Residuals to Uniform")
      newmat <- convert_u(
        sce = sce[qc_gene_idx,],
        assay_use = assay_use,
        marginal_list = marginal_list[qc_gene_idx],
        data = input_data,
        DT = DT,
        pseudo_obs = pseudo_obs,
        family_use = family_use,
        n_cores = n_cores,
        epsilon = epsilon,
        parallelization = parallelization,
        BPPARAM = BPPARAM
      )
      message("Converting End")
    }
    
    ## select important genes
    if(is.vector(important_feature) & methods::is(important_feature,"logical")){
      if(length(important_feature) != dim(sce)[1]){
        stop("The important_feature should either be 'auto' or a logical vector with the length equals to the number of genes in the input data")
      }
    }else{
      if(is.numeric(important_feature)){
        stopifnot(important_feature <= 1)
        gene_zero_prop <- apply(as.matrix(SummarizedExperiment::assay(sce, assay_use)), 1, function(y){
          sum(y < 1e-5) / dim(sce)[2]
        })
        important_feature <- gene_zero_prop < important_feature ## default zero proportion in scDesign2 is 0.8.
        names(important_feature) <- rownames(sce)
      }else{
        stop("The important_feature should either be a numeric value or a logical vector with the length equals to the number of genes in the input data")
      }
    }
    
    important_feature <- important_feature[qc_gene_idx]
    
    
    corr_group <- as.data.frame(input_data$corr_group)
    colnames(corr_group) <- "corr_group"
   
    
    newmvn.list <-
      lapply(group_index, function(x,
                                   sce,
                                   newmat,
                                   corr_group,
                                   ind,
                                   n_cores,
                                   important_feature) {
        message(paste0("Copula group ", x, " starts"))
        curr_index <- which(corr_group[, 1] == x)
       
        if (copula == "gaussian") {
          #message(paste0("Group ", group_index, " Start"))
          curr_mat <- newmat[curr_index, , drop = FALSE]
          #message("Cal MVN")
          cor.mat <- cal_cor(
            curr_mat,
            important_feature = important_feature,
            if_sparse = if_sparse,
            correlation_function = correlation_function,
            lambda = 0.05,
            tol = 1e-8,
            ind = ind
          )
          
          #message("Sample MVN")
          #new_mvu <- sampleMVN(n = curr_ncell,
          #                     Sigma = cor.mat)
          #message("MVN Sampling End")
          #rownames(new_mvu) <- curr_ncell_idx
          
          #message("Cal AIC/BIC Start")
          model_aic <- cal_aic(norm.mat = newmat,
                               cor.mat = cor.mat,
                               ind = ind)
          model_bic <- cal_bic(norm.mat = newmat,
                               cor.mat = cor.mat,
                               ind = ind)
          #message("Cal AIC/BIC End")
          
        } else if (copula == "vine") {
          message("Vine Copula Estimation Starts")
          start <- Sys.time()
          curr_mat <- newmat[curr_index, , drop = FALSE]
          curr_mat <- curr_mat[,which(important_feature)]
          vine.fit <- rvinecopulib::vinecop(
            data = curr_mat,
            family_set = family_set,
            show_trace = FALSE,
            par_method = "mle",
            cores = n_cores
          )
          end <- Sys.time()
          print(end - start)
          message("Vine Copula Estimation Ends")
          # if (curr_ncell != 0) {
          #   message("Sampling Vine Copula Starts")
          #   new_mvu <- rvinecopulib::rvinecop(
          #     curr_ncell,
          #     vine = vine.fit,
          #     cores = n_cores,
          #     qrng = TRUE
          #   )
          #   message("Sampling Vine Copula Ends")
          #   rownames(new_mvu) <- curr_ncell_idx
          # } else{
          #   new_mvu <- NULL
          # }
          model_aic <- stats::AIC(vine.fit)
          model_bic <- stats::BIC(vine.fit)
          cor.mat <- vine.fit
        } else{
          stop("Copula must be one of 'vine' or 'gaussian'")
        }
        return(
          list(
            #new_mvu = new_mvu,
            model_aic = model_aic,
            model_bic = model_bic,
            cor.mat = cor.mat
          )
        )
      }, sce = sce, 
      newmat = newmat, 
      ind = ind, 
      n_cores = n_cores, 
      corr_group = corr_group,
      important_feature = important_feature)
    
    #newmvn <-
    #  do.call(rbind, lapply(newmvn.list, function(x)
    #    x$new_mvu))
    
    copula.aic <- sum(sapply(newmvn.list, function(x)
      x$model_aic))
    marginal.aic <- sum(sapply(marginals[qc_gene_idx], stats::AIC))
    
    copula.bic <- sum(sapply(newmvn.list, function(x)
      x$model_bic))
    marginal.bic <- sum(sapply(marginals[qc_gene_idx], stats::BIC))
    
    model_aic <- c(marginal.aic, copula.aic, marginal.aic + copula.aic)
    names(model_aic) <- c("aic.marginal", "aic.copula", "aic.total")
    
    model_bic <- c(marginal.bic, copula.bic, marginal.bic + copula.bic)
    names(model_bic) <- c("bic.marginal", "bic.copula", "bic.total")
    
    copula_list <- lapply(newmvn.list, function(x)
      x$cor.mat)
    names(copula_list) <- group_index
    
    
    #new_mvu <- as.data.frame(newmvn)
    
    return(
      list(
        #new_mvu = new_mvu,
        model_aic = model_aic,
        model_bic = model_bic,
        copula_list = copula_list,
        important_feature = important_feature
      )
    )
  }
}


## Calculate the correlation matrix. If use sparse cor estimation, package spcov will be used (it can be VERY SLOW).
cal_cor <- function(norm.mat,
                    important_feature,
                    correlation_function = "default",
                    if_sparse = FALSE,
                    lambda = 0.05,
                    tol = 1e-8,
                    ind = FALSE) {
  if (ind) {
    cor.mat <- diag(rep(1, dim(norm.mat)[2]))
    return(cor.mat)
  }
  else {
    cor.mat <- diag(rep(1, dim(norm.mat)[2]))
    rownames(cor.mat) <- colnames(norm.mat)
    colnames(cor.mat) <- colnames(norm.mat)
    important.mat <- norm.mat[,which(important_feature)]
    if (if_sparse) {
      important_cor.mat <- sparse_cov(important.mat, 
                               method = 'qiu', 
                               operator = 'hard', 
                               corr = TRUE)
    } else {
      important_cor.mat <- correlation(important.mat, correlation_function = correlation_function)
    }

    #s_d <- apply(norm.mat, 2, stats::sd)
    s_d <- matrixStats::colSds(important.mat,na.rm = TRUE)
    if (any(0 == s_d)) {
      important_cor.mat[is.na(important_cor.mat)] <- 0
    }
    cor.mat[rownames(important_cor.mat), colnames(important_cor.mat)] <- important_cor.mat
  }

  n <- dim(cor.mat)[1]

  ### We currently gave up this because the sparse corr calculation is too slow.
  # if (if_sparse) {
  #   ifposd <- matrixcalc::is.positive.definite(cor.mat, tol = tol)
  #   if (!ifposd) {
  #     warning("Cor matrix is not positive defnite! Add tol to the diagnol.")
  #     #diag(cor.mat) <- diag(cor.mat) + tol
  #   }
  #   ## Call spcov
  #   lambda_matrix <- matrix(rep(1, n ^ 2), nrow = n) * lambda
  #   diag(lambda_matrix) <- 0
  #   scor <- spcov::spcov(
  #     cor.mat,
  #     cor.mat,
  #     lambda = lambda_matrix,
  #     step.size  = 100,
  #     trace = 1,
  #     n.inner.steps = 200,
  #     thr.inner = tol
  #   )
  #   cor.mat <- scor$Sigma
  # }
  if (if_sparse) {
  cor.mat <- methods::as(methods::as(methods::as(cor.mat, "dMatrix"), "symmetricMatrix"), "CsparseMatrix")
  }
  cor.mat
}

## Convert marginal distributions to standard normals.
convert_n <- function(sce,
                      assay_use,
                      marginal_list,
                      data,
                      DT = TRUE,
                      pseudo_obs = FALSE,
                      epsilon = 1e-6,
                      family_use,
                      n_cores,
                      parallelization,
                      BPPARAM) {
  ## Extract count matrix
  count_mat <-
      t(as.matrix(SummarizedExperiment::assay(sce, assay_use)))
  removed_cell_list <- lapply(marginal_list, function(x){x$removed_cell})
  marginal_list <- lapply(marginal_list, function(x){x$fit})
  # n cell
  ncell <- dim(count_mat)[1]


  mat_function <- function(x, y) {
    fit <- marginal_list[[x]]
    removed_cell <- removed_cell_list[[x]]
    if(length(removed_cell) > 0 && !any(is.na(removed_cell))){
      data<- data[-removed_cell,]
    }
    if (methods::is(fit, "gamlss")) {
      mean_vec <- stats::predict(fit, type = "response", what = "mu", data = data)
      if (y == "poisson" | y == "binomial") {
        theta_vec <- rep(NA, length(mean_vec))
      } else if (y == "gaussian") {
        theta_vec <-
          stats::predict(fit, type = "response", what = "sigma", data = data) # called the theta_vec but actually used as sigma_vec for Gaussian
      } else if (y == "nb") {
        theta_vec <-
          1 / stats::predict(fit, type = "response", what = "sigma", data = data)
        #theta_vec[theta_vec < 1e-3] <- 1e-3
      } else if (y == "zip") {
        theta_vec <- rep(NA, length(mean_vec))
        zero_vec <-
          stats::predict(fit, type = "response", what = "sigma", data = data)
      } else if (y == "zinb") {
        theta_vec <- stats::predict(fit, type = "response", what = "sigma", data = data)
        zero_vec <-
          stats::predict(fit, type = "response", what = "nu", data = data)
      } else {
        stop("Distribution of gamlss must be one of gaussian, poisson, nb, zip or zinb!")
      }
    } else {
      ## if input is from mgcv
      ## Check the family (since sometimes zip and zinb may degenerate into poisson or nb)

      y <- stats::family(fit)$family[1]
      if (grepl("Negative Binomial", y)) {
        y <- "nb"
      }

      mean_vec <- stats::predict(fit, type = "response")
      if (y == "poisson" | y == "binomial") {
        theta_vec <- rep(NA, length(mean_vec))
      } else if (y == "gaussian") {
        theta_vec <- rep(sqrt(fit$sig2), length(mean_vec)) # called the theta_vec but actually used as sigma_vec for Gaussian
      } else if (y == "nb") {
        theta <- fit$family$getTheta(TRUE)
        theta_vec <- rep(theta, length(mean_vec))
      } else {
        stop("Distribution of mgcv must be one of gaussian, poisson or nb!")
      }
    }

    ## Mean for Each Cell


    Y <- count_mat[names(mean_vec), x]


    ## Frame
    if (!exists("zero_vec")) {
      zero_vec <- 0
    }
    family_frame <- cbind(Y, mean_vec, theta_vec, zero_vec)
    if (y == "binomial") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::pbinom(x[1], prob = x[2], size = 1)
      })
    } else if (y == "poisson") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::ppois(x[1], lambda = x[2])
      })
    } else if (y == "gaussian") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pNO(x[1], mu = x[2], sigma = abs(x[3]))
      })
    } else if (y == "nb") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::pnbinom(x[1], mu = x[2], size = x[3])
      })
    } else if (y == "zip") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pZIP(x[1], mu = x[2], sigma = abs(x[4]))
      })
    }
    else if (y == "zinb") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pZINBI(x[1],
                            mu = x[2],
                            sigma = abs(x[3]),
                            nu = x[4])
      })
    } else {
      stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb, zip or zinb!")
    }

    ## CHECK ABOUT THE FIRST PARAM!!!!!
    if (DT) {
      if (y == "poisson") {
        pvec2 <- apply(family_frame, 1, function(x) {
          stats::ppois(x[1] - 1, lambda = x[2]) * as.integer(x[1] > 0)
        })
      } else if (y == "gaussian" | y == "binomial") {
        ## Gaussian is continuous, thus do not need DT.
        ## Binomial seems to be weird to have DT.
        message("Continuous gaussian does not need DT.")
        pvec2 <- pvec
        # pvec2 <- apply(family_frame, 1, function(x){
        # pNO(x[1]-1, mu = x[2], sigma = abs(x[3]))*as.integer(x[1] > 0)
        #})
      } else if (y == "nb") {
        pvec2 <- apply(family_frame, 1, function(x) {
          stats::pnbinom(x[1] - 1, mu = x[2], size = x[3]) * as.integer(x[1] > 0)
        })
      } else if (y == "zip") {
        pvec2 <- apply(family_frame, 1, function(x) {
          ifelse(x[1] > 0, gamlss.dist::pZIP(x[1] - 1, mu = x[2], sigma = abs(x[4])), 0)
        })
      } else if (y == "zinb") {
        pvec2 <- apply(family_frame, 1, function(x) {
          ifelse(x[1] > 0,
                 gamlss.dist::pZINBI(
                   x[1] - 1,
                   mu = x[2],
                   sigma = abs(x[3]),
                   nu = x[4]
                 ),
                 0)
        })
      } else {
        stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb, zip or zinb!")
      }

      u1 <- pvec
      u2 <- pvec2

      v <- stats::runif(length(mean_vec))
      ## Random mapping
      r <- u1 * v + u2 * (1 - v)
    } else {
      r <- pvec
    }

    if(length(r) < dim(sce)[2]){
      new_r <- rep(1, dim(sce)[2])
      names(new_r) <- colnames(sce)
      new_r[names(r)] <- r
      r <- new_r
    }

    ## Avoid Inf
    idx_adjust <- which(1 - r < epsilon)
    r[idx_adjust] <- r[idx_adjust] - epsilon
    idx_adjust <- which(r < epsilon)
    r[idx_adjust] <- r[idx_adjust] + epsilon

    if (pseudo_obs) {
      r <- rvinecopulib::pseudo_obs(r)
    }

    stats::qnorm(
      r,
      mean = 0,
      sd = 1,
      lower.tail = TRUE,
      log.p = FALSE
    )
  }

  paraFunc <- parallel::mcmapply
  if(.Platform$OS.type == "windows"){
    BPPARAM <- BiocParallel::SnowParam()
    parallelization <- "bpmapply"
  }
  if(parallelization == "bpmapply"){
    paraFunc <- BiocParallel::bpmapply
  }
  if(parallelization == "pbmcmapply"){
    paraFunc <- pbmcapply::pbmcmapply
  }
  if(parallelization == "bpmapply"){
    if(class(BPPARAM)[1] != "SerialParam"){
      BPPARAM$workers <- n_cores
    }
    mat <- paraFunc(mat_function, x = seq_len(dim(sce)[1]), y = family_use, SIMPLIFY = TRUE, BPPARAM = BPPARAM)
  }else{
    mat <- paraFunc(mat_function, x = seq_len(dim(sce)[1]), y = family_use, SIMPLIFY = TRUE, mc.cores = n_cores)
  }
  colnames(mat) <- rownames(sce)
  rownames(mat) <- colnames(sce)

  ## Remove inf
  mat[is.infinite(mat)] <- NA

  for (i in 1:ncol(mat)) {
    mat[is.na(mat[, i]), i] <- mean(mat[, i], na.rm = TRUE)
  }

  return(mat)
}

## convert marginals to Unif[0, 1].
convert_u <- function(sce,
                      assay_use,
                      marginal_list,
                      data,
                      DT = TRUE,
                      pseudo_obs = FALSE,
                      epsilon = 1e-6,
                      n_cores,
                      family_use,
                      parallelization,
                      BPPARAM) {
  ## Extract count matrix
  count_mat <-
      t(as.matrix(SummarizedExperiment::assay(sce, assay_use)))
  removed_cell_list <- lapply(marginal_list, function(x){x$removed_cell})
  marginal_list <- lapply(marginal_list, function(x){x$fit})


  # n cell
  ncell <- dim(count_mat)[1]

  mat_function <- function(x, y) {
    fit <- marginal_list[[x]]
    removed_cell <- removed_cell_list[[x]]
    if(length(removed_cell) > 0 && !any(is.na(removed_cell))){
      data<- data[-removed_cell,]
    }
    if (methods::is(fit, "gamlss")) {
      mean_vec <- stats::predict(fit, type = "response", what = "mu", data = data) #
      if (y == "poisson" | y == "binomial") {
        theta_vec <- rep(NA, length(mean_vec))
      } else if (y == "gaussian") {
        theta_vec <-
          stats::predict(fit, type = "response", what = "sigma", data = data) # called the theta_vec but actually used as sigma_vec for Gaussian
      } else if (y == "nb") {
        theta_vec <-
          1 / stats::predict(fit, type = "response", what = "sigma", data = data) #
        #theta_vec[theta_vec < 1e-3] <- 1e-3
      } else if (y == "zip") {
        theta_vec <- rep(NA, length(mean_vec))
        zero_vec <-
          stats::predict(fit, type = "response", what = "sigma", data = data)
      } else if (y == "zinb") {
        theta_vec <- stats::predict(fit, type = "response", what = "sigma", data = data)
        zero_vec <-
          stats::predict(fit, type = "response", what = "nu", data = data)
      } else {
        stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb, zip or zinb!")
      }
    } else {
      ## if input is from mgcv, update the family
      y <- stats::family(fit)$family[1]
      if (grepl("Negative Binomial", y)) {
        y <- "nb"
      }

      mean_vec <- stats::predict(fit, type = "response")
      if (y == "poisson" | y == "binomial") {
        theta_vec <- rep(NA, length(mean_vec))
      } else if (y == "gaussian") {
        theta_vec <- rep(sqrt(fit$sig2), length(mean_vec)) # called the theta_vec but actually used as sigma_vec for Gaussian
      } else if (y == "nb") {
        theta <- fit$family$getTheta(TRUE)
        theta_vec <- rep(theta, length(mean_vec))
      } else {
        stop("Distribution of mgcv must be one of gaussian, binomial, poisson or nb!")
      }
    }

    ## Mean for Each Cell


    Y <- count_mat[names(mean_vec), x]


    ## Frame
    if (!exists("zero_vec")) {
      zero_vec <- 0
    }
    family_frame <- cbind(Y, mean_vec, theta_vec, zero_vec)

    if (y == "binomial") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::pbinom(x[1], prob = x[2], size = 1)
      })
    } else if (y == "poisson") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::ppois(x[1], lambda = x[2])
      })
    } else if (y == "gaussian") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pNO(x[1], mu = x[2], sigma = abs(x[3]))
      })
    } else if (y == "nb") {
      pvec <- apply(family_frame, 1, function(x) {
        stats::pnbinom(x[1], mu = x[2], size = x[3])
      })
    } else if (y == "zip") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pZIP(x[1], mu = x[2], sigma = abs(x[4]))
      })
    }
    else if (y == "zinb") {
      pvec <- apply(family_frame, 1, function(x) {
        gamlss.dist::pZINBI(x[1],
                            mu = x[2],
                            sigma = abs(x[3]),
                            nu = x[4])
      })
    } else {
      stop("Distribution of gamlss must be one of gaussian, poisson, nb, zip or zinb!")
    }

    ## CHECK ABOUT THE FIRST PARAM!!!!!
    if (DT) {
      if (y == "poisson") {
        pvec2 <- apply(family_frame, 1, function(x) {
          stats::ppois(x[1] - 1, lambda = x[2]) * as.integer(x[1] > 0)
        })
      } else if (y == "gaussian" | y == "binomial") {
        ## Gaussian is continuous, thus do not need DT.
        message("Continuous gaussian doesnot need DT.")
        pvec2 <- pvec
        # pvec2 <- apply(family_frame, 1, function(x){
        # pNO(x[1]-1, mu = x[2], sigma = abs(x[3]))*as.integer(x[1] > 0)
        #})
      } else if (y == "nb") {
        pvec2 <- apply(family_frame, 1, function(x) {
          stats::pnbinom(x[1] - 1, mu = x[2], size = x[3]) * as.integer(x[1] > 0)
        })
      } else if (y == "zip") {
        pvec2 <- apply(family_frame, 1, function(x) {
          ifelse(x[1] > 0, gamlss.dist::pZIP(x[1] - 1, mu = x[2], sigma = abs(x[4])), 0)
        })
      } else if (y == "zinb") {
        pvec2 <- apply(family_frame, 1, function(x) {
          ifelse(x[1] > 0,
                 gamlss.dist::pZINBI(
                   x[1] - 1,
                   mu = x[2],
                   sigma = abs(x[3]),
                   nu = x[4]
                 ),
                 0)
        })
      } else {
        stop("Distribution of gamlss must be one of gaussian, binomial, poisson, nb, zip or zinb!")
      }

      u1 <- pvec
      u2 <- pvec2

      v <- stats::runif(length(mean_vec))
      ## Random mapping
      r <- u1 * v + u2 * (1 - v)
    } else {
      r <- pvec
    }

    if(length(r) < dim(sce)[2]){
      new_r <- rep(1, dim(sce)[2])
      names(new_r) <- colnames(sce)
      new_r[names(r)] <- r
      r <- new_r
    }

    ## Avoid Inf
    idx_adjust <- which(1 - r < epsilon)
    r[idx_adjust] <- r[idx_adjust] - epsilon
    idx_adjust <- which(r < epsilon)
    r[idx_adjust] <- r[idx_adjust] + epsilon

    if (pseudo_obs) {
      r <- rvinecopulib::pseudo_obs(r)
    }

    r
  }

  paraFunc <- parallel::mcmapply
  if(.Platform$OS.type == "windows"){
    BPPARAM <- BiocParallel::SnowParam()
    parallelization <- "bpmapply"
  }
  if(parallelization == "bpmapply"){
    paraFunc <- BiocParallel::bpmapply
  }
  if(parallelization == "pbmcmapply"){
    paraFunc <- pbmcapply::pbmcmapply
  }
  if(parallelization == "bpmapply"){
    if(class(BPPARAM)[1] != "SerialParam"){
      BPPARAM$workers <- n_cores
    }
    mat <- paraFunc(mat_function, x = seq_len(dim(sce)[1]), y = family_use, SIMPLIFY = TRUE, BPPARAM = BPPARAM)
  }else{
    mat <- paraFunc(mat_function, x = seq_len(dim(sce)[1]), y = family_use, SIMPLIFY = TRUE, mc.cores = n_cores)
  }
  colnames(mat) <- rownames(sce)
  rownames(mat) <- colnames(sce)

  ## Remove inf
  mat[is.infinite(mat)] <- NA

  ## Use mean to replace the missing value
  for (i in 1:ncol(mat)) {
    mat[is.na(mat[, i]), i] <- mean(mat[, i], na.rm = TRUE)
  }

  quantile_mat <- mat
  return(quantile_mat)
}


## Calculate Gaussian copula AIC
cal_aic <- function(norm.mat,
                    cor.mat,
                    ind) {
  if (ind) {
    copula.aic = 0
  } else {
    cor.mat <- as.matrix(cor.mat)
    copula.nop <- (as.integer(sum(cor.mat != 0)) - dim(cor.mat)[1]) / 2

    copula.aic <-
      -2 * (sum(
        mvtnorm::dmvnorm(
          x = norm.mat,
          mean = rep(0, dim(cor.mat)[1]),
          sigma = cor.mat,
          log = TRUE
        )
      ) - sum(rowSums(stats::dnorm(norm.mat, log = TRUE)))) + 2 * copula.nop
  }

  copula.aic
}

## Calculate Gaussian copula BIC
cal_bic <- function(norm.mat,
                    cor.mat,
                    ind) {
  n_obs <- dim(norm.mat)[1]
  if (ind) {
    copula.bic = 0
  } else {
    cor.mat <- as.matrix(cor.mat)
    copula.nop <- (as.integer(sum(cor.mat != 0)) - dim(cor.mat)[1]) / 2

    copula.bic <-
      -2 * (sum(
        mvtnorm::dmvnorm(
          x = norm.mat,
          mean = rep(0, dim(cor.mat)[1]),
          sigma = cor.mat,
          log = TRUE
        )
      ) - sum(rowSums(stats::dnorm(norm.mat, log = TRUE)))) + log(n_obs) * copula.nop
  }

  copula.bic
}

## Similar to the cora function from "Rfast" but uses different functions to calculate column means and row sums.
correlation <- function(x, correlation_function = "default") {
  if(correlation_function == "default") {
    mat <- t(x) - matrixStats::colMeans2(x)
    mat <- mat / sqrt(matrixStats::rowSums2(mat^2))
    tcrossprod(mat)
  } else if (correlation_function == "coop") {
    coop::pcor(x, inplace = TRUE)
  } else {
    stop("correlation_function is either default or coop.")
  }
  
  
}

## Similar to the cova function from "Rfast" but uses different functions to calculate column means and row sums.
covariance <- function(x, correlation_function = "default") {
  if(correlation_function == "default") {
  n <- dim(x)[1]
  m <- sqrt(n) * matrixStats::colMeans2(x)
  s <- (crossprod(x) - tcrossprod(m))/(n - 1)
  s} else if (correlation_function == "coop") {
    coop::covar(x, inplace = TRUE)
  } else {
  stop("correlation_function is either default or coop.")
  }
}

### Sample MVN based on cor
sampleMVN <- function(n,
                      Sigma, 
                      n_cores = n_cores,
                      fastmvn = fastmvn) {
  if_sparse <- methods::is(Sigma, "dsCMatrix")
  p <- dim(Sigma)[1]
  if(if_sparse) {
    CH <- Matrix::Cholesky(methods::as(methods::as(methods::as(Sigma, "dMatrix"), "symmetricMatrix"), "CsparseMatrix"))
    mvnrv <- sparseMVN::rmvn.sparse(n = n, rep(0, p), CH, prec = FALSE)
  } else {
    if(fastmvn) {
      mvnrv <- mvnfast::rmvn(n, mu = rep(0, p), sigma = Sigma, ncores = n_cores)
    } else {
      mvnrv <-
        rmvnorm(n, mean = rep(0, p), sigma = Sigma, checkSymmetry = FALSE, method="eigen")
    }
  }
  
  mvnrvq <- apply(mvnrv, 2, stats::pnorm)

  return(mvnrvq)
}

## fix integer overflow issue when # of gene* # of cell is too larger in rnorm(n * ncol(sigma))
## This function comes from package Mvnorm.
rmvnorm <- function(n, mean = rep(0, nrow(sigma)), sigma = diag(length(mean)),
                    method=c("eigen", "svd", "chol"), pre0.9_9994 = FALSE, checkSymmetry = TRUE)
{
  if (checkSymmetry && !isSymmetric(sigma, tol = sqrt(.Machine$double.eps),
                                    check.attributes = FALSE)) {
    stop("sigma must be a symmetric matrix")
  }
  if (length(mean) != nrow(sigma))
    stop("mean and sigma have non-conforming size")

  method <- match.arg(method)

  R <- if(method == "eigen") {
    ev <- eigen(sigma, symmetric = TRUE)
    if (!all(ev$values >= -sqrt(.Machine$double.eps) * abs(ev$values[1]))){
      warning("Sigma is numerically not positive semidefinite")
    }
    ## ev$vectors %*% diag(sqrt(ev$values), length(ev$values)) %*% t(ev$vectors)
    ## faster for large  nrow(sigma):
    t(ev$vectors %*% (t(ev$vectors) * sqrt(pmax(ev$values, 0))))
  }
  else if(method == "svd"){
    s. <- svd(sigma)
    if (!all(s.$d >= -sqrt(.Machine$double.eps) * abs(s.$d[1]))){
      warning("Sigma is numerically not positive semidefinite")
    }
    t(s.$v %*% (t(s.$u) * sqrt(pmax(s.$d, 0))))
  }
  else if(method == "chol"){
    R <- chol(sigma, pivot = TRUE)
    R[, order(attr(R, "pivot"))]
  }

  retval <- matrix(stats::rnorm(as.double(n) * ncol(sigma)), nrow = n, byrow = !pre0.9_9994) %*%  R
  retval <- sweep(retval, 2, mean, "+")
  colnames(retval) <- names(mean)
  retval
}

#' Perform the likelihood ratio test
#'
#' \code{perform_lrt} performs the likelihood ratio test to compare two list of marginal models.
#'
#' The function takes two lists of marginal models (by default, the first list is the alternative and the second is the null)
#' from \code{\link{fit_marginal}}. Note that LRT only makes sense for NESTED models. This can be quite tricky if you use penalized-splines (e.g., for trajectory data).
#'
#' @param alter_marginal A list of marginal models from the alternative hypothesis.
#' @param null_marginal A list of marginal models from the null hypothesis. It must be strictly nested in the alternative model.
#'
#' @return A data.frame of the LRT result.
#' @examples
#' data(example_sce)
#' my_data <- construct_data(
#' sce = example_sce,
#' assay_use = "counts",
#' celltype = "cell_type",
#' pseudotime = "pseudotime",
#' spatial = NULL,
#' other_covariates = NULL,
#' corr_by = "cell_type"
#' )
#'
#' my_data2 <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "pseudotime",
#'   ncell = 10000
#' )
#'
#' my_marginal1 <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "1",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#' )
#' my_marginal2 <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#' )
#' my_fit1 <- lapply(my_marginal1, function(x)x$fit)
#' my_fit2 <- lapply(my_marginal2, function(x)x$fit)
#' my_pvalue <- perform_lrt(my_fit2, my_fit1)
#'
#' @export perform_lrt

perform_lrt <- function(alter_marginal,
                        null_marginal) {

  p_tab <- mapply(function(fit1, fit2) {
    ## Check if same models for fit1/2
    sm <- identical(class(fit1), class(fit2))

    ## get predicted probabilities for both models
    m1y <- fit1$y
    m2y <- fit2$y
    m1n <- length(m1y)
    m2n <- length(m2y)
    if(m1n==0 | m2n==0)
      stop("Could not extract dependent variables from models.")

    if(m1n != m2n)
      stop(paste("Models appear to have different numbers of observations.\n",
                 "Model 1 has ",m1n," observations.\n",
                 "Model 2 has ",m2n," observations.\n",
                 sep="")
      )

    if(any(m1y != m2y)){
      stop(paste("Models appear to have different values on dependent variables.\n"))
    }

    ## gather up degrees of freedom
    if(methods::is(fit1, "gamlss")) {
      k1 <- fit1$df.fit
    } else if (methods::is(fit1, "gam"))
    {
      k1 <- sum(fit1$edf2 + fit1$edf1 - fit1$edf)
    } else {
      stop("Model must be either gamlss or mgcv::gam!")
    }

    if(methods::is(fit2, "gamlss")) {
      k2 <- fit2$df.fit
    } else if (methods::is(fit2, "gam"))
    {
      k2 <- sum(fit2$edf2 + fit2$edf1 - fit2$edf)
    } else {
      stop("Model must be either gamlss or mgcv::gam!")
    }

    #k1 <- length(coef(m1))
    #k2 <- length(coef(m2))

    ll1 <- stats::logLik(fit1)
    ll2 <- stats::logLik(fit2)

    sign_k <- sign(k1 - k2)
    lr <- -2*(ll2 - ll1)

    llk1 <- ll1/k1
    llk2 <- ll1/k2

    lrt.p <- ifelse(lr > 0, stats::pchisq(lr, (k1-k2), lower.tail = FALSE), NA)

    res <- c(sm, ll1, ll2, k1, k2, lrt.p)
    names(res) <- c("same_model", "LogLik_alter", "LogLik_null", "df_alter", "df_null", "p_value")
    return(res)
  }, fit1 = alter_marginal, fit2 = null_marginal)
  p_tab <- as.data.frame(t(p_tab))

  return(p_tab)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    message = FALSE,
    collapse = TRUE,
    comment = "#>",
    crop = NULL ## Related to https://stat.ethz.ch/pipermail/bioc-devel/2020-April/016656.html
)
tools::R_user_dir("scDesign3", which="cache")

## ----message=FALSE, warning=FALSE, results='hide'-----------------------------
library(scDesign3)
library(SingleCellExperiment)
library(ggplot2)
theme_set(theme_bw())

## -----------------------------------------------------------------------------
#example_sce <- readRDS((url("https://figshare.com/ndownloader/files/40581992")))
#print(example_sce)
#example_sce <- example_sce[1:100, ]

data("example_count", package = "scDesign3")
data("pseudotime", package = "scDesign3")

example_sce <- SingleCellExperiment::SingleCellExperiment(assays = list(counts = example_count, logcounts = log1p(example_count)),
  colData = DataFrame(pseudotime = pseudotime))

## ----message=FALSE, warning=FALSE, results='hide'-----------------------------
set.seed(123)
example_simu <- scdesign3(
    sce = example_sce,
    assay_use = "counts",
    celltype = NULL,
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    mu_formula = "s(pseudotime, k = 10, bs = 'cr')", # mgcg::gam stype formula input.
    sigma_formula = "1", # If you want your dispersion also varies along pseudotime, use "s(pseudotime, k = 5, bs = 'cr')". This will significantly increase the computational cost.
    family_use = "nb",
    n_cores = 2, # Consider the balance between your ROM and threads.
    usebam = FALSE,
    corr_formula = "1",
    copula = "gaussian",
    DT = TRUE,
    pseudo_obs = FALSE,
    return_model = FALSE,
    nonzerovar = FALSE
  )

## -----------------------------------------------------------------------------
dim(example_simu$new_count)

## -----------------------------------------------------------------------------
logcounts(example_sce) <- log1p(counts(example_sce))
simu_sce <- SingleCellExperiment(list(counts = example_simu$new_count), colData = example_simu$new_covariate)
logcounts(simu_sce) <- log1p(counts(simu_sce))

## -----------------------------------------------------------------------------
set.seed(123)
compare_figure <- plot_reduceddim(ref_sce = example_sce, 
                                  sce_list = list(simu_sce), 
                                  name_vec = c("Reference", "scDesign3"),
                                  assay_use = "logcounts", 
                                  if_plot = TRUE, 
                                  color_by = "pseudotime", 
                                  n_pc = 20)
plot(compare_figure$p_umap)

## -----------------------------------------------------------------------------
sessionInfo()

#' Fit the marginal models
#'
#' \code{fit_marginal} fits the per-feature regression models.
#'
#' The function takes the result from \code{\link{construct_data}} as the input,
#' and fit the regression models for each feature based on users' specification.
#'
#' @param data An object from \code{\link{construct_data}}.
#' @param predictor A string of the predictor for the gam/gamlss model. Default is "gene". This is just a name.
#' @param mu_formula A string of the mu parameter formula. It follows the format of formula in \code{\link[mgcv]{bam}}. Note: if the formula has multiple smoothers (\code{s()}) (we do not recommend this), please put the one with largest k (most complex one) as the first one. 
#' @param sigma_formula A string of the sigma parameter formula
#' @param family_use A string or a vector of strings of the marginal distribution.
#' Must be one of 'binomial', 'poisson', 'nb', 'zip', 'zinb' or 'gaussian', which represent 'poisson distribution',
#' 'negative binomial distribution', 'zero-inflated poisson distribution', 'zero-inflated negative binomial distribution',
#' and 'gaussian distribution' respectively.
#' @param n_cores An integer. The number of cores to use.
#' @param usebam A logic variable. If use \code{\link[mgcv]{bam}} for acceleration.
#' @param edf_flexible A logic variable. It uses simpler model to accelerate the marginal fitting with a mild loss of accuracy. If TRUE, the fitted regression model will use the fitted relationship between Gini coefficient and the effective degrees of freedom on a random selected gene sets. Default is FALSE.
#' @param parallelization A string indicating the specific parallelization function to use.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#' @param trace A logic variable. If TRUE, the warning/error log and runtime for gam/gamlss will be returned.
#' will be returned, FALSE otherwise. Default is FALSE.
#' @param simplify A logic variable. If TRUE, the fitted regression model will only keep the essential contains for \code{predict}. Default is FALSE.
#' @param filter_cells A logic variable. If TRUE, when all covariates used for fitting the GAM/GAMLSS model are categorical, the code will check each unique combination of categories and remove cells in that category if it has all zero gene expression for each fitted gene.
#' @return A list of fitted regression models. The length is equal to the total feature number.
#' @examples
#'   data(example_sce)
#'   my_data <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "1"
#'   )
#'   my_marginal <- fit_marginal(
#'   data = my_data,
#'   mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
#'   sigma_formula = "1",
#'   family_use = "nb",
#'   n_cores = 1,
#'   usebam = FALSE
#'   )
#'
#' @export fit_marginal
#'
fit_marginal <- function(data,
                         predictor = "gene", ## Fix this later.
                         mu_formula,
                         sigma_formula,
                         family_use,
                         n_cores,
                         usebam = FALSE,
                         edf_flexible = FALSE,
                         parallelization = "mcmapply",
                         BPPARAM = NULL,
                         trace = FALSE, 
                         simplify = FALSE,
                         filter_cells = FALSE) {
  count_mat <-  data$count_mat
  dat_cov <- data$dat
  filtered_gene <- data$filtered_gene
  feature_names <- colnames(count_mat)
  
  
  # Extract K from mu formula
  matches <- regexpr("k\\s*=\\s*([0-9]+)", mu_formula, perl = TRUE)
  extracted_value <- regmatches(mu_formula, matches)
  extracted_K <- as.numeric(sub("k\\s*=\\s*", "", extracted_value))
  if(identical(extracted_K, numeric(0))) {
    extracted_K <- 0
  }
  
  # Randomly select genes for edf fitting
  num <- 100
  if(dim(count_mat)[2] > num & extracted_K >= 200 & edf_flexible == TRUE){
    edf_fitting <- TRUE
    
    # genes for fitting edf-gini relationship
    edf_gini_genes <- sample(seq_len(dim(count_mat)[2]), num)
    edf_gini_count_mat <-  count_mat[,edf_gini_genes]
    edf_gini_feature_names <- feature_names[edf_gini_genes]
    
    # genes for flexible edf
    edf_flexible_genes <- seq_len(dim(count_mat)[2])[-edf_gini_genes]
    edf_flexible_count_mat <- count_mat[,-edf_gini_genes]
    edf_flexible_feature_names <- feature_names[-edf_gini_genes]
    
  }else{
    edf_fitting <- FALSE
  }
  
  
  ## Check family_use
  if(length(family_use) == 1) {
    if(edf_fitting == TRUE) {
      edf_gini_family_use <- rep(family_use, length(edf_gini_feature_names))
      edf_flexible_family_use <- rep(family_use, length(edf_flexible_feature_names))
    }
    family_use <- rep(family_use, length(feature_names))
  }
  if(length(family_use) != length(feature_names)) {
    stop("The family_use must be either a single string or a vector with the same length as all features!")
  }
  
  
  fit_model_func <- function(gene,
                             family_gene,
                             dat_use,
                             #mgcv_formula,
                             mu_formula,
                             sigma_formula,
                             predictor,
                             count_mat,
                             edf=NULL
  ) {
    
    if(!is.null(edf)){
      mu_formula_ex <- sub("(k\\s*=).*", "\\1", mu_formula)
      mu_formula = paste0(mu_formula_ex, round(edf[[gene]]), ")")
    }
    
    mgcv_formula <-
      stats::formula(paste0(predictor, "~", mu_formula))
    
    ## If use the mgcv s() smoother
    mu_mgcvform <- grepl("s\\(", mu_formula) | grepl("te\\(", mu_formula)
    
    ## If use bam to fit marginal distribution
    usebam <- usebam & mu_mgcvform ## If no smoothing terms, no need to to use bam.
    if(usebam){
      fitfunc = mgcv::bam
    }else{
      fitfunc = mgcv::gam
    }
    
    if (mu_mgcvform) {
      terms <- attr(stats::terms(mgcv_formula), "term.labels")
      terms_smooth <- terms[which(grepl("s\\(", terms))] 
      
      if(usebam){
        terms_smooth_update <- sapply(terms_smooth, function(x){paste0("ba(~", x, ", method = 'fREML', gc.level = 0, discrete = TRUE)")})
        if(length(terms_smooth) == length(terms)){## only contain smooth terms
          mu_formula <-
            stats::formula(paste0(predictor, "~", paste0(terms_smooth_update, collapse = "+")))
        }else{
          terms_linear <- terms[which(!grepl("s\\(", terms))] 
          terms_update <- c(terms_linear, terms_smooth_update)
          mu_formula <-
            stats::formula(paste0(predictor, "~", paste0(terms_update, collapse = "+")))
        }
      }else{
        terms_smooth_update <- sapply(terms_smooth, function(x){paste0("ga(~", x, ", method = 'REML')")})
        if(length(terms_smooth) == length(terms)){## only contain smooth terms
          mu_formula <-
            stats::formula(paste0(predictor, "~", paste0(terms_smooth_update, collapse = "+")))
        }else{
          terms_linear <- terms[which(!grepl("s\\(", terms))] 
          terms_update <- c(terms_linear, terms_smooth_update)
          mu_formula <-
            stats::formula(paste0(predictor, "~", paste0(terms_update, collapse = "+")))
        }
      }
    }
    else {
      mu_formula <- stats::formula(paste0(predictor, "~", mu_formula))
    }
    
    sigma_mgcvform <- grepl("s\\(", sigma_formula) | grepl("te\\(", sigma_formula)
    if (sigma_mgcvform) {
      temp_sigma_formula <- stats::formula(paste0(predictor, "~", sigma_formula))
      terms <- attr(stats::terms(temp_sigma_formula), "term.labels")
      terms_smooth <- terms[which(grepl("s\\(", terms))]
      if(usebam){
        terms_smooth_update <- sapply(terms_smooth, function(x){paste0("ba(~", x, ", method = 'fREML', gc.level = 0, discrete = TRUE)")})
        if(length(terms_smooth) == length(terms)){## only contain smooth terms
          sigma_formula <-
            stats::formula(paste0("~", paste0(terms_smooth_update, collapse = "+")))
        }else{
          terms_linear <- terms[which(!grepl("s\\(", terms))] 
          terms_update <- c(terms_linear, terms_smooth_update)
          sigma_formula <-
            stats::formula(paste0("~", paste0(terms_update, collapse = "+")))
        }
      }else{
        terms_smooth_update <- sapply(terms_smooth, function(x){paste0("ga(~", x, ", method = 'REML')")})
        if(length(terms_smooth) == length(terms)){## only contain smooth terms
          sigma_formula <-
            stats::formula(paste0("~", paste0(terms_smooth_update, collapse = "+")))
        }else{
          terms_linear <- terms[which(!grepl("s\\(", terms))] 
          terms_update <- c(terms_linear, terms_smooth_update)
          sigma_formula <-
            stats::formula(paste0("~", paste0(terms_update, collapse = "+")))
        }
      }
      
    } else {
      sigma_formula <- stats::formula(paste0("~", sigma_formula))
    }
    
    
    ## Add gene expr
    dat_use$gene <- count_mat[, gene]
    
    ## For error/warning logging
    add_log <- function(function_name, type, message) {
      new_l <- logs
      new_log <- list(function_name = function_name,
                      type = type,
                      message =  message)
      new_l[[length(new_l) + 1]]  <- new_log
      logs <<- new_l
    }
    
    logs <- list()
    ## Don't fit marginal if gene only have two or less non-zero expression
    if(!is.null(filtered_gene) & gene %in% filtered_gene){
      add_log("fit_marginal","warning", paste0(gene, "is expressed in too few cells."))
      return(list(fit = NA, warning = logs, time = c(NA,NA)))
    }
    
    if(filter_cells){
      all_covariates <- all.vars(mgcv_formula)[-1]
      dat_cova <- dat_use[, all_covariates]
      check_factor <- all(sapply(dat_cova,is.factor))
      if (length(all_covariates) > 0 & check_factor){
        remove_idx_list <- lapply(all_covariates, function(x){
          curr_x <- tapply(dat_use$gene, dat_use[,x], sum)
          zero_group <- which(curr_x==0)
          if(length(zero_group) == 0){
            return(list(idx = NA, changeFormula = FALSE))
          }else{
            type <- names(curr_x)[zero_group]
            if(length(type) == length(unique(dat_use[,x])) - 1){
              return(list(idx = NA, changeFormula = TRUE))
            }
            return(list(idx = which(dat_use[,x] %in% type), changeFormula = FALSE))
          }
          
        })
        names(remove_idx_list) <- all_covariates
        remove_idx <- lapply(remove_idx_list, function(x)x$idx)
        remove_cell <- unlist(remove_idx)
        if(all(is.na(remove_cell))){
          remove_cell <- NA
        }else{
          remove_cell <- unique(stats::na.omit(remove_cell))
        }
        if(length(remove_cell) > 0 && !any(is.na(remove_cell))){
          dat_use <- dat_use[-remove_cell,]
        }
        
        changeFormula <-  sapply(remove_idx_list, function(x)x$changeFormula)
        if(length(which(changeFormula)) > 0){
          changeVars <- names(which(changeFormula))
          formulaUpdate <- paste0(changeVars, collapse = "-")
          mgcv_formula <- stats::update.formula(mgcv_formula, stats::as.formula(paste0("~.-",formulaUpdate)))
          mu_formula <- stats::update.formula(mu_formula, stats::as.formula(paste0("~.-",formulaUpdate)))
          sigmaVars <- which(changeVars %in% as.character(sigma_formula))
          if(length(sigmaVars) > 0){
            formulaUpdate <- paste0(changeVars[sigmaVars], collapse = "-")
          }
          sigma_formula = stats::update.formula(sigma_formula, stats::as.formula(paste0("~.-",formulaUpdate)))
        }
        
      }else{
        remove_cell <- NA
      }
    }else{
      remove_cell <- NA
    }
    
    
    time_list <- c(NA,NA)
    
    if (family_gene == "binomial") {
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <-fitfunc(formula = mgcv_formula, data = dat_use, family = "binomial", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
          
        })
      
      if (sigma_formula != "~1") {
        gamlss.fit <- withCallingHandlers(
          tryCatch({
            start.time = Sys.time()
            res <- gamlss::gamlss(
              formula = mu_formula,
              #sigma.formula = sigma_formula, ## Binomial is one para dist.
              data = dat_use,
              family = gamlss.dist::BI,
              control = gamlss::gamlss.control(trace = FALSE, c.crit = 0.1)
            )
            end.time = Sys.time()
            time = as.numeric(end.time - start.time)
            time_list[2] <- time
            res
          }, error=function(e) {
            add_log("gamlss","error", toString(e))
            NULL
          }), warning=function(w) {
            add_log("gamlss","warning", toString(w))
          })
        
      } else {
        gamlss.fit <- NULL
      }
    } else if (family_gene == "poisson") {
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <-fitfunc(formula = mgcv_formula, data = dat_use, family = "poisson", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
        })
      
      if (sigma_formula != "~1") {
        gamlss.fit <- withCallingHandlers(
          tryCatch({
            start.time = Sys.time()
            res <- gamlss::gamlss(
              formula = mu_formula,
              #sigma.formula = sigma_formula, ## Poisson has constant mean
              data = dat_use,
              family = gamlss.dist::PO,
              control = gamlss::gamlss.control(trace = FALSE, c.crit = 0.1))
            end.time = Sys.time()
            time = as.numeric(end.time - start.time)
            time_list[2] <- time
            res
          }, error=function(e) {
            add_log("gamlss","error", toString(e))
            NULL
          }), warning=function(w) {
            add_log("gamlss","warning", toString(w))
          })
      } else {
        gamlss.fit <- NULL
      }
    } else if (family_gene == "gaussian") {
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <- fitfunc(formula = mgcv_formula, data = dat_use, family = "gaussian", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
        })
      
      if (sigma_formula != "~1") {
        gamlss.fit<- withCallingHandlers(
          tryCatch({
            start.time = Sys.time()
            res <- gamlss::gamlss(
              formula = mu_formula,
              sigma.formula = sigma_formula,
              data = dat_use,
              family = gamlss.dist::NO,
              control = gamlss::gamlss.control(trace = FALSE, c.crit = 0.1)
            )
            end.time = Sys.time()
            time = as.numeric(end.time - start.time)
            time_list[2] <- time
            res
          }, error=function(e) {
            add_log("gamlss","error", toString(e))
            NULL
          }), warning=function(w) {
            add_log("gamlss","warning", toString(w))
          })
        
      } else {
        gamlss.fit <- NULL
      }
    } else if (family_gene == "nb"){
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <-fitfunc(formula = mgcv_formula, data = dat_use, family = "nb", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
        })
      
      if (sigma_formula != "~1") {
        gamlss.fit <- withCallingHandlers(
          tryCatch({
            start.time = Sys.time()
            res <- gamlss::gamlss(
              formula = mu_formula,
              sigma.formula = sigma_formula,
              data = dat_use,
              family = gamlss.dist::NBI,
              control = gamlss::gamlss.control(trace = FALSE,  c.crit = 0.1)
            )
            end.time = Sys.time()
            time = as.numeric(end.time - start.time)
            time_list[2] <- time
            res
          }, error=function(e) {
            add_log("gamlss","error", toString(e))
            NULL
          }), warning=function(w) {
            add_log("gamlss","warning", toString(w))
          })
      } else {
        gamlss.fit <- NULL
      }
    } else if (family_gene == "zip") {
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <-fitfunc(formula = mgcv_formula, data = dat_use, family = "poisson", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
        })
      gamlss.fit <- withCallingHandlers(
        tryCatch({
          start.time = Sys.time()
          res <- gamlss::gamlss(
            formula = mu_formula,
            sigma.formula = mu_formula, ## Here sigma is the dropout prob, not variance!
            data = dat_use,
            family = gamlss.dist::ZIP,
            control = gamlss::gamlss.control(trace = FALSE, c.crit = 0.1)
            
          )
          end.time = Sys.time()
          time = as.numeric(end.time - start.time)
          time_list[2] <- time
          res
        }, error=function(e) {
          add_log("gamlss","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gamlss","warning", toString(w))
        })
      
    } else if (family_gene == "zinb"){
      mgcv.fit <- withCallingHandlers(
        tryCatch({
          start.time <- Sys.time()
          res <- fitfunc(formula = mgcv_formula, data = dat_use, family = "nb", discrete = usebam)
          end.time <- Sys.time()
          time <- as.numeric(end.time - start.time)
          time_list[1] <- time
          res
        }, error=function(e) {
          add_log("gam","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gam","warning", toString(w))
        })
      gamlss.fit <- withCallingHandlers(
        tryCatch({
          start.time = Sys.time()
          res <- gamlss::gamlss(
            formula = mu_formula,
            sigma.formula = sigma_formula,
            nu.formula = mu_formula, ## Here nu is the dropout probability!
            data = dat_use,
            family = gamlss.dist::ZINBI,
            control = gamlss::gamlss.control(trace = FALSE, c.crit = 0.1)
          )
          end.time = Sys.time()
          time = as.numeric(end.time - start.time)
          time_list[2] <- time
          res
        }, error=function(e) {
          add_log("gamlss","error", toString(e))
          NULL
        }), warning=function(w) {
          add_log("gamlss","warning", toString(w))
        })
      
    } else {
      stop("The regression distribution must be one of gaussian, poisson, nb, zip or zinb!")
    }
    
    ## Check if gamlss is fitted.
    if (!"gamlss" %in% class(gamlss.fit)) {
      if (sigma_formula != "~1") {
        message(paste0(gene, " uses mgcv::gam due to gamlss's error!"))
        ## gamlss.fit contains warning message
        if(!is.null(gamlss.fit)){
          ## check whether gam has warning messages
          if(is.null(warn)){
            warn = gamlss.fit
          }else{
            warn = c(warn, gamlss.fit)
          }
        }
      }
      
      fit <- mgcv.fit
    } else {
      
      mean_vec <- stats::predict(gamlss.fit, type = "response", what = "mu", data = dat_use)
      theta_vec <-
        stats::predict(gamlss.fit, type = "response", what = "sigma", data = dat_use)
      
      if_infinite <- (sum(is.infinite(mean_vec + theta_vec)) > 0)
      if_overmax <- (max(mean_vec, na.rm = TRUE) > 10* max(dat_use$gene, na.rm = TRUE))
      if(family_gene %in% c("nb","zinb")){
        #if_overdisp <- (min(theta_vec, na.rm = TRUE) < 1/ 1000)
        if_overdisp <- (max(theta_vec, na.rm = TRUE) > 1000)
        
      }else{
        if_overdisp <- FALSE
      }
      
      
      if (if_infinite | if_overmax | if_overdisp) {
        add_log("fit_marginal","warning", paste0(gene, " gamlss returns abnormal fitting values!"))
        #message(paste0(gene, " gamlss returns abnormal fitting values!"))
        fit <- mgcv.fit
      } else if (stats::AIC(mgcv.fit) - stats::AIC(gamlss.fit) < -Inf) {
        message(paste0(
          gene,
          "'s gamlss AIC is not signifincantly smaller than gam!"
        ))
        fit <- mgcv.fit
      }
      else {
        fit <- gamlss.fit
      }
    }
    
    if(simplify) {
      fit <- simplify_fit(fit)
    }
    
    if(trace){
      return(list(fit = fit, warning = logs, time = time_list, removed_cell = remove_cell))
    }
    return(list(fit = fit,removed_cell = remove_cell))
    #return(fit)
  }
  
  paraFunc <- parallel::mcmapply
  if(.Platform$OS.type == "windows"){
    BPPARAM <- BiocParallel::SnowParam()
    parallelization <- "bpmapply"
  }
  if(parallelization == "bpmapply"){
    paraFunc <- BiocParallel::bpmapply
  }
  if(parallelization == "pbmcmapply"){
    paraFunc <- pbmcapply::pbmcmapply
  }
  # If not using edf flexible fitting
  if(edf_fitting==FALSE){
    if(parallelization == "bpmapply"){
      if(class(BPPARAM)[1] != "SerialParam"){
        BPPARAM$workers <- n_cores
      }
      model_fit <- suppressMessages(paraFunc(fit_model_func, gene = feature_names,
                                             family_gene = family_use,
                                             MoreArgs = list(dat_use = dat_cov,
                                                             #mgcv_formula = mgcv_formula,
                                                             mu_formula = mu_formula,
                                                             sigma_formula = sigma_formula,
                                                             predictor = predictor,
                                                             count_mat = count_mat),
                                             SIMPLIFY = FALSE, BPPARAM = BPPARAM))
    }else{
      model_fit <-  suppressMessages(paraFunc(fit_model_func, gene = feature_names,
                                              family_gene = family_use,
                                              mc.cores = n_cores,
                                              MoreArgs = list(dat_use = dat_cov,
                                                              #mgcv_formula = mgcv_formula,
                                                              mu_formula = mu_formula,
                                                              sigma_formula = sigma_formula,
                                                              predictor = predictor,
                                                              count_mat = count_mat),
                                              SIMPLIFY = FALSE))
    }
  }else{ 
    # If using edf flexible fitting
    
    if(parallelization == "bpmapply"){
      if(class(BPPARAM)[1] != "SerialParam"){
        BPPARAM$workers <- n_cores
      }
      # Fit model to selected edf_gini_genes 
      model_fit_edf_gini <- suppressMessages(paraFunc(fit_model_func, gene = edf_gini_feature_names,
                                                      family_gene = edf_gini_family_use,
                                                      MoreArgs = list(dat_use = dat_cov,
                                                                      #mgcv_formula = mgcv_formula,
                                                                      mu_formula = mu_formula,
                                                                      sigma_formula = sigma_formula,
                                                                      predictor = predictor,
                                                                      count_mat = edf_gini_count_mat),
                                                      SIMPLIFY = FALSE, BPPARAM = BPPARAM))
    }else{
      
      # Fit model to selected edf_gini_genes
      model_fit_edf_gini <-  suppressMessages(paraFunc(fit_model_func, gene = edf_gini_feature_names,
                                                       family_gene = edf_gini_family_use,
                                                       mc.cores = n_cores,
                                                       MoreArgs = list(dat_use = dat_cov,
                                                                       #mgcv_formula = mgcv_formula,
                                                                       mu_formula = mu_formula,
                                                                       sigma_formula = sigma_formula,
                                                                       predictor = predictor,
                                                                       count_mat = edf_gini_count_mat),
                                                       SIMPLIFY = FALSE))
    }
    
    
    # Extract the fitted edf
    edf <- rep(NA, length(model_fit_edf_gini))
    for(i in 1:length(model_fit_edf_gini)){
      res_ind <- model_fit_edf_gini[i]
      if(lengths(res_ind)==2){
        res_ind <- res_ind[[names(res_ind)]]
        edf[i] <- sum(res_ind$fit$edf)
      }
    }
    
    # Fit a edf-gini relationship for edf_gini_genes
    edf_gini_count_gini <- apply(log(edf_gini_count_mat+1), MARGIN=2, FUN=gini)
    edf_gini_df <- data.frame(edf=edf, gini=edf_gini_count_gini)
    lm_edf_gini <- stats::lm(edf~gini, data=edf_gini_df)
    # Upper bound for the lm coef
    #coef <- confint(lm_edf_gini)[,2]
    
    
    # Predict edf for edf_flexible_genes
    edf_flexible_count_gini <- apply(log(edf_flexible_count_mat+1), MARGIN=2, FUN=gini)
    edf_flexible_df <- data.frame(gini=edf_flexible_count_gini)
    edf_flexible_predicted <- stats::predict(lm_edf_gini, edf_flexible_df, se.fit = TRUE, interval = "confidence", level = 0.95)
    edf_flexible_predicted_upr <- edf_flexible_predicted$fit[,3]
    
    
    # Fit again for the rest genes
    if(parallelization == "bpmapply"){
      if(class(BPPARAM)[1] != "SerialParam"){
        BPPARAM$workers <- n_cores
      }
      model_fit_edf_flexible <- suppressMessages(paraFunc(fit_model_func, gene = edf_flexible_feature_names,
                                                          family_gene = edf_flexible_family_use,
                                                          MoreArgs = list(dat_use = dat_cov,
                                                                          #mgcv_formula = mgcv_formula,
                                                                          mu_formula = mu_formula,
                                                                          sigma_formula = sigma_formula,
                                                                          predictor = predictor,
                                                                          count_mat = edf_flexible_count_mat,
                                                                          edf=edf_flexible_predicted_upr),
                                                          SIMPLIFY = FALSE, BPPARAM = BPPARAM))
    }else{
      model_fit_edf_flexible <-  suppressMessages(paraFunc(fit_model_func, gene = edf_flexible_feature_names,
                                                           family_gene = edf_flexible_family_use,
                                                           mc.cores = n_cores,
                                                           MoreArgs = list(dat_use = dat_cov,
                                                                           #mgcv_formula = mgcv_formula,
                                                                           mu_formula = mu_formula,
                                                                           sigma_formula = sigma_formula,
                                                                           predictor = predictor,
                                                                           count_mat = edf_flexible_count_mat,
                                                                           edf=edf_flexible_predicted_upr),
                                                           SIMPLIFY = FALSE))
    }
    
    
    # Combine model_fit_edf_gini and model_fit_edf_flexible
    model_fit <- vector(mode = "list", length = length(feature_names))
    names(model_fit) <- feature_names
    
    # Populate the new list based on indices:
    for (index in names(model_fit_edf_gini)) {
      model_fit[[index]] <- model_fit_edf_gini[[index]]
    }
    for (index in names(model_fit_edf_flexible)) {
      model_fit[[index]] <- model_fit_edf_flexible[[index]]
    }
  } 
  
  # if(!is.null(model_fit$warning)) {
  #   #stop("Model has warning!")
  #   model_fit <- model_fit$value
  # }
  return(model_fit)
}

simplify_fit <- function(cm) {
  ## This function is modified from https://win-vector.com/2014/05/30/trimming-the-fat-from-glm-models-in-r/
  cm$y = c()
  #cm$model = c()
  
  cm$residuals = c()
  cm$fitted.values = c()
  cm$effects = c()
  cm$qr$qr = c()  
  cm$linear.predictors = c()
  cm$weights = c()
  cm$prior.weights = c()
  cm$data = c()
  
  #cm$mu.x = c()
  #cm$sigma.x = c()
  #cm$nu.x = c()
  
  #cm$family$variance = c()
  #cm$family$dev.resids = c()
  #cm$family$aic = c()
  #cm$family$validmu = c()
  #cm$family$simulate = c()
  attr(cm$terms,".Environment") = c()
  attr(cm$formula,".Environment") = c()
  
  attr(cm$mu.terms,".Environment") = c()
  attr(cm$mu.formula,".Environment") = c()
  
  attr(cm$sigma.terms,".Environment") = c()
  attr(cm$sigma.formula,".Environment") = c()
  
  attr(cm$nu.terms,".Environment") = c()
  attr(cm$nu.formula,".Environment") = c()
  cm
}

## Function from R package reldist by Dr. Mark S. Handcock      
gini <- function(x, weights=rep(1,length=length(x))){
  ox <- order(x)
  x <- x[ox]
  weights <- weights[ox]/sum(weights)
  p <- cumsum(weights)
  nu <- cumsum(weights*x)
  n <- length(nu)
  nu <- nu / nu[n]
  sum(nu[-1]*p[-n]) - sum(nu[-n]*p[-1])
}

#' Construct the input data (covariate matrix and expression matrix)
#'
#' This function constructs the input data for \code{\link{fit_marginal}}.
#'
#' This function takes a \code{SingleCellExperiment} object as the input.
#' Based on users' choice, it constructs the matrix of covariates
#' (explanatory variables) and the expression matrix (e.g., count matrix for scRNA-seq).
#'
#' @param sce A \code{SingleCellExperiment} object.
#' @param assay_use A string which indicates the assay you will use in the sce. Default is 'counts'.
#' @param celltype A string of the name of cell type variable in the \code{colData} of the sce. Default is 'cell_type'.
#' @param pseudotime A string or a string vector of the name of pseudotime and (if exist)
#' multiple lineages. Default is NULL.
#' @param spatial A length two string vector of the names of spatial coordinates. Default is NULL.
#' @param other_covariates A string or a string vector of the other covariates you want to include in the data.
#' @param ncell The number of cell you want to simulate. Default is \code{dim(sce)[2]} (the same number as the input data).
#' If an arbitrary number is provided, the function will use Vine Copula to simulate a new covariate matrix.
#' @param corr_by A string or a string vector which indicates the groups for correlation structure. If '1', all cells have one estimated corr. If 'ind', no corr (features are independent). If others, this variable decides the corr structures.
#' @param parallelization A string indicating the specific parallelization function to use.
#' Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
#' \code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.
#' @param BPPARAM A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
#' this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
#' \code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.
#'
#' @return A list with the components:
#' \describe{
#'   \item{\code{count_mat}}{The expression matrix}
#'   \item{\code{dat}}{The original covariate matrix}
#'   \item{\code{newCovariate}}{The simulated new covariate matrix, is NULL if the parameter ncell is default}
#'   \item{\code{filtered_gene}}{The genes that are excluded in the marginal and copula fitting 
#' steps because these genes only express in less than two cells.}
#' }
#' @examples
#'   data(example_sce)
#'   my_data <- construct_data(
#'   sce = example_sce,
#'   assay_use = "counts",
#'   celltype = "cell_type",
#'   pseudotime = "pseudotime",
#'   spatial = NULL,
#'   other_covariates = NULL,
#'   corr_by = "1"
#'   )
#'
#'
#' @export construct_data
construct_data <- function(sce,
                           assay_use = "counts",
                           celltype,
                           pseudotime,
                           spatial,
                           other_covariates,
                           ncell = dim(sce)[2],
                           corr_by,
                           parallelization = "mcmapply",
                           BPPARAM = NULL) {
  
  ## check unique cell names and gene names
  if(length(unique(colnames(sce))) != dim(sce)[2]){
    stop("Please make sure your inputted SingleCellExperiment object does not have duplicate cell names")
  }
  if(length(unique(rownames(sce))) != dim(sce)[1]){
    stop("Please make sure your inputted SingleCellExperiment object does not have duplicate gene names")
  }
  ## Extract expression matrix
  count_mat <-
    t(as.matrix(SummarizedExperiment::assay(sce, assay_use)))
  ## Extract col data
  coldata_mat <- data.frame(SummarizedExperiment::colData(sce))
  
  ##
  if (is.null(celltype) & is.null(pseudotime) & is.null(spatial)) {
    stop("One of celltype, pseudotime and spatial must be provided!")
  } else {
    primary_covariate <- c(celltype, pseudotime, spatial)
    dat <- as.data.frame(coldata_mat[, primary_covariate, drop = FALSE])
  }
  
  if(!is.null(celltype)) {
    dat[, celltype] <- as.factor(dat[, celltype])}
  
  # ## Extract pseudotime / cell type / spatial
  # if (!is.null(celltype)) {
  #   celltype <- as.matrix(coldata_mat[, celltype, drop = FALSE])
  # }
  #
  # if (!is.null(pseudotime)) {
  #   pseudotime <- as.matrix(coldata_mat[, pseudotime, drop = FALSE])
  # }
  #
  # if (!is.null(spatial)) {
  #   spatial <- as.matrix(coldata_mat[, spatial, drop = FALSE])
  # }
  #
  #
  # if (covariate_use == "celltype") {
  #   dat <- data.frame(celltype)
  #   dat$cell_type <- as.factor(dat$cell_type)
  # } else if (covariate_use == "pseudotime") {
  #   n_l <- dim(pseudotime)[2]
  #   dat <- data.frame(pseudotime)
  # } else if (covariate_use == "spatial") {
  #   dat <- data.frame(spatial)
  # } else {
  #   stop("Covairate_use must be one of 'celltype', 'pseudotime' or 'spatial'!")
  # }
  
  ## Convert NA to -1
  #pseudotime[is.na(pseudotime)] <- -1
  
  ## dat is the input covariate matrix
  if (!is.null(other_covariates)) {
    other_covariates <- (coldata_mat[, other_covariates, drop = FALSE])
    dat <- dplyr::bind_cols(dat, other_covariates)
    if("condition" %in% colnames(other_covariates)){
      dat$condition <- as.factor(dat$condition)
    }
    if("batch" %in% colnames(other_covariates)){
      dat$batch <- as.factor(dat$batch)
    }
  }
  
  ## check if user wants to simulate new number of cells
  if(ncell != dim(dat)[1]){
    newCovariate <- as.data.frame(simuCovariateMat(dat,ncell, parallelization, BPPARAM))
  }else{
    newCovariate <- dat
  }
  
  # identify groups
  n_gene <- dim(sce)[1]
  n_cell <- dim(sce)[2]
  group <- unlist(corr_by)
  if(ncell != dim(dat)[1]){
    if (group[1] == "1") {
      corr_group <- rep(1, n_cell)
      corr_group2 <- rep(1, dim(newCovariate)[1])
    } else if (group[1] == "ind"){
      corr_group <- rep("ind", n_cell)
      corr_group2 <- rep("ind", dim(newCovariate)[1])
    } else if (group[1] == "pseudotime" | length(group) > 1) {
      ## For continuous pseudotime, discretize it
      corr_group <- SummarizedExperiment::colData(sce)[, group]
      mclust_mod <- mclust::Mclust(corr_group, G = seq_len(5))
      corr_group <- mclust_mod$classification
      
      corr_group2 <- newCovariate[, group]
      corr_group2 <- mclust::predict.Mclust(mclust_mod, newdata = corr_group2)$classification
      
    } else {
      corr_group <- SummarizedExperiment::colData(sce)[, group]
      corr_group2 <- newCovariate[, group]
    }
    newCovariate$corr_group <- corr_group2
  }else{
    if (group[1] == "1") {
      corr_group <- rep(1, n_cell)
    } else if (group[1] == "ind"){
      corr_group <- rep("ind", n_cell)
    }else if (group[1] == "pseudotime" | length(group) > 1) {
      ## For continuous pseudotime, discretize it
      corr_group <- SummarizedExperiment::colData(sce)[, group]
      mclust_mod <- mclust::Mclust(corr_group, G = seq_len(5))
      
      corr_group <- mclust_mod$classification
      
    } else {
      corr_group <- SummarizedExperiment::colData(sce)[, group]
    }
  }
  dat$corr_group <- corr_group
  
  qc <- apply(count_mat, 2, function(x){
    return(length(which(x < 1e-5)) > length(x) - 2)
  })
  if(length(which(qc)) == 0){
    filtered_gene <- NULL
  }else{
    filtered_gene <- names(which(qc)) 
  }
  return(list(count_mat = count_mat, dat = dat, newCovariate = newCovariate, filtered_gene = filtered_gene))
}


## Simulate covariate matrix
simuCovariateMat <- function(covariate_mat,
                             n_cell_new = 50000,
                             parallelization = "mcmapply",
                             BPPARAM = NULL) {
  
  n_cell_ori <- dim(covariate_mat)[1]
  n_covraite_ori <- dim(covariate_mat)[2]
  
  if_factor_exist <- sum(sapply(covariate_mat, is.factor))
  if_numeric_exist <- sum(sapply(covariate_mat, is.numeric))
  
  
  
  df <- covariate_mat
  
  if(if_factor_exist) {
    df_all <-  dplyr::mutate(df, discrete_group = interaction(dplyr::select_if(df, is.factor), sep = "-"))
    
    df_list <- split(df_all, df_all$discrete_group)
    
    group_prop <-  table(df_all$discrete_group)/dim(df_all)[1]
    group_name <- names(group_prop)
    group_n_new <- stats::rmultinom(1, size = n_cell_new, prob = group_prop)
    paraFunc <- parallel::mcmapply
    if(.Platform$OS.type == "windows"){
      BPPARAM <- BiocParallel::SnowParam()
      parallelization <- "bpmapply"
    }
    if(parallelization == "bpmapply"){
      paraFunc <- BiocParallel::bpmapply
    }
    if(parallelization == "pbmcmapply"){
      paraFunc <- pbmcapply::pbmcmapply
    }
    
    if(if_numeric_exist) {
      dat_function <- {function(df, n) {
        df <- dplyr::select(df, -"discrete_group")
        df_numeric <- dplyr::select_if(df, is.numeric)
        df_factor <- dplyr::select_if(df, is.factor)
        if(n == 0){
          return(NA)
        }
        fit_kde <- rvinecopulib::vine(df_numeric, cores = 1)
        new_dat <- rvinecopulib::rvine(n, fit_kde)
        
        new_dat <- as.data.frame(new_dat)
        new_dat[colnames(df_factor)] <- df_factor[1, ]
        new_dat
        
      }}
      
      if(parallelization == "bpmapply"){
        new_dat_list <- paraFunc(FUN = dat_function , df = df_list, n = group_n_new, BPPARAM = BPPARAM,SIMPLIFY = FALSE)
      }else{
        new_dat_list <- paraFunc(FUN = dat_function , df = df_list, n = group_n_new,SIMPLIFY = FALSE)
      }
      covariate_new <- do.call("rbind", new_dat_list[!is.na(new_dat_list)])
    }
    else {
      dat_function <- function(df, n) {
        df <- dplyr::select(df, -"discrete_group")
        df_factor <- dplyr::select_if(df, is.factor)
        df_onerow <- as.data.frame(df_factor[1, ])
        colnames(df_onerow) <- colnames(df_factor)
        new_dat <- as.data.frame(df_onerow[rep(1, n), ])
        colnames(new_dat) <- colnames(df_onerow)
        new_dat
      }
      if(parallelization == "bpmapply"){
        new_dat_list <- paraFunc(FUN = dat_function, df = df_list, n = group_n_new, BPPARAM = BPPARAM, SIMPLIFY = FALSE)
      }else{
        new_dat_list <- paraFunc(FUN = dat_function, df = df_list, n = group_n_new,SIMPLIFY = FALSE)
      }
      covariate_new <- do.call("rbind", new_dat_list)
      
    }
  }
  else {
    fit_kde <- rvinecopulib::vine(df, cores = 1)
    covariate_new <- rvinecopulib::rvine(n_cell_new, fit_kde)
  }
  
  rownames(covariate_new) <- paste0("Cell", seq_len(n_cell_new))
  return(covariate_new)
}

#' A SingleCellExperiment object containing both cell type and pseudotime
#' @format A dataset with 10 rows (genes) and 1289 cols (cells)
#' @return The corresponding SingleCellExperiment object
#' @usage data("example_sce")
"example_sce"

#' A sparse matrix with example data
#' @format A dgCMatrix with 100 rows (genes) and 2087 cols (cells)
#' @return The corresponding sparse matrix object 'example_count'
#' @usage data("example_count")
"example_count"

#' A numeric vector of pseudotime values
#' @format A numeric vector with 2087 values
#' @return The corresponding pseudotime of the 'example_count'
#' @usage data("pseudotime")
"pseudotime"

#' Dimensionality reduction and visualization
#' 
#' \code{plot_reduceddim} performs the dimensionality reduction
#' 
#' This function takes a reference sce and a list of new sces, performs the dimensionality reduction on the reference data, 
#' projects the synthetic datasets on the same low dimensional space,
#' then visualize the results.
#'
#' @param ref_sce The reference sce.
#' @param sce_list A list of synthetic sce.
#' @param name_vec A vector of the names of each dataset. The length should be \code{length(sce_list) + 1}, where the first name is for \code{ref_sce}.
#' @param assay_use A string which indicates the assay you will use in the sce.
#' Default is 'logcounts'.
#' @param n_pc An integer of the number of PCs.
#' @param pc_umap A logic value of whether using PCs as the input of UMAP. Default is TRUE.
#' @param center A logic value of whether centering the data before PCA. Default is TRUE.
#' @param scale. A logic value of whether scaling the data before PCA. Default is TRUE.
#' @param if_plot A logic value of whether returning the plot. If FALSE, return the reduced dimensions of each dataset.
#' @param shape_by A string which indicates the column in \code{colData} used for shape.
#' @param color_by A string which indicates the column in \code{colData} used for color.
#' @param point_size A numeric value of the point size in the final plot. Default is 1.
#'
#' @return The ggplot or the data.frame of reduced dimensions.
#'
#' @import ggplot2
#'
#' @export plot_reduceddim

plot_reduceddim <- function(ref_sce,
                            sce_list,
                            name_vec,
                            assay_use = "logcounts",
                            pc_umap = TRUE,
                            n_pc = 50,
                            center = TRUE,
                            scale. = TRUE,
                            if_plot = TRUE,
                            shape_by = NULL,
                            color_by,
                            point_size = 1) {

  Method <- NULL ## Avoid check note.
  stopifnot(length(name_vec) == (length(sce_list) + 1))

  mat_ref <- t(as.matrix(SummarizedExperiment::assay(ref_sce, assay_use)))

  if(sum(matrixStats::colVars(mat_ref) == 0) > 0) {
    stop("The ref dataset contains 0 variance features. Please remove them.")
  }

  if(!is.null(shape_by)){
    if(!(shape_by %in% colnames(SummarizedExperiment::colData(ref_sce)))) {
      stop("The shape_by in not in your ref_sce's colData. Please double check the variable name for shape_by.")
    }
    shape_by_check <- sapply(sce_list, function(x){shape_by %in% colnames(SummarizedExperiment::colData(x))})
    if(!all(shape_by_check)){
      stop("The shape_by in not in your sce_list's colData. Please double check the variable name for shape_by.")
    }
  }

  mat_list <- lapply(sce_list, function(x){
    mat <- t(as.matrix(SummarizedExperiment::assay(x, assay_use)))
    mat
  })

  ref_pca_fit <- irlba::prcomp_irlba(mat_ref,
                                     center = center,
                                     scale. = scale.,
                                     n = n_pc)
  ref_pca <- ref_pca_fit$x
  if(pc_umap) {
    ref_umap_fit <- umap::umap(ref_pca_fit$x)
  } else {
    ref_umap_fit <- umap::umap(mat_ref)
  }
  ref_umap <- ref_umap_fit$layout
  colnames(ref_umap) <- c("UMAP1", "UMAP2")

  SingleCellExperiment::reducedDim(ref_sce, "PCA") <- ref_pca
  SingleCellExperiment::reducedDim(ref_sce, "UMAP") <- ref_umap

  sce_list <- lapply(sce_list, function(x) {
    mat <- t(as.matrix(SummarizedExperiment::assay(x, assay_use)))
    SingleCellExperiment::reducedDim(x, "PCA") <- stats::predict(ref_pca_fit, newdata = mat)
    if(pc_umap) {
      res <- stats::predict(object = ref_umap_fit, data = SingleCellExperiment::reducedDim(x, "PCA"))
    } else {
      res <- stats::predict(object = ref_umap_fit, data = mat)
    }
     colnames(res) <- c("UMAP1", "UMAP2")
    SingleCellExperiment::reducedDim(x, "UMAP") <- res
    return(x)
  })

  sce_list_new <- c(list(ref_sce), sce_list)
  names(sce_list_new) <- name_vec

  rd_list <- lapply(sce_list_new, function(x) {
    rd <- tibble::as_tibble(SummarizedExperiment::colData(x))
    if(is.null(shape_by)){
      rd <- dplyr::select(rd, color_by)
    }else{
      rd <- dplyr::select(rd, c(color_by,shape_by))
    }


    rd_pca <- tibble::as_tibble(SingleCellExperiment::reducedDim(x, "PCA"))
    rd_umap <- tibble::as_tibble(SingleCellExperiment::reducedDim(x, "UMAP"))
    rd <- dplyr::bind_cols(rd, rd_pca)
    rd <- dplyr::bind_cols(rd, rd_umap)
    rd
  })

  names(rd_list) <- names(sce_list_new)

  rd_tbl <- dplyr::bind_rows(rd_list, .id = "Method")
  rd_tbl <- dplyr::mutate(rd_tbl, Method = factor(Method, levels = name_vec))

  if(if_plot) {

    aes_args <- list(x = quote(.data[["UMAP1"]]),
                     y = quote(.data[["UMAP2"]]),
                     color = substitute(.data[[color_by]]))
    
    if (!is.null(shape_by)) {
      aes_args$shape <- substitute(.data[[shape_by]])
    }
    
    p_pca <- ggplot(rd_tbl, do.call(aes, aes_args)) +
      geom_point(alpha = 0.5, size = point_size) +
      facet_wrap(~Method, nrow = 1) +
      theme(aspect.ratio = 1, legend.position = "bottom") 
      
    p_umap <- ggplot(rd_tbl, do.call(aes, aes_args)) +
      geom_point(alpha = 0.5, size = point_size) +
      facet_wrap(~Method, nrow = 1) +
      theme(aspect.ratio = 1, legend.position = "bottom") 
      

    if(is.numeric(unlist(rd_tbl[, color_by]))) {
      p_pca <- p_pca + viridis::scale_color_viridis()
      p_umap <- p_umap + viridis::scale_color_viridis()
    } else {
      p_pca <- p_pca + guides(color = guide_legend(override.aes = list(size = 2, alpha = 1)))
      p_umap <- p_umap + guides(color = guide_legend(override.aes = list(size = 2, alpha = 1)))
    }

    return(list(p_pca = p_pca, p_umap = p_umap))

  } else {
    return(rd_tbl)
  }

}

---
title: "scDesign3 Quickstart"
author: 
  - name: Dongyuan Song
    affiliation:
    - Department of Genetics and Genome Sciences, UConn Health, Farmington, CT
    - Bioinformatics IDP, University of California, Los Angeles, CA
    email: dongyuansong@ucla.edu
  - name: Qingyang Wang
    affiliation:
    - Department of Statistics, University of California, Los Angeles, CA
    email: qw802@g.ucla.edu
output: 
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    code_folding: show
date: "`r doc_date()`"
package: "`r pkg_ver('scDesign3')`"
vignette: >
  %\VignetteIndexEntry{scDesign3-quickstart-vignette}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}  
---
```{css, echo=FALSE}
pre {
  white-space: pre !important;
  overflow-x: scroll !important;
}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    message = FALSE,
    collapse = TRUE,
    comment = "#>",
    crop = NULL ## Related to https://stat.ethz.ch/pipermail/bioc-devel/2020-April/016656.html
)
tools::R_user_dir("scDesign3", which="cache")
```

```{r, message=FALSE, warning=FALSE, results='hide'}
library(scDesign3)
library(SingleCellExperiment)
library(ggplot2)
theme_set(theme_bw())
```

## Introduction
scDesign3 is a unified probabilistic framework that generates realistic in silico high-dimensional single-cell omics data of various cell states, including discrete cell types, continuous trajectories, and spatial locations by learning from real datasets. Since the functions of scDesign3 is very comprehensive, here we only introduce how scDesign3 simulates an scRNA-seq dataset with one continuous developmental trajectory. For more information, please check the Articles on our website: (https://songdongyuan1994.github.io/scDesign3/docs/index.html).

## Read in the reference data

The raw data is from the [scvelo](https://scvelo.readthedocs.io/scvelo.datasets.pancreas/), which describes pancreatic endocrinogenesis. We pre-select the top 1000 highly variable genes and filter out some cell types to ensure a **single trajectory**. To save computational time, we only use the top 100 genes.

```{r}
#example_sce <- readRDS((url("https://figshare.com/ndownloader/files/40581992")))
#print(example_sce)
#example_sce <- example_sce[1:100, ]

data("example_count", package = "scDesign3")
data("pseudotime", package = "scDesign3")

example_sce <- SingleCellExperiment::SingleCellExperiment(assays = list(counts = example_count, logcounts = log1p(example_count)),
  colData = DataFrame(pseudotime = pseudotime))
```


## Simulation
The function `scdesign3()` takes in a `SinglecellExperiment` object with the cell covariates (such as cell types, pseudotime, or spatial coordinates) stored in the `colData` of the `SinglecellExperiment` object.
```{r, message=FALSE, warning=FALSE, results='hide'}
set.seed(123)
example_simu <- scdesign3(
    sce = example_sce,
    assay_use = "counts",
    celltype = NULL,
    pseudotime = "pseudotime",
    spatial = NULL,
    other_covariates = NULL,
    mu_formula = "s(pseudotime, k = 10, bs = 'cr')", # mgcg::gam stype formula input.
    sigma_formula = "1", # If you want your dispersion also varies along pseudotime, use "s(pseudotime, k = 5, bs = 'cr')". This will significantly increase the computational cost.
    family_use = "nb",
    n_cores = 2, # Consider the balance between your ROM and threads.
    usebam = FALSE,
    corr_formula = "1",
    copula = "gaussian",
    DT = TRUE,
    pseudo_obs = FALSE,
    return_model = FALSE,
    nonzerovar = FALSE
  )
```
The output of `scdesign3()` is a list which includes:

* `new_count`: This is the synthetic count matrix generated by `scdesign3()`.
* `new_covariate`: 
  + If the parameter `ncell` is set to a number that is different from the number of cells in the input data, this will be a matrix that has the new cell covariates that are used for generating new data.
  + If the parameter `ncell` is the default value, this will be `NULL`.
* `model_aic`: This is a vector include the genes' marginal models' AIC, fitted copula's AIC, and total AIC, which is the sum of the previous two AIC. 
* `model_bic`: This is a vector include the genes' marginal models' BIC, fitted copula's BIC, and total BIC, which is the sum of the previous two BIC.
* `marginal_list`:
  + If the parameter `return_model` is set to `TRUE`, this will be a list which contains the fitted gam or gamlss model for all genes in the input data.
  + If the parameter `return_model` is set to the default value `FALSE`, this will be `NULL`.
* `corr_list`:
  + If the parameter `return_model` is set to `TRUE`, this will be a list which contains the either a correlation matrix (when `copula = "gaussian"`) or the fitted vine copula (when `copula = "vine`) for each user specified correlation groups (based on the parameter `corr_by`).
  + If the parameter `return_model` is set to the default value `FALSE`, this will be `NULL`.

In this example, since we did not change the parameter `ncell`, the synthetic count matrix will have the same dimension as the input count matrix. 
```{r}
dim(example_simu$new_count)
```
Then, we can create the `SinglecellExperiment` object using the synthetic count matrix and store the `logcounts` to the input and synthetic `SinglecellExperiment` objects.
```{r}
logcounts(example_sce) <- log1p(counts(example_sce))
simu_sce <- SingleCellExperiment(list(counts = example_simu$new_count), colData = example_simu$new_covariate)
logcounts(simu_sce) <- log1p(counts(simu_sce))
```

## Visualization
```{r}
set.seed(123)
compare_figure <- plot_reduceddim(ref_sce = example_sce, 
                                  sce_list = list(simu_sce), 
                                  name_vec = c("Reference", "scDesign3"),
                                  assay_use = "logcounts", 
                                  if_plot = TRUE, 
                                  color_by = "pseudotime", 
                                  n_pc = 20)
plot(compare_figure$p_umap)
```

## Session information
```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simu_new.R
\name{simu_new}
\alias{simu_new}
\title{Simulate new data}
\usage{
simu_new(
  sce,
  assay_use = "counts",
  mean_mat,
  sigma_mat,
  zero_mat,
  quantile_mat = NULL,
  copula_list,
  n_cores,
  fastmvn = FALSE,
  family_use,
  nonnegative = TRUE,
  nonzerovar = FALSE,
  input_data,
  new_covariate,
  important_feature = "all",
  parallelization = "mcmapply",
  BPPARAM = NULL,
  filtered_gene
)
}
\arguments{
\item{sce}{A \code{SingleCellExperiment} object.}

\item{assay_use}{A string which indicates the assay you will use in the sce. Default is 'counts'.}

\item{mean_mat}{A cell by feature matrix of the mean parameter.}

\item{sigma_mat}{A cell by feature matrix of the sigma parameter.}

\item{zero_mat}{A cell by feature matrix of the zero-inflation parameter.}

\item{quantile_mat}{A cell by feature matrix of the multivariate quantile.}

\item{copula_list}{A list of copulas for generating the multivariate quantile matrix. If provided, the \code{quantile_mat} must be NULL.}

\item{n_cores}{An integer. The number of cores to use.}

\item{fastmvn}{An logical variable. If TRUE, the sampling of multivariate Gaussian is done by \code{mvnfast}, otherwise by \code{mvtnorm}. Default is FALSE.}

\item{family_use}{A string of the marginal distribution.
Must be one of 'poisson', "binomial", 'nb', 'zip', 'zinb' or 'gaussian'.}

\item{nonnegative}{A logical variable. If TRUE, values < 0 in the synthetic data will be converted to 0. Default is TRUE (since the expression matrix is nonnegative).}

\item{nonzerovar}{A logical variable. If TRUE, for any gene with zero variance, a cell will be replaced with 1. This is designed for avoiding potential errors, for example, PCA.}

\item{input_data}{A input count matrix.}

\item{new_covariate}{A data.frame which contains covariates of targeted simulated data from  \code{\link{construct_data}}.}

\item{important_feature}{important_feature A string or vector which indicates whether a gene will be used in correlation estimation or not. If this is a string, then
this string must be either "all" (using all genes) or "auto", which indicates that the genes will be automatically selected based on the proportion of zero expression across cells
for each gene. Gene with zero proportion greater than 0.8 will be excluded form gene-gene correlation estimation. If this is a vector, then this should
be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
The default value for is "all".}

\item{parallelization}{A string indicating the specific parallelization function to use.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}

\item{filtered_gene}{A vector or NULL which contains genes that are excluded in the marginal and copula fitting 
steps because these genes only express in less than two cells. This can be obtain from  \code{\link{construct_data}}}
}
\value{
A feature by cell matrix of the new simulated count (expression) matrix or sparse matrix.
}
\description{
\code{simu_new} generates new simulated data based on fitted marginal and copula models.
}
\details{
The function takes the new covariate (if use) from \code{\link{construct_data}},
parameter matrices from \code{\link{extract_para}} and multivariate Unifs from \code{\link{fit_copula}}.
}
\examples{
  data(example_sce)
  my_data <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "1"
  )
  my_marginal <- fit_marginal(
  data = my_data,
  mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
  )
  my_copula <- fit_copula(
  sce = example_sce,
  assay_use = "counts",
  marginal_list = my_marginal,
  family_use = c(rep("nb", 5), rep("zip", 5)),
  copula = "vine",
  n_cores = 1,
  input_data = my_data$dat
  )
  my_para <- extract_para(
    sce = example_sce,
    marginal_list = my_marginal,
    n_cores = 1,
    family_use = c(rep("nb", 5), rep("zip", 5)),
    new_covariate = my_data$new_covariate,
    data = my_data$dat
  )
  my_newcount <- simu_new(
  sce = example_sce,
  mean_mat = my_para$mean_mat,
  sigma_mat = my_para$sigma_mat,
  zero_mat = my_para$zero_mat,
  quantile_mat = NULL,
  copula_list = my_copula$copula_list,
  n_cores = 1,
  family_use = c(rep("nb", 5), rep("zip", 5)),
  input_data = my_data$dat,
  new_covariate = my_data$new_covariate,
  important_feature = my_copula$important_feature,
  filtered_gene = my_data$filtered_gene
  )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamlss_fix.R
\name{ga}
\alias{ga}
\title{Functions from gamlss/gamlss.add with bugs fixed}
\usage{
ga(formula, envir, control = ga.control(...), ...)
}
\arguments{
\item{formula}{A formula of the model.}

\item{envir}{The environment.}

\item{control}{The control of the model fitting.}

\item{...}{Other arguments.}
}
\value{
A xvar list.
}
\description{
An additive function to be used while fitting GAMLSS models. The interface for \code{gam()} in the \pkg{mgcv} package.
}
\section{ga}{
NA
}

\examples{
print("No example")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdesign3.R
\name{scdesign3}
\alias{scdesign3}
\title{The wrapper for the whole scDesign3 pipeline}
\usage{
scdesign3(
  sce,
  assay_use = "counts",
  celltype,
  pseudotime = NULL,
  spatial = NULL,
  other_covariates,
  ncell = dim(sce)[2],
  mu_formula,
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 2,
  correlation_function = "default",
  usebam = FALSE,
  edf_flexible = FALSE,
  corr_formula,
  empirical_quantile = FALSE,
  copula = "gaussian",
  if_sparse = FALSE,
  fastmvn = FALSE,
  DT = TRUE,
  pseudo_obs = FALSE,
  family_set = c("gauss", "indep"),
  important_feature = "all",
  nonnegative = TRUE,
  nonzerovar = FALSE,
  return_model = FALSE,
  simplify = FALSE,
  parallelization = "mcmapply",
  n_rep = 1,
  BPPARAM = NULL,
  trace = FALSE
)
}
\arguments{
\item{sce}{A \code{SingleCellExperiment} object.}

\item{assay_use}{A string which indicates the assay you will use in the sce. Default is 'counts'.}

\item{celltype}{A string of the name of cell type variable in the \code{colData} of the sce. Default is 'cell_type'.}

\item{pseudotime}{A string or a string vector of the name of pseudotime and (if exist)
multiple lineages. Default is NULL.}

\item{spatial}{A length two string vector of the names of spatial coordinates. Default is NULL.}

\item{other_covariates}{A string or a string vector of the other covariates you want to include in the data.}

\item{ncell}{The number of cell you want to simulate. Default is \code{dim(sce)[2]} (the same number as the input data).}

\item{mu_formula}{A string of the mu parameter formula}

\item{sigma_formula}{A string of the sigma parameter formula}

\item{family_use}{A string of the marginal distribution.
Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian'.}

\item{n_cores}{An integer. The number of cores to use.}

\item{correlation_function}{A string. If 'default', the function from \code{Rfast}; if 'coop', the function from \code{coop}, which calls BLAS.}

\item{usebam}{A logic variable. If use \code{\link[mgcv]{bam}} for acceleration in marginal fitting.}

\item{edf_flexible}{A logic variable. It is used for accelerating for spatial model if k is large in 'mu_formula'. Default is FALSE.}

\item{corr_formula}{A string of the correlation structure.}

\item{empirical_quantile}{Please only use it if you clearly know what will happen! A logic variable. If TRUE, DO NOT fit the copula and use the EMPIRICAL CDF values of the original data; it will make the simulated data fixed (no randomness). Default is FALSE. Only works if ncell is the same as your original data.}

\item{copula}{A string of the copula choice. Must be one of 'gaussian' or 'vine'. Default is 'gaussian'. Note that vine copula may have better modeling of high-dimensions, but can be very slow when features are >1000.}

\item{if_sparse}{A logic variable. Only works for Gaussian copula (\code{family_set = "gaussian"}). If TRUE, a thresholding strategy will make the corr matrix sparse.}

\item{fastmvn}{An logical variable. If TRUE, the sampling of multivariate Gaussian is done by \code{mvnfast}, otherwise by \code{mvtnorm}. Default is FALSE. It only matters for Gaussian copula.}

\item{DT}{A logic variable. If TRUE, perform the distributional transformation
to make the discrete data 'continuous'. This is useful for discrete distributions (e.g., Poisson, NB).
Default is TRUE. Note that for continuous data (e.g., Gaussian), DT does not make sense and should be set as FALSE.}

\item{pseudo_obs}{A logic variable. If TRUE, use the empirical quantiles instead of theoretical quantiles for fitting copula.
Default is FALSE.}

\item{family_set}{A string or a string vector of the bivariate copula families. Default is c("gauss", "indep"). For more information please check package \code{rvinecoplib}.}

\item{important_feature}{A numeric value or vector which indicates whether a gene will be used in correlation estimation or not. If this is a numeric value, then
gene with zero proportion greater than this value will be excluded form gene-gene correlation estimation. If this is a vector, then this should
be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
The default value is "all" (a special string which means no filtering).}

\item{nonnegative}{A logical variable. If TRUE, values < 0 in the synthetic data will be converted to 0. Default is TRUE (since the expression matrix is nonnegative).}

\item{nonzerovar}{A logical variable. If TRUE, for any gene with zero variance, a cell will be replaced with 1. This is designed for avoiding potential errors, for example, PCA. Default is FALSE.}

\item{return_model}{A logic variable. If TRUE, the marginal models and copula models will be returned. Default is FALSE.}

\item{simplify}{A logic variable. If TRUE, the fitted regression model will only keep the essential contains for \code{predict}, otherwise the fitted models can be VERY large. Default is FALSE.}

\item{parallelization}{A string indicating the specific parallelization function to use.}

\item{n_rep}{An integer number. The number of replicates of simulated new count matrix. Default is 1.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}

\item{trace}{A logic variable. If TRUE, the warning/error log and runtime for gam/gamlss
will be returned, FALSE otherwise. Default is FALSE.}
}
\value{
A list with the components:
\describe{
  \item{\code{new_count}}{A matrix of the new simulated count (expression) matrix.}
  \item{\code{new_covariate}}{A data.frame of the new covariate matrix.}
  \item{\code{model_aic}}{The model AIC.}
  \item{\code{marginal_list}}{A list of marginal regression models if return_model = TRUE.}
  \item{\code{corr_list}}{A list of correlation models (conditional copulas) if return_model = TRUE.}
}
}
\description{
\code{scdesign3} takes the input data, fits the model and
}
\examples{
data(example_sce)
my_simu <- scdesign3(
sce = example_sce,
assay_use = "counts",
celltype = "cell_type",
pseudotime = "pseudotime",
spatial = NULL,
other_covariates = NULL,
mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
sigma_formula = "1",
family_use = "nb",
n_cores = 2,
usebam = FALSE,
edf_flexible = FALSE,
corr_formula = "pseudotime",
copula = "gaussian",
if_sparse = TRUE,
DT = TRUE,
pseudo_obs = FALSE,
ncell = 1000,
return_model = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_para.R
\name{extract_para}
\alias{extract_para}
\title{Extract the parameters of each cell's distribution}
\usage{
extract_para(
  sce,
  assay_use = "counts",
  marginal_list,
  n_cores,
  family_use,
  new_covariate,
  parallelization = "mcmapply",
  BPPARAM = NULL,
  data
)
}
\arguments{
\item{sce}{A \code{SingleCellExperiment} object.}

\item{assay_use}{A string which indicates the assay you will use in the sce. Default is 'counts'.}

\item{marginal_list}{A list of fitted regression models from \code{\link{fit_marginal}} for each gene in sce.}

\item{n_cores}{An integer. The number of cores to use.}

\item{family_use}{A string of the marginal distribution.
Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian', which represent 'poisson distribution',
'negative binomial distribution', 'zero-inflated poisson distribution', 'zero-inflated negative binomial distribution',
and 'gaussian distribution' respectively.}

\item{new_covariate}{A data.frame which contains covariates of targeted simulated data from  \code{\link{construct_data}} and the
correlation group assignment for each cell in the column 'corr_group'.}

\item{parallelization}{A string indicating the specific parallelization function to use.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}

\item{data}{A dataframe which is used when fitting the gamlss model}
}
\value{
A list with the components:
\describe{
  \item{\code{mean_mat}}{A cell by feature matrix of the mean parameter.}
  \item{\code{sigma_mat}}{A cell by feature matrix of the sigma parameter (for Gaussian, the variance; for NB, the dispersion.).}
  \item{\code{zero_mat}}{A cell by feature matrix of the zero-inflation parameter (only non-zero for ZIP and ZINB).}
}
}
\description{
\code{extract_para} generates parameter matrices which determine each cell's distribution
}
\details{
The function takes the new covariate (if use) from \code{\link{construct_data}} and
marginal models from \code{\link{fit_marginal}}.
}
\examples{
  data(example_sce)
  my_data <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "1"
  )
  my_marginal <- fit_marginal(
  data = my_data,
  mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
  )
  my_copula <- fit_copula(
  sce = example_sce,
  assay_use = "counts",
  marginal_list = my_marginal,
  family_use = c(rep("nb", 5), rep("zip", 5)),
  copula = "vine",
  n_cores = 1,
  input_data = my_data$dat
  )
  my_para <- extract_para(
    sce = example_sce,
    marginal_list = my_marginal,
    n_cores = 1,
    family_use = c(rep("nb", 5), rep("zip", 5)),
    new_covariate = my_data$new_covariate,
    data = my_data$dat
  )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_cov.R
\name{sparse_cov}
\alias{sparse_cov}
\title{This function computes the thresholding sparse covariance/correlation estimator 
with the optimal threshold level.}
\usage{
sparse_cov(
  data,
  method = c("cv", "qiu"),
  operator = c("hard", "soft", "scad", "al"),
  corr = TRUE
)
}
\arguments{
\item{data}{The data matrix.}

\item{method}{The choice of method to select the optimal threshold level.}

\item{operator}{The choice of the thresholding operator.}

\item{corr}{The indicator of computing correlation or covariance matrix.}
}
\value{
The thresholding sparse covariance/correlation estimator.
}
\description{
Part from Chenxin Jiang
}
\examples{
print("No example")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/construct_data.R
\name{construct_data}
\alias{construct_data}
\title{Construct the input data (covariate matrix and expression matrix)}
\usage{
construct_data(
  sce,
  assay_use = "counts",
  celltype,
  pseudotime,
  spatial,
  other_covariates,
  ncell = dim(sce)[2],
  corr_by,
  parallelization = "mcmapply",
  BPPARAM = NULL
)
}
\arguments{
\item{sce}{A \code{SingleCellExperiment} object.}

\item{assay_use}{A string which indicates the assay you will use in the sce. Default is 'counts'.}

\item{celltype}{A string of the name of cell type variable in the \code{colData} of the sce. Default is 'cell_type'.}

\item{pseudotime}{A string or a string vector of the name of pseudotime and (if exist)
multiple lineages. Default is NULL.}

\item{spatial}{A length two string vector of the names of spatial coordinates. Default is NULL.}

\item{other_covariates}{A string or a string vector of the other covariates you want to include in the data.}

\item{ncell}{The number of cell you want to simulate. Default is \code{dim(sce)[2]} (the same number as the input data).
If an arbitrary number is provided, the function will use Vine Copula to simulate a new covariate matrix.}

\item{corr_by}{A string or a string vector which indicates the groups for correlation structure. If '1', all cells have one estimated corr. If 'ind', no corr (features are independent). If others, this variable decides the corr structures.}

\item{parallelization}{A string indicating the specific parallelization function to use.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}
}
\value{
A list with the components:
\describe{
  \item{\code{count_mat}}{The expression matrix}
  \item{\code{dat}}{The original covariate matrix}
  \item{\code{newCovariate}}{The simulated new covariate matrix, is NULL if the parameter ncell is default}
  \item{\code{filtered_gene}}{The genes that are excluded in the marginal and copula fitting 
steps because these genes only express in less than two cells.}
}
}
\description{
This function constructs the input data for \code{\link{fit_marginal}}.
}
\details{
This function takes a \code{SingleCellExperiment} object as the input.
Based on users' choice, it constructs the matrix of covariates
(explanatory variables) and the expression matrix (e.g., count matrix for scRNA-seq).
}
\examples{
  data(example_sce)
  my_data <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "1"
  )


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_copula.R
\name{fit_copula}
\alias{fit_copula}
\title{Fit the copula model}
\usage{
fit_copula(
  sce,
  assay_use,
  input_data,
  empirical_quantile = FALSE,
  marginal_list,
  family_use,
  copula = "gaussian",
  DT = TRUE,
  pseudo_obs = FALSE,
  epsilon = 1e-06,
  family_set = c("gaussian", "indep"),
  important_feature = "all",
  if_sparse = FALSE,
  correlation_function = "default",
  n_cores,
  parallelization = "mcmapply",
  BPPARAM = NULL
)
}
\arguments{
\item{sce}{A \code{SingleCellExperiment} object.}

\item{assay_use}{A string which indicates the assay you will use in the sce.
Default is 'counts'.}

\item{input_data}{The input data, which is one of the output from \code{\link{construct_data}}.}

\item{empirical_quantile}{Please only use it if you clearly know what will happen! A logic variable. If TRUE, DO NOT fit the copula and use the EMPIRICAL CDF values of the original data; it will make the simulated data fixed (no randomness). Default is FALSE. Only works if ncell is the same as your original data.}

\item{marginal_list}{A list of fitted regression models from \code{\link{fit_marginal}}.}

\item{family_use}{A string or a vector of strings of the marginal distribution. Must be one of 'poisson', 'nb', 'zip', 'zinb' or 'gaussian'.}

\item{copula}{A string of the copula choice. Must be one of 'gaussian' or 'vine'. Default is 'gaussian'. Note that vine copula may have better modeling of high-dimensions, but can be very slow when features are >1000.}

\item{DT}{A logic variable. If TRUE, perform the distributional transformation
to make the discrete data 'continuous'. This is useful for discrete distributions (e.g., Poisson, NB).
Default is TRUE. Note that for continuous data (e.g., Gaussian), DT does not make sense and should be set as FALSE.}

\item{pseudo_obs}{A logic variable. If TRUE, use the empirical quantiles instead of theoretical quantiles for fitting copula.
Default is FALSE.}

\item{epsilon}{A numeric variable for preventing the transformed quantiles to collapse to 0 or 1.}

\item{family_set}{A string or a string vector of the bivariate copula families. Default is c("gaussian", "indep").}

\item{important_feature}{A numeric value or vector which indicates whether a gene will be used in correlation estimation or not. If this is a numeric value, then
gene with zero proportion greater than this value will be excluded form gene-gene correlation estimation. If this is a vector, then this should
be a logical vector with length equal to the number of genes in \code{sce}. \code{TRUE} in the logical vector means the corresponding gene will be included in
gene-gene correlation estimation and \code{FALSE} in the logical vector means the corresponding gene will be excluded from the gene-gene correlation estimation.
The default value for is "all" (a special string which means no filtering).}

\item{if_sparse}{A logic variable. Only works for Gaussian copula (\code{family_set = "gaussian"}). If TRUE, a thresholding strategy will make the corr matrix sparse.}

\item{correlation_function}{A string. If 'default', the function from \code{Rfast}; if 'coop', the function from \code{coop}, which calls BLAS.}

\item{n_cores}{An integer. The number of cores to use.}

\item{parallelization}{A string indicating the specific parallelization function to use.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}
}
\value{
A list with the components:
\describe{
  \item{\code{new_mvu}}{A matrix of the new multivariate uniform distribution from the copula.}
  \item{\code{copula_list}}{A list of the fitted copula model. If using Gaussian copula, a list of correlation matrices; if vine, a list of vine objects.}
  \item{\code{model_aic}}{A vector of the marginal AIC and the copula AIC.}
  \item{\code{model_bic}}{A vector of the marginal BIC and the copula BIC.}
}
}
\description{
\code{fit_copula} fits the copula model.
}
\details{
This function takes the result from \code{\link{fit_marginal}} as the input and
and fit the copula model on the residuals.
}
\examples{
  data(example_sce)
  my_data <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "1"
  )
  my_marginal <- fit_marginal(
  data = my_data,
  mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
  )
  my_copula <- fit_copula(
  sce = example_sce,
  assay_use = "counts",
  marginal_list = my_marginal,
  family_use = c(rep("nb", 5), rep("zip", 5)),
  copula = "vine",
  n_cores = 1,
  input_data = my_data$dat
  )
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example_count}
\alias{example_count}
\title{A sparse matrix with example data}
\format{
A dgCMatrix with 100 rows (genes) and 2087 cols (cells)
}
\usage{
data("example_count")
}
\value{
The corresponding sparse matrix object 'example_count'
}
\description{
A sparse matrix with example data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/perform_lrt.R
\name{perform_lrt}
\alias{perform_lrt}
\title{Perform the likelihood ratio test}
\usage{
perform_lrt(alter_marginal, null_marginal)
}
\arguments{
\item{alter_marginal}{A list of marginal models from the alternative hypothesis.}

\item{null_marginal}{A list of marginal models from the null hypothesis. It must be strictly nested in the alternative model.}
}
\value{
A data.frame of the LRT result.
}
\description{
\code{perform_lrt} performs the likelihood ratio test to compare two list of marginal models.
}
\details{
The function takes two lists of marginal models (by default, the first list is the alternative and the second is the null)
from \code{\link{fit_marginal}}. Note that LRT only makes sense for NESTED models. This can be quite tricky if you use penalized-splines (e.g., for trajectory data).
}
\examples{
data(example_sce)
my_data <- construct_data(
sce = example_sce,
assay_use = "counts",
celltype = "cell_type",
pseudotime = "pseudotime",
spatial = NULL,
other_covariates = NULL,
corr_by = "cell_type"
)

my_data2 <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "pseudotime",
  ncell = 10000
)

my_marginal1 <- fit_marginal(
  data = my_data,
  mu_formula = "1",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
)
my_marginal2 <- fit_marginal(
  data = my_data,
  mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
)
my_fit1 <- lapply(my_marginal1, function(x)x$fit)
my_fit2 <- lapply(my_marginal2, function(x)x$fit)
my_pvalue <- perform_lrt(my_fit2, my_fit1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_marginal.R
\name{fit_marginal}
\alias{fit_marginal}
\title{Fit the marginal models}
\usage{
fit_marginal(
  data,
  predictor = "gene",
  mu_formula,
  sigma_formula,
  family_use,
  n_cores,
  usebam = FALSE,
  edf_flexible = FALSE,
  parallelization = "mcmapply",
  BPPARAM = NULL,
  trace = FALSE,
  simplify = FALSE,
  filter_cells = FALSE
)
}
\arguments{
\item{data}{An object from \code{\link{construct_data}}.}

\item{predictor}{A string of the predictor for the gam/gamlss model. Default is "gene". This is just a name.}

\item{mu_formula}{A string of the mu parameter formula. It follows the format of formula in \code{\link[mgcv]{bam}}. Note: if the formula has multiple smoothers (\code{s()}) (we do not recommend this), please put the one with largest k (most complex one) as the first one.}

\item{sigma_formula}{A string of the sigma parameter formula}

\item{family_use}{A string or a vector of strings of the marginal distribution.
Must be one of 'binomial', 'poisson', 'nb', 'zip', 'zinb' or 'gaussian', which represent 'poisson distribution',
'negative binomial distribution', 'zero-inflated poisson distribution', 'zero-inflated negative binomial distribution',
and 'gaussian distribution' respectively.}

\item{n_cores}{An integer. The number of cores to use.}

\item{usebam}{A logic variable. If use \code{\link[mgcv]{bam}} for acceleration.}

\item{edf_flexible}{A logic variable. It uses simpler model to accelerate the marginal fitting with a mild loss of accuracy. If TRUE, the fitted regression model will use the fitted relationship between Gini coefficient and the effective degrees of freedom on a random selected gene sets. Default is FALSE.}

\item{parallelization}{A string indicating the specific parallelization function to use.
Must be one of 'mcmapply', 'bpmapply', or 'pbmcmapply', which corresponds to the parallelization function in the package
\code{parallel},\code{BiocParallel}, and \code{pbmcapply} respectively. The default value is 'mcmapply'.}

\item{BPPARAM}{A \code{MulticoreParam} object or NULL. When the parameter parallelization = 'mcmapply' or 'pbmcmapply',
this parameter must be NULL. When the parameter parallelization = 'bpmapply',  this parameter must be one of the
\code{MulticoreParam} object offered by the package 'BiocParallel. The default value is NULL.}

\item{trace}{A logic variable. If TRUE, the warning/error log and runtime for gam/gamlss will be returned.
will be returned, FALSE otherwise. Default is FALSE.}

\item{simplify}{A logic variable. If TRUE, the fitted regression model will only keep the essential contains for \code{predict}. Default is FALSE.}

\item{filter_cells}{A logic variable. If TRUE, when all covariates used for fitting the GAM/GAMLSS model are categorical, the code will check each unique combination of categories and remove cells in that category if it has all zero gene expression for each fitted gene.}
}
\value{
A list of fitted regression models. The length is equal to the total feature number.
}
\description{
\code{fit_marginal} fits the per-feature regression models.
}
\details{
The function takes the result from \code{\link{construct_data}} as the input,
and fit the regression models for each feature based on users' specification.
}
\examples{
  data(example_sce)
  my_data <- construct_data(
  sce = example_sce,
  assay_use = "counts",
  celltype = "cell_type",
  pseudotime = "pseudotime",
  spatial = NULL,
  other_covariates = NULL,
  corr_by = "1"
  )
  my_marginal <- fit_marginal(
  data = my_data,
  mu_formula = "s(pseudotime, bs = 'cr', k = 10)",
  sigma_formula = "1",
  family_use = "nb",
  n_cores = 1,
  usebam = FALSE
  )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example_sce}
\alias{example_sce}
\title{A SingleCellExperiment object containing both cell type and pseudotime}
\format{
A dataset with 10 rows (genes) and 1289 cols (cells)
}
\usage{
data("example_sce")
}
\value{
The corresponding SingleCellExperiment object
}
\description{
A SingleCellExperiment object containing both cell type and pseudotime
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamlss_fix.R
\name{ba}
\alias{ba}
\title{Functions from gamlss/gamlss.add with bugs fixed}
\usage{
ba(formula, control = ba.control(...), ...)
}
\arguments{
\item{formula}{A formula of the model.}

\item{control}{The control of the model fitting.}

\item{...}{Other arguments.}
}
\value{
A xvar list.
}
\description{
An additive function to be used while fitting GAMLSS models. The interface for \code{bam()} in the \pkg{mgcv} package.
}
\section{ba}{
NA
}

\examples{
print("No example")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_reduceddim.R
\name{plot_reduceddim}
\alias{plot_reduceddim}
\title{Dimensionality reduction and visualization}
\usage{
plot_reduceddim(
  ref_sce,
  sce_list,
  name_vec,
  assay_use = "logcounts",
  pc_umap = TRUE,
  n_pc = 50,
  center = TRUE,
  scale. = TRUE,
  if_plot = TRUE,
  shape_by = NULL,
  color_by,
  point_size = 1
)
}
\arguments{
\item{ref_sce}{The reference sce.}

\item{sce_list}{A list of synthetic sce.}

\item{name_vec}{A vector of the names of each dataset. The length should be \code{length(sce_list) + 1}, where the first name is for \code{ref_sce}.}

\item{assay_use}{A string which indicates the assay you will use in the sce.
Default is 'logcounts'.}

\item{pc_umap}{A logic value of whether using PCs as the input of UMAP. Default is TRUE.}

\item{n_pc}{An integer of the number of PCs.}

\item{center}{A logic value of whether centering the data before PCA. Default is TRUE.}

\item{scale.}{A logic value of whether scaling the data before PCA. Default is TRUE.}

\item{if_plot}{A logic value of whether returning the plot. If FALSE, return the reduced dimensions of each dataset.}

\item{shape_by}{A string which indicates the column in \code{colData} used for shape.}

\item{color_by}{A string which indicates the column in \code{colData} used for color.}

\item{point_size}{A numeric value of the point size in the final plot. Default is 1.}
}
\value{
The ggplot or the data.frame of reduced dimensions.
}
\description{
\code{plot_reduceddim} performs the dimensionality reduction
}
\details{
This function takes a reference sce and a list of new sces, performs the dimensionality reduction on the reference data, 
projects the synthetic datasets on the same low dimensional space,
then visualize the results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamlss_fix.R
\name{gamlss.ga}
\alias{gamlss.ga}
\title{Support for Function ga()}
\usage{
gamlss.ga(x, y, w, xeval = NULL, ...)
}
\arguments{
\item{x}{The explanatory variables}

\item{y}{Iterative y variable}

\item{w}{Iterative weights}

\item{xeval}{If xeval=TRUE then prediction is used}

\item{...}{Other arguments}
}
\value{
Not used
}
\description{
This is support for the  smoother functions \code{ga()} interfaces for Simon Wood's \code{gam()} functions from package \pkg{mgcv}. It is not intended to be called directly by users. From \code{gamlss.add::gamlss.ga}.
}
\examples{
print("No example")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{pseudotime}
\alias{pseudotime}
\title{A numeric vector of pseudotime values}
\format{
A numeric vector with 2087 values
}
\usage{
data("pseudotime")
}
\value{
The corresponding pseudotime of the 'example_count'
}
\description{
A numeric vector of pseudotime values
}
\keyword{datasets}
